/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.TreeSet;
import java.util.Vector;

import at.icbi.gpviz.data.DataChangedEvent.CHANGEACTION;
import at.icbi.gpviz.data.DataChangedEvent.WHATCHANGED;
import at.icbi.gpviz.data.loader.anno.EnsemblUniprotLoader;
import at.icbi.gpviz.data.loader.anno.RefseqUniprotLoader;
import at.icbi.gpviz.data.regions.GPVHighlightedRegion;
import at.icbi.gpviz.data.regions.GPVHighlightedRegionSet;
import at.icbi.gpviz.data.regions.GPVVariant;
import at.icbi.gpviz.data.regions.GPVVariantConstants;
import at.icbi.gpviz.data.sorting.AlphabeticalGeneSorter;
import at.icbi.gpviz.data.util.GPVUtils;
import at.tugraz.genome.biojava.seq.DNAStrand;

public class GPVDataContainer
{
  private HashMap<String,GPVGene> genes = new HashMap<String,GPVGene>();

  private TreeSet<GPVChromosome> discoveredChromosomes = new TreeSet<GPVChromosome>();
  
  private ArrayList<GPVHighlightedRegionSet> highlightedRegionSets = new ArrayList<GPVHighlightedRegionSet>();
  
  private ArrayList<DataChangedListener> changeListeners = new ArrayList<DataChangedListener>();
  
  // The highlighted regions in here are stored REDUNDANTLY to the highligtedRegions field. 
  // This is just a cache that should NOT be manipulated from outside and needs to be refreshed
  // whenever the gene list or region list changes! It is necessary to speed up drawing the image,
  // because otherwise we'd need to search for the regions every time we draw an image for a gene
  private HashMap<String, TreeSet<GPVHighlightedRegion>> regionsPerChromosomeCache = new HashMap<String,TreeSet<GPVHighlightedRegion>>();
  private HashSet<String> unknownVariantTypes = new HashSet<String>();

  private HashSet<String> additionalDomainSetInfoKeys = new HashSet<String>();
  
  private ArrayList<GPVDomainSet> domainSets = new ArrayList<GPVDomainSet>();
  
  public enum GENE_SOURCE {
    REFSEQ,
    ENSEMBL
  }
  
  
  public enum DOMAIN_SOURCE {
    UNIPROT,
    GENERIC
  }
  
  
  private GENE_SOURCE source = null;
  private DOMAIN_SOURCE domainSource = DOMAIN_SOURCE.GENERIC;
  
  public HashMap<String,GPVGene> getGenes()
  {
    return genes;
  }

  public void setGenes(HashMap<String,GPVGene> genes)
  {
    this.genes = genes;
    determineGeneIDSource();
    refreshRegionsPerChromosomeCache();
    loadSourceSpecificAnnotations();
    fireDataChangeListeners(WHATCHANGED.EXONS,CHANGEACTION.LOADED);
  }
  
  private void determineGeneIDSource(){
    if(this.genes.size()>0){
      String some_gene_id = this.genes.entrySet().iterator().next().getKey();
      if(some_gene_id.toUpperCase().startsWith("ENSG")){
        //this is from the ENSEMBL database
        this.source  = GENE_SOURCE.ENSEMBL;
      }else{
        this.source  = GENE_SOURCE.REFSEQ;
      }
    }
  }
  
  private void loadSourceSpecificAnnotations(){
    if(source == GENE_SOURCE.REFSEQ){
      new RefseqUniprotLoader().loadRefseqAndUniprotData(null);      
    }else if(source == GENE_SOURCE.ENSEMBL){
      new EnsemblUniprotLoader().loadEnsemblAndUniprotData(null);
    }
  }

  public Vector<String> getGeneNameVector(){
    Vector<String> ret = new Vector<String>(new TreeSet<String>(genes.keySet()));
    return ret;
  }
  
  public Vector<GPVGene> getSortedGeneVector(){
    return getSortedGeneVector(new AlphabeticalGeneSorter(GPVGene.DISPLAYNAME.SYMBOL, false));
  }
  
  
  public HashSet<String> getUnknownVariantTypes()
  {
    return unknownVariantTypes;
  }

  
  public Vector<GPVGene> getSortedGeneVector(Comparator<GPVGene> comp){
    TreeSet<GPVGene> sorted =new TreeSet<GPVGene>(comp);
    sorted.addAll(genes.values());
    Vector<GPVGene> ret = new Vector<GPVGene>(sorted);
    return ret;
  }
  
  
  public GPVChromosome getChromosomeByName(String name){
    for(GPVChromosome c:getDiscoveredChromosomes()){
      if(c.getTrimmedName().toLowerCase().equals(name.toLowerCase())){
        return c;
      }
    }
    //none found. create new
    GPVChromosome ret = new GPVChromosome(name);  
    discoveredChromosomes.add(ret);
    return ret;
  }
  
  public TreeSet<GPVChromosome> getDiscoveredChromosomes()
  {
    return discoveredChromosomes;
  }

  public void setDiscoveredChromosomes(TreeSet<GPVChromosome> discoveredChromosomes)
  {
    this.discoveredChromosomes = discoveredChromosomes;
    fireDataChangeListeners(WHATCHANGED.CHROMOSOMES,CHANGEACTION.LOADED);
  }

  public void addHighlightedRegions(GPVHighlightedRegionSet set){
    this.highlightedRegionSets.add(set);
    refreshRegionsPerChromosomeCache();
    refreshUnknownVariantsFromList(set.getRegions());
    fireDataChangeListeners(WHATCHANGED.REGIONS,CHANGEACTION.LOADED);
  }
  public void addHighlightedRegions(Collection<GPVHighlightedRegionSet> sets){
    for(GPVHighlightedRegionSet set:sets){
      this.highlightedRegionSets.add(set);
      refreshUnknownVariantsFromList(set.getRegions());
    }
    refreshRegionsPerChromosomeCache();
    fireDataChangeListeners(WHATCHANGED.REGIONS,CHANGEACTION.LOADED);
  }
  
  
  
  public GPVHighlightedRegionSet getHighlightedRegionsSet(String name, boolean create_if_not_exist){
    GPVHighlightedRegionSet ret = null;
    for(GPVHighlightedRegionSet set:this.highlightedRegionSets){
      if(set.getSetName().equals(name)){
        ret = set; break;
      }
    }
    if(ret==null){
      ret = new GPVHighlightedRegionSet();
      ret.setSetName(name);
      highlightedRegionSets.add(ret);
    }
    
    return ret;
  }
  
  public ArrayList<GPVHighlightedRegionSet> getHighlightedRegionSets()
  {
    return highlightedRegionSets;
  }

  public void removeHighlightedRegionsSet(String name){
    GPVHighlightedRegionSet todelete = getHighlightedRegionsSet(name,false);
    if(todelete!=null)this.highlightedRegionSets.remove(todelete);
    refreshRegionsPerChromosomeCache();
    refreshUnknownVariantTypeList();
    fireDataChangeListeners(WHATCHANGED.REGIONS,CHANGEACTION.PURGED);
  }
  
  public void addRegionToHighlight(GPVChromosome chrom, int start, int stop, DNAStrand strand){
    GPVHighlightedRegionSet set = getHighlightedRegionsSet("Custom", true);
    if(start==stop){
      GPVVariant toadd = new GPVVariant(chrom,start,strand);
      toadd.setVariantType(GPVVariantConstants.Other);
      set.addRegion(toadd);
    }else{
      set.addRegion(new GPVHighlightedRegion(chrom,start,stop,strand));
    }
    
    refreshRegionsPerChromosomeCache();
    fireDataChangeListeners(WHATCHANGED.REGIONS,CHANGEACTION.ADDED);
  }
  
  public void removeHighlightedRegion(GPVHighlightedRegion toremove){
    for(GPVHighlightedRegionSet set:this.highlightedRegionSets){
      if(set.getRegions()!=null){
        set.getRegions().remove(toremove);
        TreeSet<GPVHighlightedRegion> regionlist = regionsPerChromosomeCache.get(toremove.getChromosome().getTrimmedName());
        if(regionlist!=null){
          regionlist.remove(toremove);
        }
        fireDataChangeListeners(WHATCHANGED.REGIONS,CHANGEACTION.REMOVED);
      }
    }
  }
  
  public void removeHighlightedRegionThatContains(GPVHighlightedRegion child){
    for(GPVHighlightedRegionSet set:this.highlightedRegionSets){
      if(set.getRegions()!=null){

        GPVHighlightedRegion toremove = null;

        TreeSet<GPVHighlightedRegion> regionlist = regionsPerChromosomeCache.get(child.getChromosome().getTrimmedName());
        if(regionlist!=null){
          for(GPVHighlightedRegion region:regionlist){
            if(region.getParentSet() == set && region.contains(child) ){
              toremove = region;
              break;
            }
          }
        }
        if(toremove!=null){
          regionlist.remove(toremove);
          set.getRegions().remove(toremove);
          fireDataChangeListeners(WHATCHANGED.REGIONS,CHANGEACTION.REMOVED);
        }
      }
    }
  }
  


  public ArrayList<GPVDomainSet> getDomainSets()
  {
    return domainSets;
  }

  public void setDomainSets(ArrayList<GPVDomainSet> domainSets)
  {
    this.domainSets = domainSets;
    fireDataChangeListeners(WHATCHANGED.DOMAINS,CHANGEACTION.LOADED);
  }

  public HashSet<String> getAdditionalDomainSetInfoKeys()
  {
    return additionalDomainSetInfoKeys;
  }

  public void setAdditionalDomainSetInfoKeys(HashSet<String> additionalDomainSetInfoKeys)
  {
    this.additionalDomainSetInfoKeys = additionalDomainSetInfoKeys;
//    fireDataChangeListeners(WHATCHANGED.DOMAINS,CHANGEACTION.LOADED);
  }

  /**
   * Important! This always needs to be called when the regions or gene list change!
   */
  public void refreshRegionsPerChromosomeCache(){
    //sort the regions by chrom
    if(this.highlightedRegionSets!=null){
      HashMap<String,TreeSet<GPVHighlightedRegion>> organized_regions = new HashMap<String,TreeSet<GPVHighlightedRegion>>();
      TreeSet<GPVHighlightedRegion> chromregions = null;
      String curchrom = null;
      TreeSet<GPVHighlightedRegion> allregions = new TreeSet<GPVHighlightedRegion>();
      for(GPVHighlightedRegionSet set:this.highlightedRegionSets){
        allregions.addAll(set.getRegions());
      }
      for(GPVHighlightedRegion region:allregions){
        if(curchrom == null || !curchrom.equals(region.getChromosome().getTrimmedName())){
          curchrom = region.getChromosome().getTrimmedName();
          chromregions = new TreeSet<GPVHighlightedRegion>();
          organized_regions.put(curchrom, chromregions);
        }
        chromregions.add(region);
      }

      this.regionsPerChromosomeCache = organized_regions;
    }
  }
  
  private void refreshUnknownVariantTypeList(){
    unknownVariantTypes = new HashSet<String>();
    for(GPVHighlightedRegionSet set:highlightedRegionSets){
      refreshUnknownVariantsFromList(set.getRegions());
    }
  }
  
  private void refreshUnknownVariantsFromList(Collection<GPVHighlightedRegion> list){
    for(GPVHighlightedRegion region:list){
      if(region instanceof GPVVariant){
        GPVVariant v  = (GPVVariant) region;
        if(v.getVariantType()==null){
          unknownVariantTypes.add(v.getOriginalVariantType());
        }
      }
    }
  }

  
  public TreeSet<GPVHighlightedRegion> getHighlightedRegionsForGene(GPVGene selectedGene, boolean enabled_only){
    return getHighlightedRegionsForGene(selectedGene,true,true,enabled_only);
  }
  
  public TreeSet<GPVHighlightedRegion> getHighlightedRegionsForGene(GPVGene selectedGene, boolean include_variants, boolean include_regular_regions, boolean enabled_only){
    if(regionsPerChromosomeCache!=null && selectedGene !=null){
      TreeSet<GPVHighlightedRegion> regions_in_chrom = regionsPerChromosomeCache.get(selectedGene.getChromosome().getTrimmedName()); 
      if(regions_in_chrom!=null){
        ArrayList<GPVHighlightedRegion> ret_unsplit = new ArrayList<GPVHighlightedRegion>();
        for(GPVHighlightedRegion region:regions_in_chrom){
          if(region.getParentSet().isEnabled() || !enabled_only) {
            if((include_variants && region instanceof GPVVariant)|| (include_regular_regions && !(region instanceof GPVVariant))){
              if(region.isInGene(selectedGene)){
                ret_unsplit.add(region);
              }
            }
          }
        }
        return new TreeSet<GPVHighlightedRegion>(ret_unsplit);
      }
    }
    return new TreeSet<GPVHighlightedRegion>();
  }
  
  public boolean isGeneHasHighlightedRegions(GPVGene selectedGene, boolean include_variants, boolean include_regular_regions, boolean enabled_only){
    if(regionsPerChromosomeCache!=null && selectedGene !=null){
      TreeSet<GPVHighlightedRegion> regions_in_chrom = regionsPerChromosomeCache.get(selectedGene.getChromosome().getTrimmedName()); 
      if(regions_in_chrom!=null){
        for(GPVHighlightedRegion region:regions_in_chrom){
          if(region.getParentSet().isEnabled() || !enabled_only) {
            if((include_variants && region instanceof GPVVariant)|| (include_regular_regions && !(region instanceof GPVVariant))){
              if(region.isInGene(selectedGene)){
                return true;
              }
            }
          }
        }
      }
    }
    return false;
  }
  
  
  public HashMap<String, ArrayList<GPVDomain>> getProteinDomainsForGene(GPVGene selectedGene){
    return getProteinDomainsForGene(selectedGene,false);
  }
  /**
   * key: protein id
   * val: list of domains
   */
  public HashMap<String, ArrayList<GPVDomain>> getProteinDomainsForGene(GPVGene selectedGene, boolean include_disabled){
    HashMap<String, ArrayList<GPVDomain>> ret = new  HashMap<String, ArrayList<GPVDomain>>();
    Collection<GPVDomainSet> all_sets = (include_disabled?domainSets:getEnabledDomainSets());
    for(GPVTranscript transcript:selectedGene.getTranscripts().values()){
      if(transcript.isCoding()){
        String protid = transcript.getResultingProtein().getIDFrom(domainSource);
        if(protid!=null && !protid.equals("")){
          ArrayList<GPVDomain> domains_for_this_p = new ArrayList<GPVDomain>();
          ret.put(protid, domains_for_this_p);
          
          for(GPVDomainSet domainset: all_sets){
            List<GPVDomain> domains = domainset.getDomainsForProteinID(protid);
            if(domains!=null) domains_for_this_p.addAll(domains);
          }
        }
      }
    }
    return ret;
  }
  
  
  /**
   * val: list of domain sets
   */
  public HashSet<String> getDomainSetsInGene(GPVGene selectedGene, boolean include_disabled){
    HashSet<String> ret = new  HashSet<String>();
    
    Collection<GPVDomainSet> all_sets = (include_disabled?domainSets:getEnabledDomainSets());
    for(GPVDomainSet domainset:all_sets){
      for(GPVTranscript transcript:selectedGene.getTranscripts().values()){
        if(transcript.isCoding()){
          String protid = transcript.getResultingProtein().getIDFrom(domainSource);
          if(protid!=null && !protid.equals("")){
            if(domainset.getDomainsForProteinID(protid)!=null){
              ret.add(domainset.getSetName());
              break; // found something for this domainset, can break
            }
          }
        }
      }  
    }
    return ret;
  }
  
  public ArrayList<GPVDomainSet> getEnabledDomainSets(){
    ArrayList<GPVDomainSet> ret = new ArrayList<GPVDomainSet>();
    for(GPVDomainSet set:domainSets){
      if(set.isEnabled())ret.add(set);
    }
    return ret;
  }
  
  private void fireDataChangeListeners(DataChangedEvent.WHATCHANGED what, DataChangedEvent.CHANGEACTION action){
    for(DataChangedListener listener:changeListeners){
      listener.dataChanged(new DataChangedEvent(what,action));
    }
  }
  
  public void addDataChangeListener(DataChangedListener toadd){
    this.changeListeners.add(toadd);
  }

  public DOMAIN_SOURCE getDomainSource()
  {
    return domainSource;
  }

  public void setDomainSource(DOMAIN_SOURCE domainSource)
  {
    this.domainSource = domainSource;
  }

}
