/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data;

import java.util.Arrays;

import at.tugraz.genome.biojava.seq.DNAStrand;

public class GPVExon implements Comparable<GPVExon>
{
  private int startPos;
  private int endPos;
  private boolean highlighted = false;
  
  //when drawing the union genome map, exons from different transcripts can overlap
  //then, on the union genome map, those exons are merged. but we still want to show were
  //they have been merged, so that they can be splitted into "artificial" sub-exons.
  private int[] mergeSites;
  
  private DNAStrand strand;
  
  public GPVExon(int start, int stop, DNAStrand strand){
    this(start,stop,strand,new int[0]);
  }
  
  public GPVExon(int start, int stop, DNAStrand strand, int[] mergeSites){
    setStartPos(start);
    setEndPos(stop);
    setStrand(strand);
    setMergeSites(mergeSites);
  }
  
  public int getStartPos()
  {
    return startPos;
  }

  public void setStartPos(int startPos)
  {
    this.startPos = startPos;
  }

  public int getEndPos()
  {
    return endPos;
  }

  public void setEndPos(int endPos)
  {
    this.endPos = endPos;
  }

  public DNAStrand getStrand()
  {
    return strand;
  }

  public void setStrand(DNAStrand strand)
  {
    this.strand = strand;
  }
  
  public boolean equals(int start, int stop, DNAStrand strand){
    if(this.startPos==start&&this.endPos==stop&&this.strand==strand) return true;
    return false;
  }
  
  public boolean equals(GPVExon o){
    return compareTo(o)==0;
  }

  @Override
  public int compareTo(GPVExon o)
  {
    if(o==null) return Integer.MIN_VALUE;
    int ret = startPos-o.startPos;
    if(ret!=0) return ret;
    ret = endPos-o.endPos;
    if(ret!=0) return ret;
    if(this.strand!=o.strand){
      if(strand == DNAStrand.FORWARD)return 1;
      if(strand == DNAStrand.REVERSE && o.strand == DNAStrand.UNKNOWN)return 1;
      return -1;
    }
    return 0;
  }

  public boolean isHighlighted()
  {
    return highlighted;
  }

  public void setHighlighted(boolean highlighted)
  {
    this.highlighted = highlighted;
  }
  
  public int[] getMergeSites()
  {
    return mergeSites;
  }

  public void setMergeSites(int[] mergeSites)
  {
    Arrays.sort(mergeSites);
    this.mergeSites = mergeSites;
  }

  private int[] addToArrayIfNotDuplicate(int[] list, int toadd){
    int[] ret = new int[list.length+1];
    for(int i = 0; i < list.length; i++){
      if(list[i] == toadd) return list;
      
      ret[i] = list[i];
    }
    ret[list.length] = toadd; 
    return ret;
  }
  
  public void absorbExon(GPVExon newexon){
    int newMergeSiteA = Math.max(newexon.getStartPos(), this.getStartPos());
    int newMergeSiteB = Math.min(newexon.getEndPos(), this.getEndPos());
    int[] newMergeSites = getMergeSites();
    
    if(newexon.getStartPos() != this.getStartPos()) newMergeSites= addToArrayIfNotDuplicate(newMergeSites, newMergeSiteA);
    if(newexon.getEndPos() != this.getEndPos()) newMergeSites= addToArrayIfNotDuplicate(newMergeSites, newMergeSiteB);

    for(int i = 0; i < newexon.getMergeSites().length; i++){
      newMergeSites= addToArrayIfNotDuplicate(newMergeSites, newexon.getMergeSites()[i]);
    }
    setStartPos(Math.min(newexon.getStartPos(),this.getStartPos()));
    setEndPos(Math.max(newexon.getEndPos(),this.getEndPos()));
    setMergeSites(newMergeSites);
  }
  
  public GPVExon clone(){
    GPVExon ret = new GPVExon(getStartPos(), getEndPos(), getStrand(), getMergeSites().clone());
    return ret;
  }
}
