/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data;

import java.util.Collection;
import java.util.HashMap;
import java.util.TreeSet;

import at.icbi.gpviz.GPViz;
import at.icbi.gpviz.data.util.GPVUtils;
import at.tugraz.genome.biojava.seq.DNAStrand;

public class GPVGene implements Comparable<GPVGene>
{
  private String geneSymbol;
  
  private String geneFullName;
  
  private String geneID;
  
  private String uniGeneID;
  
  private DNAStrand strand;
  
  private GPVChromosome chromosome;
  
  private HashMap<String, GPVTranscript> transcripts = new HashMap<String,GPVTranscript>(1);
  
  private int first_known_position;
  private int last_known_position;
  
  public enum DISPLAYNAME{
    SYMBOL, ID, LONGNAME
  }
  
  /* A simple caching mechanism to speed up drawing images with lots of transcripts **/
  private TreeSet<GPVExon> allExonsForAllTranscriptsCache = null;
  private TreeSet<GPVExon> allExonsForEnabledTranscriptsCache = null;
  
  public GPVGene(String name, GPVChromosome chromosome){
    setGeneID(name);
    setChromosome(chromosome);
  }
  
  public GPVExon getExonByProperties(int start, int stop, DNAStrand strand){
    for(GPVExon totest:getExons()){
      if(totest.equals(start,stop,strand)) return totest;
    }
    return null;
  }
  
  public TreeSet<GPVExon> getExons(boolean include_out_of_coding_range, boolean only_enabled_transcripts)
  {
    TreeSet<GPVExon> ret = new TreeSet<GPVExon>();
    Collection<GPVTranscript> transcripts;
    if(only_enabled_transcripts){
      transcripts = getEnabledTranscriptsSorted();
    }else{
      transcripts = getTranscripts().values();
    }
    
    for(GPVTranscript transcript:transcripts){
      Collection<GPVExon> subexons;
      if(include_out_of_coding_range){
        subexons = transcript.getExons();
      }else{
        subexons = transcript.getAllExonRegionsSeparately(include_out_of_coding_range);
      }
      for(GPVExon exon:subexons){
        mergeExons(exon.clone(), ret);
      }
    }
    
    splitExonsAtStartStopCodons(ret);
    
    return ret;
  }
  
  public TreeSet<GPVExon> getExons()
  {
    return getExons(true,false);
  }
  
  public int getFirstKnownPosition()
  {
//    int ret = Integer.MAX_VALUE;
//    for(GPVTranscript transcript:getTranscripts().values()){
//      ret = Math.min(transcript.getExons().first().getStartPos(), ret);
//    }
//    return ret;
    return first_known_position;
  }

  public int getLastKnownPosition()
  {
//    int ret = Integer.MIN_VALUE;
//    for(GPVTranscript transcript:getTranscripts().values()){
//      ret = Math.max(transcript.getExons().last().getEndPos(), ret);
//    }
//    return ret;
    return last_known_position;
  }

  public void updateFirstAndLastPosition(){
    first_known_position = Integer.MAX_VALUE;
    last_known_position = Integer.MIN_VALUE;
    for(GPVTranscript transcript:getTranscripts().values()){
      first_known_position = Math.min(transcript.getExons().first().getStartPos(), first_known_position);
      last_known_position = Math.max(transcript.getExons().last().getEndPos(), last_known_position);
    }
  }
  
  public GPVChromosome getChromosome()
  {
    return chromosome;
  }

  public void setChromosome(GPVChromosome chromosome)
  {
    this.chromosome = chromosome;
  }

  public String getGeneSymbol()
  {
    return geneSymbol;
  }

  public void setGeneSymbol(String geneName)
  {
    this.geneSymbol = geneName;
  }
  
  public String getGeneFullName()
  {
    return geneFullName;
  }

  public void setGeneFullName(String geneFullName)
  {
    this.geneFullName = geneFullName;
  }

  public HashMap<String,GPVTranscript> getTranscripts()
  {
    return transcripts;
  }
  
  public TreeSet<GPVTranscript> getSortedTranscripts(){
    return new TreeSet<GPVTranscript>(transcripts.values());
  }
  
  public String getGeneID()
  {
    return geneID;
  }

  private void setGeneID(String geneID)
  {
    this.geneID = geneID;
  }
  
  public String getUniGeneID()
  {
    return uniGeneID;
  }

  public void setUniGeneID(String uniGeneID)
  {
    this.uniGeneID = uniGeneID;
  }

  public TreeSet<GPVTranscript> getEnabledTranscriptsSorted(){
    TreeSet<GPVTranscript> ret = new TreeSet<GPVTranscript>();
    for(GPVTranscript transcript:getTranscripts().values()){
      if(transcript.isDisplayed()) ret.add(transcript);
    }
    return ret;
  }
  
  public GPVTranscript getTranscriptFromEnabledTranscripts(int tr){
    int i = 1;
    for(GPVTranscript transcript:getEnabledTranscriptsSorted()){
      if(i++==tr) return transcript;
    }
    return null;
  }

  public void setTranscripts(HashMap<String,GPVTranscript> transcripts)
  {
    this.transcripts = transcripts;
    invalidateAllTranscriptsExonCache();
  }
  
  public DNAStrand getStrand()
  {
    return strand;
  }

  public void setStrand(DNAStrand strand)
  {
    this.strand = strand;
  }
  
  public TreeSet<GPVExon> getAllExonRegionsSeparately(){
    return getAllExonRegionsSeparately(true,false);
  }
  
  public TreeSet<GPVExon> getAllExonRegionsSeparately(boolean include_out_of_transcription_range, boolean only_enabled_transcripts){
    if(include_out_of_transcription_range){
      if(allExonsForAllTranscriptsCache == null){
        allExonsForAllTranscriptsCache = new GPVUtils().getAllExonRegionsSeparately(getExons(include_out_of_transcription_range,only_enabled_transcripts)); 
      }
      return allExonsForAllTranscriptsCache;
    }else{
      if(allExonsForEnabledTranscriptsCache == null){
        allExonsForEnabledTranscriptsCache = new GPVUtils().getAllExonRegionsSeparately(getExons(include_out_of_transcription_range,only_enabled_transcripts),getEarliestLowerTranscriptionRegionBoundary(), getLatestUpperTranscriptionRegionBoundary()); 
      }
      return allExonsForEnabledTranscriptsCache;
    }
  }

  @Override
  public int compareTo(GPVGene o)
  {
    int ret = getChromosome().compareTo(o.getChromosome());
    if(ret!=0)return ret;
    ret = getFirstKnownPosition()-o.getFirstKnownPosition();
    if(ret!=0)return ret;
    ret = getLastKnownPosition()-o.getLastKnownPosition();
    return ret;
  }
  
  
  public String toString(){
    return getDisplayName();
  }
  
  public int getEarliestStartCodon(){
    boolean reverse = strand == DNAStrand.REVERSE;
    int ret = reverse?Integer.MIN_VALUE:Integer.MAX_VALUE;
    for(GPVTranscript transcript: getEnabledTranscriptsSorted()){
      if(transcript.isCoding()){
        if(reverse){
          ret = Math.max(transcript.getStartCodon(),ret);
        }else{
          ret = Math.min(transcript.getStartCodon(),ret);
        }
      }
    }
    return ret;
  }

  public int getLatestStopCodon(){
    boolean reverse = strand == DNAStrand.REVERSE;
    int ret = reverse?Integer.MAX_VALUE:Integer.MIN_VALUE;
    for(GPVTranscript transcript: getEnabledTranscriptsSorted()){
      if(transcript.isCoding()){
        if(reverse){
          ret = Math.min(transcript.getStopCodon(),ret);
        }else{
          ret = Math.max(transcript.getStopCodon(),ret);
        }
      }
    }
    return ret;
  }
  
  public int getEarliestLowerTranscriptionRegionBoundary(){
    int ret = Integer.MAX_VALUE;
    for(GPVTranscript transcript: getEnabledTranscriptsSorted()){
      if(transcript.isCoding()){
        ret = Math.min(transcript.getLowerTranscriptionRegionBoundary(),ret);
      }
    }
    return ret;
  }
  
  public int getLatestUpperTranscriptionRegionBoundary(){
    int ret = Integer.MIN_VALUE;
    for(GPVTranscript transcript: getEnabledTranscriptsSorted()){
      if(transcript.isCoding()){
        ret = Math.max(transcript.getUpperTranscriptionRegionBoundary(),ret);
      }
    }
    return ret;
  }
  
  private void mergeExons(GPVExon new_exon, TreeSet<GPVExon> other_exons){
    GPVExon mother_exon = new_exon.clone();
    GPVExon[] inrange = new GPVUtils().getExonsInRange(new_exon.getStartPos(), new_exon.getEndPos(), new_exon.getStrand(),other_exons);
    if(inrange.length!=0){
      for(int i = 0; i < inrange.length; i++){
        mother_exon.absorbExon(inrange[i]);
        other_exons.remove(inrange[i]);
      }
    }
    other_exons.add(mother_exon);
  }
  
  public String getDisplayName(){
    String ret = null;
    switch(GPViz.getInstance().getOptions().getGeneNameDisplayed()){
      case ID:
        ret = geneID;
        break;
      case LONGNAME:
        ret = geneFullName;
        if(ret==null)ret=geneSymbol;
        break;
      default: case SYMBOL:
        if(geneSymbol!=null && !geneSymbol.equals("")){ 
          ret = geneSymbol;
        }else{
        ret = geneID;
        }
    } 
    if(ret==null)ret = geneID;
    return ret;
  }
  
  /**
   * Doesn't really split the exon, but markes "merge sites" in the exon
   * In the union transcript it is then split at those merge sites.
   * this makes handling of transcriptions/proteins that start mid-exon much easier.
   */
  private void splitExonsAtStartStopCodons(TreeSet<GPVExon> exons){
    for(GPVTranscript transcript:getEnabledTranscriptsSorted()){
      GPVExon[] inrange = new GPVUtils().getExonsInRange(transcript.getStartCodon(), transcript.getStartCodon(), getStrand(),exons);
      if(transcript.getStartCodon()!=0)splitExonAtPosition(transcript.getStartCodon(),inrange);
      inrange = new GPVUtils().getExonsInRange(transcript.getStopCodon(), transcript.getStopCodon(), getStrand(),exons);
      if(transcript.getStopCodon()!=0) splitExonAtPosition(transcript.getStopCodon(),inrange);
      
      for(GPVTranscript transcript_b:getEnabledTranscriptsSorted()){
        inrange = transcript_b.getExonsInRange(transcript.getStartCodon(), transcript.getStartCodon(), getStrand());
        if(transcript.getStartCodon()!=0)splitExonAtPosition(transcript.getStartCodon(),inrange);
        inrange = transcript_b.getExonsInRange(transcript.getStopCodon(), transcript.getStopCodon(), getStrand());
        if(transcript.getStopCodon()!=0)splitExonAtPosition(transcript.getStopCodon(),inrange);
      }
    }
  }
  
  private void splitExonAtPosition(int splice_location, GPVExon[] exons) {
    if(exons.length > 0){
      GPVExon exon = exons[0];
      
      //a start/stop codon is 3 nucleotides long. We can thus assume that the start/end pos of 
      //a exon is still within the start/stop codon if it is 1 nucleotide away
      if(Math.abs(exon.getStartPos()-splice_location)>1 && Math.abs(exon.getEndPos()-splice_location)>1){
        boolean already_known = false;
        int[] old_mergesites = exon.getMergeSites();
        
        for(int i = 0; i < old_mergesites.length; i++){
          if(Math.abs(old_mergesites[i]-splice_location)<=2){
            //ok, we already know this one. don't add it again
            already_known = true;
          }
        }
        if(!already_known){
          int[] new_mergesites = new int[old_mergesites.length+1];
          for(int i = 0; i < old_mergesites.length; i++){
            new_mergesites[i] = old_mergesites[i];
          }
          new_mergesites[new_mergesites.length-1]=splice_location;
          exon.setMergeSites(new_mergesites);
        }
      }
    }
  }
  
  protected void invalidateEnabledTranscriptsExonCache(){
    this.allExonsForEnabledTranscriptsCache = null;
  }
  
  
  private void invalidateAllTranscriptsExonCache(){
    this.allExonsForAllTranscriptsCache = null;
  }


}
