/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data;

import java.io.File;

import javax.swing.JOptionPane;
import javax.xml.bind.JAXB;
import javax.xml.bind.annotation.XmlTransient;

import at.icbi.gpviz.data.GPVGene.DISPLAYNAME;
import at.icbi.gpviz.drawing.colorscheme.AbstractGPVColorScheme;
import at.icbi.gpviz.drawing.colorscheme.DefaultGPVColorScheme;
import at.icbi.gpviz.drawing.colorscheme.UserDefinedGPVColorScheme;
import at.icbi.gpviz.weblinks.WebLink;

public class GPVOptions implements Cloneable
{
  private String NMtoNPFile = null;
  private String ensemblToUniprotFile = null;
  
  private String lastInputDir = null;
  private String lastImageSaveDir = null;
  
  private AbstractGPVColorScheme preferredColorScheme = null;
  
  //this one needs to be separate, because its easier to marshall/unmarshall it that way
  private String preferredColorSchemeName = null;
  
  private String geneInfoLinkBase = "http://www.ncbi.nlm.nih.gov/gene/"+WebLink.decorateVariable(WebLink.VARIABLE_UNIGENEID);
  private String transcriptInfoLinkBase = "http://www.ncbi.nlm.nih.gov/nuccore/"+WebLink.decorateVariable(WebLink.VARIABLE_TRANSCRIPT_REFSEQ);
  private String proteinInfoLinkBase = "http://www.ncbi.nlm.nih.gov/protein/"+WebLink.decorateVariable(WebLink.VARIABLE_PROTEIN_REFSEQ);
  
  private String regionLinkBase = "http://genome.ucsc.edu/cgi-bin/hgTracks?hgHubConnect.destUrl=../cgi-bin/hgTracks&clade=mammal&org=Human&db=hg19&position=chr"+
  WebLink.decorateVariable(WebLink.VARIABLE_CHROMOSOME)+":"+WebLink.decorateVariable(WebLink.VARIABLE_STARTPOS)+"-"+WebLink.decorateVariable(WebLink.VARIABLE_STOPPOS);
  
  private DISPLAYNAME geneNameDisplayed = DISPLAYNAME.SYMBOL;
  
  private boolean useProteinDomainCache = true;
  
  public String getNMtoNPFile()
  {
    if(NMtoNPFile==null || NMtoNPFile.trim().equals("")) return null;
    return NMtoNPFile;
  }
  
  public String getEnsemblToUniprotFile()
  {
    if(ensemblToUniprotFile==null || ensemblToUniprotFile.trim().equals("")) return null;
    return ensemblToUniprotFile;
  }

  public void setEnsemblToUniprotFile(String ensemblToUniprotFile)
  {
    this.ensemblToUniprotFile = ensemblToUniprotFile;
  }
  
  public void setNMtoNPFile(String nMtoNPFile)
  {
    this.NMtoNPFile = nMtoNPFile;
  }
  
  public String getLastInputDir()
  {
    return lastInputDir;
  }

  public void setLastInputDir(String lastInputDir)
  {
    this.lastInputDir = lastInputDir;
  }

  public String getLastImageSaveDir()
  {
    return lastImageSaveDir;
  }

  public void setLastImageSaveDir(String lastImageSaveDir)
  {
    this.lastImageSaveDir = lastImageSaveDir;
  }

  @XmlTransient
  public AbstractGPVColorScheme getPreferredColorScheme()
  {
    return preferredColorScheme;
  }

  public void setPreferredColorScheme(AbstractGPVColorScheme preferredColorScheme)
  {
    this.preferredColorScheme = preferredColorScheme;
    privateSetPreferredColorSchemeName(this.preferredColorScheme.getName());
  }

  public String getPreferredColorSchemeName()
  {
    return preferredColorSchemeName;
  }

  public String getGeneInfoLinkBase()
  {
    return geneInfoLinkBase;
  }

  public void setGeneInfoLinkBase(String geneInfoLinkBase)
  {
    this.geneInfoLinkBase = geneInfoLinkBase;
  }

  public String getTranscriptInfoLinkBase()
  {
    return transcriptInfoLinkBase;
  }

  public void setTranscriptInfoLinkBase(String transcriptInfoLinkBase)
  {
    this.transcriptInfoLinkBase = transcriptInfoLinkBase;
  }

  public String getProteinInfoLinkBase()
  {
    return proteinInfoLinkBase;
  }
  
  public String getRegionLinkBase()
  {
    return regionLinkBase;
  }

  public void setProteinInfoLinkBase(String proteinInfoLinkBase)
  {
    this.proteinInfoLinkBase = proteinInfoLinkBase;
  }
  

  public void setRegionLinkBase(String regionLinkBase)
  {
    this.regionLinkBase = regionLinkBase;
  }
  
  @XmlTransient
  public DISPLAYNAME getGeneNameDisplayed()
  {
    return geneNameDisplayed;
  }

  public void setGeneNameDisplayed(DISPLAYNAME geneNameDisplayed)
  {
    this.geneNameDisplayed = geneNameDisplayed;
  }

  private void privateSetPreferredColorSchemeName(String preferredColorSchemeName){
    this.preferredColorSchemeName = preferredColorSchemeName;
  }
  
  public void setPreferredColorSchemeName(String preferredColorSchemeName)
  {
    this.preferredColorSchemeName = preferredColorSchemeName;
    
    if(preferredColorSchemeName == null || preferredColorSchemeName.equals(DefaultGPVColorScheme.NAME)){
      this.preferredColorScheme = new DefaultGPVColorScheme();
    }else{
      for(UserDefinedGPVColorScheme scheme:UserDefinedGPVColorScheme.loadAll()){
        if(preferredColorSchemeName.equals(scheme.getName())){
          this.preferredColorScheme = scheme;
        }
      }
    }
  }

  public static final String optionsFilename = System.getProperty("user.home")+"/.gpviz/systemproperties.xml";
  
  public static GPVOptions load(){
    File optionsfile = new File(optionsFilename);
    GPVOptions ret = new GPVOptions();
    if(!optionsfile.exists()){
      JOptionPane.showMessageDialog(null, "No options file has been found. If you want to work with Refseq data, please go Tools->Options to define the annotation file. Consult the manual for more information.");
      return ret;
    }
    
    try{
      ret = JAXB.unmarshal(optionsfile, GPVOptions.class);
    }catch(Exception e){
      e.printStackTrace();
      JOptionPane.showMessageDialog(null, "Error loading options file "+optionsfile.getName());
    }
    return ret;
  }  
  
  public static void save(GPVOptions options){
    File optionsfile = new File(optionsFilename);
    if(!optionsfile.getParentFile().exists()) optionsfile.getParentFile().mkdirs();
    JAXB.marshal(options, optionsfile);
  }  
  
  public boolean isUseProteinDomainCache()
  {
    return useProteinDomainCache;
  }

  public void setUseProteinDomainCache(boolean useProteinDomainCache)
  {
    this.useProteinDomainCache = useProteinDomainCache;
  }

  public GPVOptions clone(){
    GPVOptions clone = new GPVOptions();
    clone.setNMtoNPFile(this.getNMtoNPFile());
    clone.setLastImageSaveDir(this.getLastImageSaveDir());
    clone.setLastInputDir(this.getLastInputDir());
    clone.setPreferredColorScheme(this.getPreferredColorScheme());
    return clone;
  }
  
}
