/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data;

import java.util.Collection;
import java.util.TreeSet;

import at.icbi.gpviz.data.util.GPVUtils;
import at.tugraz.genome.biojava.seq.DNAStrand;

public class GPVTranscript implements Comparable<GPVTranscript>
{
  
  private String transcriptId;
  
  /**
   * the exons that are used for this transcript
   */
  private TreeSet<GPVExon> exons = new TreeSet<GPVExon>();
  
  private GPVProtein resultingProtein;
  
  private int startCodon;
  private int stopCodon;
  
  private GPVGene parent;
  
  private DNAStrand strand;
  
  private boolean displayed = true;
  
  
  /* A simple caching mechanism to speed up drawing images with lots of transcripts **/
  private TreeSet<GPVExon> allExonsSeperatelyCache = null;
  private TreeSet<GPVExon> allCodingExonsSeparatelyCache = null;
  
  
  public GPVTranscript(String id){
    transcriptId = id;
    setResultingProtein(new GPVProtein());
  }

  public GPVTranscript(String id, String proteinId){
    transcriptId = id;
    setResultingProtein(new GPVProtein(proteinId));
  }
  
  public TreeSet<GPVExon> getExons()
  {
    return this.exons;
  }

  public void setExons(TreeSet<GPVExon> exons)
  {
    this.exons = exons;
    
    
    allExonsSeperatelyCache = null;
    allCodingExonsSeparatelyCache = null;
  }
  
  public GPVProtein getResultingProtein()
  {
    return resultingProtein;
  }

  public void setResultingProtein(GPVProtein resultingProtein)
  {
    this.resultingProtein = resultingProtein;
    this.resultingProtein.setParent(this);
  }

  public String getTranscriptId()
  {
    return transcriptId;
  }
  
  public int getStartCodon()
  {
    return startCodon;
  }
  
  public boolean isCoding(){
    return startCodon>0 && stopCodon>0;
  }
  
  public void setStartCodon(int startCodon)
  {
    this.startCodon = startCodon;
  }
  
  public void setStartCodonIfBetter(int startCodon)
  {
    if(this.startCodon==0)this.startCodon = startCodon;
    if(strand==DNAStrand.REVERSE){
      this.startCodon = Math.max(startCodon, this.startCodon);
    }else{
      this.startCodon = Math.min(startCodon, this.startCodon);
    }
  }

  public int getStopCodon()
  {
    return stopCodon;
  }

  public void setStopCodon(int stopCodon)
  {
    this.stopCodon = stopCodon;
  }
  
  /**
   * @return Start codon if forward strand, stop codon if reverse strand
   */
  public int getLowerTranscriptionRegionBoundary(){
    int ret;
    if(strand==DNAStrand.REVERSE){
      ret = this.stopCodon;
    }else{
      ret = this.startCodon;
    }
    return ret==0 ? Integer.MIN_VALUE: ret;
  }

  /**
   * @return Stop codon if forward strand, start codon if reverse strand
   */
  public int getUpperTranscriptionRegionBoundary(){
    int ret;
    if(strand==DNAStrand.REVERSE){
      ret = this.startCodon;
    }else{
      ret = this.stopCodon;
    }
    return ret==0 ? Integer.MAX_VALUE : ret;
  }
  
  public void setStopCodonIfBetter(int stopCodon)
  {
    if(this.stopCodon==0)this.stopCodon = stopCodon;
    if(strand==DNAStrand.REVERSE){
      this.stopCodon = Math.min(stopCodon, this.stopCodon);
    }else{
      this.stopCodon = Math.max(stopCodon, this.stopCodon);
    }
  }

  public GPVGene getParent()
  {
    return parent;
  }

  public void setParent(GPVGene parent)
  {
    this.parent = parent;
  }
  
  public DNAStrand getStrand()
  {
    return strand;
  }

  public void setStrand(DNAStrand strand)
  {
    this.strand = strand;
  }

  
  public boolean containsExon(GPVExon exon){
    for(GPVExon listed_exon: getExons()){
      if(exon.equals(listed_exon))return true;
    }
    return false;
  }
  
  public GPVExon[] getExonsInRange(int start, int stop, DNAStrand strand){
    return new GPVUtils().getExonsInRange(start, stop, strand, getExons());
  }
  
  public TreeSet<GPVExon> getAllExonRegionsSeparately(boolean include_out_of_transcription_range){
    if(include_out_of_transcription_range){
      if(allExonsSeperatelyCache==null) allExonsSeperatelyCache = new GPVUtils().getAllExonRegionsSeparately(getExons()); 
      return allExonsSeperatelyCache;
    }else{
      //if its not coding, nothing is in transcription range
      if(!this.isCoding()) return new TreeSet<GPVExon>();
      if(allCodingExonsSeparatelyCache == null) allCodingExonsSeparatelyCache = new GPVUtils().getAllExonRegionsSeparately(getExons(),getLowerTranscriptionRegionBoundary(), getUpperTranscriptionRegionBoundary()); 
      return allCodingExonsSeparatelyCache;
    }
  }
  
  public TreeSet<GPVExon> getAllExonRegionsSeparately(){
    return getAllExonRegionsSeparately(true);
  }
  
  public GPVExon[] getExonsAlignedToUnion(){
    TreeSet<GPVExon> union = parent.getAllExonRegionsSeparately(false,true);
    GPVExon[] ret = new GPVExon[union.size()];
    
    Collection<GPVExon> tomap = getAllExonRegionsSeparately(false);
    
    int index = 0;
    for(GPVExon cons_exon:union){
      if(isInExonList(tomap,cons_exon)){
        ret[index]=cons_exon;
      }
      index++;
    }
    return ret;
  }
  
  private boolean isInExonList(Collection<GPVExon> list, GPVExon search){
    int start = search.getStartPos();
    int stop = search.getEndPos();
    
    for(GPVExon cur:list){
      if(start>=cur.getStartPos() && stop<=cur.getEndPos()){
        return true;
      }
    }
    return false;
  }
  
  public int getProteinPositionForNucleotidePosition(int nt){
    Collection<GPVExon> exons = getAllExonRegionsSeparately(false);
    return GPVUtils.transformNTToPP(nt, exons, getStrand()==DNAStrand.REVERSE);
  }
  
  
  public boolean isDisplayed()
  {
    return displayed;
  }

  public void setDisplayed(boolean displayed)
  {
    this.displayed = displayed;
    //enabled transcripts have changed, we got to let the parent know to update the cache
    parent.invalidateEnabledTranscriptsExonCache();
  }

  @Override
  public int compareTo(GPVTranscript o)
  {
    //compare them by their start and stop codon, so that the visualization is less chaotic
    int ret = getStartCodon()-o.getStartCodon();
    if(ret!=0)return ret;
    ret = getStopCodon()-o.getStopCodon();
    if(ret!=0)return ret;
    
    ret = transcriptId.compareTo(o.getTranscriptId());
    return ret;
  }
  
}