/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data.loader;

import java.io.File;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

import at.icbi.gpviz.GPViz;
import at.icbi.gpviz.data.GPVChromosome;
import at.icbi.gpviz.data.GPVDataContainer;
import at.icbi.gpviz.data.GPVOptions;
import at.icbi.gpviz.data.loader.parser.BedEntry;
import at.icbi.gpviz.data.loader.parser.BedParser;
import at.icbi.gpviz.data.regions.GPVHighlightedRegion;
import at.icbi.gpviz.data.regions.GPVHighlightedRegionSet;
import at.icbi.gpviz.swing.ExtensionBasedFileFilter;
import at.icbi.gpviz.swing.MainWindow;
import at.tugraz.genome.biojava.io.EntryReader;
import at.tugraz.genome.biojava.io.RawEntry;
import at.tugraz.genome.biojava.seq.DNAStrand;
import at.tugraz.genome.util.FileUtils;

public class BedFileRegionLoader
{
  /**
   * loads regions from a bed file and saves them to the data container for later visualization
   */
  public void loadRegions(final MainWindow parent){


    GPVDataContainer data = GPViz.getInstance().getDataContainer();

    String lastinputdir = GPViz.getInstance().getOptions().getLastInputDir();
    if(lastinputdir==null)lastinputdir = System.getProperty("user.home");

    JFileChooser chooser = new JFileChooser(lastinputdir);
    chooser.setMultiSelectionEnabled(true);
    chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
    chooser.addChoosableFileFilter(new ExtensionBasedFileFilter("bed"));
    int selection = chooser.showDialog(parent, "Open");


    final File[] selectedfiles = chooser.getSelectedFiles();
    if(selectedfiles!=null && selectedfiles.length>0 && selection == JFileChooser.APPROVE_OPTION){
      GPViz.getInstance().getOptions().setLastInputDir(selectedfiles[0].getParent());
      int total_new_regions = 0;
      for(File selectedfile:selectedfiles){
        int newregions = 0;
        GPVOptions.save(GPViz.getInstance().getOptions());
        EntryReader<BedEntry> reader = null;
        try{
          BedParser parser = new BedParser();
          parser.setParseTrackLine(false);
          reader = new EntryReader<BedEntry>(selectedfile,parser);        

          long file_size = selectedfile.length();
          long length_read = 0;



          GPVHighlightedRegionSet set = new GPVHighlightedRegionSet();
          set.setSetName(FileUtils.removeFileExtension(selectedfile.getName()));

          RawEntry raw_entry;
          while((raw_entry = reader.next())!=null){
            length_read += raw_entry.getContent().length();


            if(!parser.isOtherHeader(raw_entry.getContent().trim()) && !parser.isTrackLine(raw_entry.getContent().trim())){
              BedEntry bed  = parser.parseEntry(raw_entry);
              DNAStrand strand =DNAStrand.UNKNOWN;
              String chrom = bed.getChrom();
              int start = bed.getStart().intValue();
              int stop = bed.getEnd().intValue();
              if(bed.getStrand() == DNAStrand.FORWARD){
                strand = DNAStrand.FORWARD;
              }else if(bed.getStrand() == DNAStrand.REVERSE){
                strand = DNAStrand.REVERSE;
              }else{
                strand = DNAStrand.UNKNOWN;
              }

              GPVHighlightedRegion region = new GPVHighlightedRegion();
              GPVChromosome chromosome = new GPVChromosome(chrom);
              region.setChromosome(chromosome);
              region.setStart(start);
              region.setStop(stop);
              region.setStrand(strand);

              set.addRegion(region);
              newregions++;
            }
            float percentRead = (float)length_read/(float)file_size*100f;
            parent.setProgress((int)percentRead);

            if(Thread.currentThread().isInterrupted()){
              return;
            }
          }

          set.updateChildren();
          data.addHighlightedRegions(set);
          total_new_regions+=newregions;
        }catch (Exception e){
          String error = "Error loading BED Input file. ";
          if(reader!=null)error+="(line "+reader.getCurrentLineNumber()+") ";
          JOptionPane.showMessageDialog(parent, error +e.getMessage());
          e.printStackTrace();
        }finally{
          parent.setProgress((int)0);
        }
      }
      if(total_new_regions>0){
        JOptionPane.showMessageDialog(parent, "Successfully loaded "+total_new_regions+" regions from BED file.");
      }
    }
  }
}