/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data.loader;

import java.io.File;
import java.util.HashMap;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

import at.icbi.gpviz.GPViz;
import at.icbi.gpviz.data.GPVChromosome;
import at.icbi.gpviz.data.GPVDataContainer;
import at.icbi.gpviz.data.GPVExon;
import at.icbi.gpviz.data.GPVGene;
import at.icbi.gpviz.data.GPVOptions;
import at.icbi.gpviz.data.GPVProtein;
import at.icbi.gpviz.data.GPVTranscript;
import at.icbi.gpviz.data.util.GPVUtils;
import at.icbi.gpviz.swing.MainWindow;
import at.tugraz.genome.biojava.ann.gtf.GTFEntry;
import at.tugraz.genome.biojava.ann.gtf.GTFParser;
import at.tugraz.genome.biojava.io.EntryReader;
import at.tugraz.genome.biojava.io.RawEntry;
import at.tugraz.genome.biojava.seq.DNAStrand;

public class GTFExonLoader
{

  private static final String FEATURE_EXON ="exon";
  private static final String FEATURE_START_CODON ="start_codon";
  private static final String FEATURE_STOP_CODON ="stop_codon";
  
  private static final String ATTRIBUTE_PROTEIN_ID ="protein_id";
  private static final String ATTRIBUTE_GENE_NAME ="gene_name";

  public void loadGenome(final MainWindow parent){
    String lastinputdir = GPViz.getInstance().getOptions().getLastInputDir();
    if(lastinputdir==null)lastinputdir = System.getProperty("user.home");
    
    JFileChooser chooser = new JFileChooser(new File(lastinputdir));
    chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
    int selection = chooser.showDialog(parent, "Open");
    
    
    final File selectedfile = chooser.getSelectedFile();
    if(selectedfile!=null && selection == JFileChooser.APPROVE_OPTION){
      GPViz.getInstance().getOptions().setLastInputDir(selectedfile.getParent());
      GPVOptions.save(GPViz.getInstance().getOptions());
      EntryReader<GTFEntry> reader = null;
      try{

        long file_size = selectedfile.length();

        GPVDataContainer data = GPViz.getInstance().getDataContainer();
        HashMap<String,GPVGene> genes = new HashMap<String,GPVGene>();
        
        GTFParser parser = new GTFParser();
        reader = new EntryReader<GTFEntry>(selectedfile,parser);
        long length_read = reader.getHeaderLength();

        RawEntry raw_entry;
        while((raw_entry = reader.next())!=null){
          
          
          GTFEntry entry = parser.parseEntry(raw_entry);
          length_read += raw_entry.getSize();
          
          GPVGene current_gene = genes.get(entry.getGeneId());
          if(current_gene == null){
            GPVChromosome chr = data.getChromosomeByName(GPVUtils.getChromsomeName(entry.getSeqname()));
            current_gene = new GPVGene(entry.getGeneId(), chr);
            if(entry.getGroupAttribute(ATTRIBUTE_GENE_NAME) !=null){
              current_gene.setGeneSymbol(entry.getGroupAttribute(ATTRIBUTE_GENE_NAME));
            }
            genes.put(current_gene.getGeneID(),current_gene);
          }

          GPVTranscript current_transcript = current_gene.getTranscripts().get(entry.getTranscriptId());
          if(current_transcript == null){
            current_transcript = new GPVTranscript(entry.getTranscriptId());
            current_transcript.setParent(current_gene);
            current_gene.getTranscripts().put(current_transcript.getTranscriptId(),current_transcript);
          }
          if(entry.getGroupAttribute(ATTRIBUTE_PROTEIN_ID)!=null){
            if(current_transcript.getResultingProtein()==null) current_transcript.setResultingProtein(new GPVProtein());
            current_transcript.getResultingProtein().setProteinID(entry.getGroupAttribute(ATTRIBUTE_PROTEIN_ID));
          }

          if(entry.getFeature().toLowerCase().equals(FEATURE_EXON)){
            DNAStrand strand = entry.getStrand();
            current_gene.setStrand(strand);
            current_transcript.setStrand(strand);
            
            GPVExon current_exon = new GPVExon((int)entry.getStart(),(int)entry.getEnd(),strand);

            current_transcript.getExons().add(current_exon);
          }else if (entry.getFeature().toLowerCase().equals(FEATURE_START_CODON)){
            //If there are multiple start/stop codons, the first one will win. Later start codons might just be for other transcripts
            if(current_transcript.getStrand()==DNAStrand.REVERSE){
              current_transcript.setStartCodonIfBetter((int)entry.getStart());  
            }else{
              current_transcript.setStartCodonIfBetter((int)entry.getEnd());
            }
            
          }else if (entry.getFeature().toLowerCase().equals(FEATURE_STOP_CODON)){
            if(current_transcript.getStrand()==DNAStrand.REVERSE){
              current_transcript.setStopCodonIfBetter((int)entry.getEnd());
            }else{
              current_transcript.setStopCodonIfBetter((int)entry.getStart());
            }
          }
          float percentRead = (float)length_read/(float)file_size*100f;
          parent.setProgress((int)percentRead);
          if(Thread.currentThread().isInterrupted()){
            return;
          }
        }
        
        for(GPVGene gene : genes.values())gene.updateFirstAndLastPosition();
        data.setGenes(genes);
        
      }
      catch (Exception e)
      {
        String error = "Error loading GTF Input file. ";
        if(reader!=null)error+="(line "+reader.getCurrentLineNumber()+") ";
        JOptionPane.showMessageDialog(parent, error +e.getMessage());
        e.printStackTrace();
      }finally{
        parent.setProgress(0);
      }
    }
  }

}
