/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data.loader.domains;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;

import javax.swing.JOptionPane;

import at.icbi.gpviz.GPViz;
import at.icbi.gpviz.data.GPVDataContainer;
import at.icbi.gpviz.data.GPVDomain;
import at.icbi.gpviz.data.GPVDomainSet;
import at.icbi.gpviz.data.GPVOptions;
import at.icbi.gpviz.data.loader.domains.parser.ProteinDomainEntry;
import at.icbi.gpviz.data.util.GPVUtils;
import at.icbi.gpviz.swing.CheckFreeMemoryThread;
import at.icbi.gpviz.swing.MainWindow;
import at.tugraz.genome.biojava.io.EntryReader;
import at.tugraz.genome.biojava.io.RawEntry;
import at.tugraz.genome.biojava.io.dsv.AbstractDSVParser;

public abstract class AbstractProteinDomainLoader <T extends ProteinDomainEntry>
{

  public void loadDomains(final MainWindow parent, final File selectedfile){
    GPVDataContainer data = GPViz.getInstance().getDataContainer();

    if(selectedfile!=null){
      GPViz.getInstance().getOptions().setLastInputDir(selectedfile.getParent());
      GPVOptions.save(GPViz.getInstance().getOptions());

      long file_size = selectedfile.length();

      AbstractDSVParser<T> parser = getParser();
      EntryReader<T> reader = null;
      try
      {
        reader = new EntryReader<T>(selectedfile, parser);
        long length_read = reader.getHeaderLength();
        RawEntry raw_entry;

        HashMap<String,GPVDomainSet> domain_sets = new HashMap<String,GPVDomainSet>();


        while((raw_entry=reader.next())!=null){

          T entry = parser.parseEntry(raw_entry);
          length_read += raw_entry.getContent().length();

          GPVDomainSet set = domain_sets.get(entry.getHitType());
          if(set==null){
            set = new GPVDomainSet();
            set.setSetName(entry.getHitType());
            domain_sets.put(entry.getHitType(), set);
          }

          GPVDomain domain = new GPVDomain();
          domain.setParentSet(set);
          domain.setDescription(entry.getShortName().intern());
          domain.setProteinStartPos(entry.getFrom()-1); //-1 because they start counting at 1 and I start at 0
          domain.setProteinEndPos(entry.getTo()-1);//-1 because they start counting at 1 and I start at 0
          domain.setDomainId(entry.getID().intern());
          domain.setProteinId(entry.getProteinID().intern());
          addAdditionalData(domain, entry);
          
          set.addDomain(domain,false);

          float percentRead = (float)length_read/(float)file_size*100f;
          parent.setProgress((int)percentRead);

          if(Thread.currentThread().isInterrupted()){
            return;
          }
        }
        if(data.getAdditionalDomainSetInfoKeys()==null){
          data.setAdditionalDomainSetInfoKeys(new HashSet<String>());  
        }

        data.getAdditionalDomainSetInfoKeys().addAll(getAdditionalDataKeys());
        for(GPVDomainSet set:domain_sets.values())set.updateCache();
        data.setDomainSets(new ArrayList<GPVDomainSet>(domain_sets.values()));

        data.setDomainSource(getDomainIDSource());
      }
      catch (Exception e)
      {
        JOptionPane.showMessageDialog(parent, "Error loading Protein Domain Input file. " +e.getMessage());
        e.printStackTrace();
      }finally{
        parent.setProgress(0);
        if(reader!=null)
          try
        {
            reader.close();
        }
        catch (IOException e)
        {
          e.printStackTrace();
        }
      }

    }
  }

  public boolean validateFormat(File inputfile){
    AbstractDSVParser<T> parser = getParser();
    EntryReader<T> reader = null;
    try
    {
      reader = new EntryReader<T>(inputfile, parser);
      T entry = reader.nextParsedEntry();
      return validateEntry(entry);
    }catch(Exception e){
      //not my file format
      return false;
    }finally{
      try{
        if(reader!=null) reader.close();
      }catch(IOException e){
        e.printStackTrace();
      }
    }
  }


  protected abstract AbstractDSVParser<T> getParser();
  protected abstract Collection<String> getAdditionalDataKeys();
  protected abstract void addAdditionalData(GPVDomain domain, T entry);
  protected abstract boolean validateEntry(T entry);
  protected abstract GPVDataContainer.DOMAIN_SOURCE getDomainIDSource();
}
