/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data.loader.parser;

/*****************************************************************************************
 * Title:       BedEntry<br>
 * Description: .<br>
 * Package:     at.tugraz.genome.biojava.seq.bed<br>
 * Project:     [CommonBioJava]<br><br>
 *
 * Copyright:  Copyright (c) 2010<br>
 * Company:    Institute of Genomics and Bioinformatics, Graz University of Technology<br>
 * Date:       Apr 8, 2010<br>
 * @author     Maria Fischer, maria.fischer@tugraz.at<br>
 * @author     Rene Snajder, rene.snajder@i-med.ac.at<br>
 * @extends    {@link ParsedEntry}<br>
 * @implements {@link Comparable<T>}<br>
 ****************************************************************************************/

import java.util.List;

import at.tugraz.genome.biojava.exception.BedEntryException;
import at.tugraz.genome.biojava.io.ParsedEntry;
import at.tugraz.genome.biojava.seq.DNAStrand;

public class BedEntry implements ParsedEntry , Comparable<BedEntry> {
  
  // track information
  private String trackName;
  private String trackDescription;
  // "common" bed information
  private String chrom;
  private Long start;
  private Long end;
  private String name;
  private Long score;
  private DNAStrand strand;
  private Long thickStart;
  private Long thickEnd;
  private String itemRGB;
  private Integer blockCount;
  private List<Integer> blockSizes;
  private List<Integer> blockStarts;
  
  /**
   * Constructor for minimum required bed entries.
   * 
   * @param chrom
   *          chromosome name on which the genome feature exists
   * @param start
   *          zero-based feature starting position in the chromosome
   * @param end
   *          one-based feature ending position in the chromosome
   * @throws BedEntryException 
   */
  public BedEntry(String chrom, long start, long end) throws BedEntryException {
    setChrom(chrom);
    setStart(start);
    setEnd(end);
  }

  /**
   * Constructor for minimum required bed entries and track information
   * 
   * @param chrom
   *          chromosome name on which the genome feature exists
   * @param start
   *          zero-based feature starting position in the chromosome
   * @param end
   *          one-based feature ending position in the chromosome
   * @param trackName
   *          track name in which genome feature is stored
   * @param trackDescription
   *          track description in which genome feature is stored
   * @throws BedEntryException
   */
  public BedEntry(String chrom, long start, long end, String trackName, String trackDescription) throws BedEntryException {
    setChrom(chrom);
    setStart(start);
    setEnd(end);
    setTrackName(trackName);
    setTrackDescription(trackDescription);
  }
  
  /**
   * Copy constructor
   * 
   * @param template
   *          template BED entry
   * @throws BedEntryException 
   */
  public BedEntry(BedEntry template) throws BedEntryException {
    this(template.getChrom(), template.getStart(), template.getEnd());
    this.name = template.getName();
    this.score = template.getScore();
    this.strand = template.getStrand();
    this.thickStart = template.getThickStart();
    this.thickEnd = template.getThickEnd();
    this.itemRGB = template.getItemRgb();
    this.blockCount = template.getBlockCount();
    this.blockSizes = template.getBlockSizes();
    this.blockStarts = template.getBlockStarts();
  }
  
  /**
   * @return track name in which genome feature is stored
   */
  public String getTrackName() {
    return trackName;
  }

  /**
   * Sets track name
   * 
   * @param trackName
   *          track name in which genome feature is stored
   */
  public void setTrackName(String trackName) {
    this.trackName = trackName;
  }
  
  /**
   * @return track description in which genome feature is stored
   */
  public String getTrackDescription() {
    return this.trackDescription;
  }

  /**
   * Sets track description
   * 
   * @param trackDescription
   *          track description in which genome feature is stored
   */
  public void setTrackDescription(String trackDescription) {
    this.trackDescription = trackDescription;
  }
  
  /**
   * @return chromosome name on which the genome feature exists
   */
  public String getChrom() {
    return this.chrom;
  }
  
  /**
   * Sets chromosome name.
   * 
   * @param chrom
   *          chromosome name on which the genome feature exists
   * @throws BedEntryException 
   */
  public void setChrom(String chrom) throws BedEntryException {
    // required field
    if (chrom == null || chrom.isEmpty()) {
      throw new BedEntryException("Chromosome name is a required field. Therefore variable chrom can not be null or empty.");
    }
    this.chrom = chrom;
  }
  
  /**
   * @return zero-based feature starting position in the chromosome
   */
  public Long getStart() {
    return start;
  }

  /**
   * Sets zero-based feature starting position.
   * 
   * @param start
   *          zero-based feature starting position in the chromosome
   * @throws BedEntryException 
   */
  public void setStart(long start) throws BedEntryException {
    // zero-based position
    if (start < 0) {
      throw new BedEntryException("Feature starting position must not be negative.");
    }
    this.start = start;
  }
  
  /**
   * @return one-based feature ending position in the chromosome
   */
  public Long getEnd() {
    return this.end;
  }

  /**
   * Sets one-based feature ending position.
   * 
   * @param end
   *          one-based feature ending position in the chromosome
   * @throws BedEntryException 
   */
  public void setEnd(long end) throws BedEntryException {
    // zero-based position
    if (end < 1) {
      throw new BedEntryException("Feature ending position is one-based. Therefore variable end must be greater than zero.");
    }
    if (end <= getStart()) {
      throw new BedEntryException("Feature ending position is before feature starting position.");
    }
    this.end = end;
  }
  
  /**
   * @return BED feature name
   */
  public String getName() {
    return this.name;
  }

  /**
   * Sets BED feature name
   * 
   * @param name
   *          BED feature name
   */
  public void setName(String name) {
    this.name = name;
  }
  
  /**
   * @return BED feature score
   */
  public Long getScore() {
    return this.score;
  }

  /**
   * A score between 0 and 1000. If the track line useScore attribute is set to
   * 1 for the annotation data set, the score value will determine the level of
   * gray in which this feature is displayed (higher numbers = darker gray) in
   * UCSC Genome Browser.
   * 
   * @param score
   *          BED feature score
   * @throws BedEntryException 
   */
  public void setScore(long score) throws BedEntryException {
    // score ranges from 0 to 1000
    if (score < 0 || score > 1000) {
      throw new BedEntryException("Feature score must be between 0 and 1000.");
    }
    this.score = score;
  }

  /**
   * @return FORWARD if feature is on the + strand otherwise REVERSE
   */
  public DNAStrand getStrand() {
    return this.strand;
  }
  
  /**
   * Sets FORWARD/REVERSE strand
   * 
   * @param strand
   *          FORWARD/REVERSE depending on the feature's strand position
   */
  public void setStrand(DNAStrand strand) {
    this.strand = strand;
  }

  /**
   * Sets FORWARD/REVERSE strand
   * 
   * @param strand
   *          +/- depending on the feature's strand position
   * @throws BedEntryException 
   */
  public void setStrand(char strand) throws BedEntryException {
    if ('+' == strand) {
      setStrand(DNAStrand.FORWARD);
    } else if ('-' == strand) {
      setStrand(DNAStrand.REVERSE);
    } else {
      throw new BedEntryException("Invalid reverse strand character("+strand+"). Allowed characters are: + and -.");
    }
  }

  /**
   * @return starting position at which the feature is drawn thickly in UCSC
   *         Genome Browser
   */
  public Long getThickStart() {
    return this.thickStart;
  }

  /**
   * Sets starting position at which the feature is drawn thickly in UCSC Genome
   * Browser.
   * 
   * @param thickStart
   *          starting position at which feature is drawn thickly
   * @throws BedEntryException 
   */
  public void setThickStart(long thickStart) throws BedEntryException {
    // unfortunately the UCSC bed definition does not say if this position is
    // zero- or one-based so it thickStart is only checked to be larger than 0
    if (thickStart < 0) {
      throw new BedEntryException("Thick start must not be negative.");
    }
    if (thickStart < getStart()) {
      throw new BedEntryException("Thick starting position must not be smaller than feature starting position.");
    }
    this.thickStart = thickStart;
  }

  /**
   * @return ending position at which the feature is drawn thickly in UCSC
   *         Genome Browser
   */
  public Long getThickEnd() {
    return this.thickEnd;
  }

  /**
   * Sets ending position at which the feature is drawn thickly in in UCSC
   * Genome Browser.
   * 
   * @param thickEnd
   *          ending position at which feature is drawn thickly
   * @throws BedEntryException 
   */
  public void setThickEnd(long thickEnd) throws BedEntryException {
    // unfortunately the UCSC bed definition does not say if this position is
    // zero- or one-based so it thickStart is only checked to be larger than 1
    if (thickEnd < 1) {
      throw new BedEntryException("Thick ending position must be greater than zero.");
    } if (thickEnd > getEnd()) {
      throw new BedEntryException("Thick ending position must not be greater than feature ending position.");
    } if (thickEnd <= getThickStart()) {
      throw new BedEntryException("Thick ending position must be greater than thick starting position.");
    }
    this.thickEnd = thickEnd;
  }

  /**
   * @return RGB value of the form R,G,B
   */
  public String getItemRgb() {
    return this.itemRGB;
  }

  /**
   * Sets RGB value of the form R,G,B
   * 
   * @param itemRgb
   *          RGB value
   * @throws BedEntryException 
   */
  public void setItemRgb(String itemRgb) throws BedEntryException {
    if (itemRgb == null || itemRgb.isEmpty()) {
      throw new BedEntryException("ItemRgb must not be null or empty.");
    }
    this.itemRGB = itemRgb;
  }
  
  /**
   * @return number of blocks (exons) in the BED entry
   */
  public Integer getBlockCount() {
    return this.blockCount;
  }

  /**
   * Sets number of blocks (exons) in the BED entry
   * 
   * @param blockCount
   *          number of blocks in the BED entry
   * @throws BedEntryException 
   */
  public void setBlockCount(int blockCount) throws BedEntryException {
    if (blockCount < 0) {
      throw new BedEntryException("Block count must not be negative.");
    }
    this.blockCount = blockCount;
  }
  
  /**
   * @return comma-separated list of block sizes
   */
  public List<Integer> getBlockSizes() {
    return this.blockSizes;
  }
  
  /**
   * Sets block sizes.
   * 
   * @param blockSizes
   *          comma-separated list of block sizes.
   * @throws BedEntryException 
   */
  public void setBlockSizes(List<Integer> blockSizes) throws BedEntryException {
    if (blockSizes == null) {
      throw new BedEntryException("List blockSizes is null.");
    }
    if (getBlockCount() != blockSizes.size()) {
      throw new BedEntryException("Number of blockSizes is inconsistent with attribute blockCount.");
    }
    for (Integer blockSize : blockSizes) {
      if (blockSize < 0) {
        throw new BedEntryException("Invalid blockSize: " + blockSize + ".");
      }
    }
    this.blockSizes = blockSizes;
  }
  
  /**
   * @return comma-separated list of block starts
   */
  public List<Integer> getBlockStarts() {
    return this.blockStarts;
  }

  /**
   * Sets block starts.
   * 
   * @param blockStarts
   *          comma-separated list of block starts.
   * @throws BedEntryException 
   */
  public void setBlockStarts(List<Integer> blockStarts) throws BedEntryException {
    if (blockStarts == null) {
      throw new BedEntryException("List blockStarts is null.");
    }
    if (getBlockCount() != blockStarts.size()) {
      throw new BedEntryException("Number of blockStarts is inconsistent with attribute blockCount.");
    }
    long end   = getEnd();
    long start = getStart();
    for (Integer blockStart : blockStarts) {
      if (blockStart < start || blockStart >= end) {
        throw new BedEntryException("Invalid blockStart: " + blockStart + ".");
      }
    }
    this.blockStarts = blockStarts;
  }
  
  public long getLength() {
    return this.end - this.start;
  }

  public int compareTo(BedEntry o) {
    // order : chromosome, start, end
    int compChrom = getChrom().compareToIgnoreCase(o.getChrom());
    if (compChrom != 0) {
      return compChrom;
    }
    
    Long thisStart = getStart();
    Long thatStart = o.getStart();
    int compStart = thisStart.compareTo(thatStart);
    if (compStart != 0) {
      return compStart;
    }
    
    Long thisEnd = getEnd();
    Long thatEnd = o.getEnd();
    return thisEnd.compareTo(thatEnd);
  }
  
}
