/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data.loader.parser;

/*****************************************************************************************
 * Title:       BedParser<br>
 * Description: Parses bed files.<br>
 * Package:     at.tugraz.genome.biojava.ann.bed<br>
 * Project:     [CommonBioJava]<br><br>
 *
 * Copyright: Copyright (c) 2010<br>
 * Company:   Institute of Genomics and Bioinformatics, Graz University of Technology<br>
 * Date:      Apr 13, 2010<br>
 * @author    Maria Fischer, maria.fischer@tugraz.at<br>
 * @implements {@link StatefulRawEntryParserInterface}<br>
 ****************************************************************************************/

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;

import at.tugraz.genome.biojava.db.FormatDefinitionInterface;
import at.tugraz.genome.biojava.exception.BedEntryException;
import at.tugraz.genome.biojava.io.EntryCreationException;
import at.tugraz.genome.biojava.io.RawEntry;
import at.tugraz.genome.biojava.io.StatefulRawEntryParserInterface;

public class BedParser implements StatefulRawEntryParserInterface<BedEntry> {

  private static final String DEFAULT_COL_DELIMITER = "\t";
  private static final String TRACK_DELIMITER       = "track";
  private static final String TRACK_NAME_PREFIX     = "^.*name=";
  private static final String TRACK_NAME_POSTFIX    = " .*";
  private static final String TRACK_DSC_PREFIX      = "^.*description=\\\"";
  private static final String TRACK_DSC_POSTFIX     = "\\\".*";

  private static final String OTHER_HEADER_REGEX    = "^[^chr].*";

  private boolean parseTrackLine = true;

  private String trackName = "default";
  private String trackDescription = "default track";
  private final Pattern trackNamePrefix;
  private final Pattern trackNamePostfix;
  private final Pattern trackDscPrefix;
  private final Pattern trackDscPostfix;

  public BedParser() {
    trackNamePrefix  = Pattern.compile(TRACK_NAME_PREFIX);
    trackNamePostfix = Pattern.compile(TRACK_NAME_POSTFIX);
    trackDscPrefix  = Pattern.compile(TRACK_DSC_PREFIX);
    trackDscPostfix = Pattern.compile(TRACK_DSC_POSTFIX);
  }

  public FormatDefinitionInterface getFormatDefinition() {
    return new BedFormatDefinition();
  }

  public String[] getIds(String content) {
    if (content == null || content.isEmpty()) return null;
    String [] cols = content.split(DEFAULT_COL_DELIMITER);
    String [] ids = new String[1]; ids[0] = cols[0];
    return ids;
  }

  /**
   * Determines whether the given content holds track information.
   * 
   * @param content
   *          content of one line in a bed file
   * @return true if content is a track line, false if content is a "normal" bed
   *         entry
   */
  public boolean isTrackLine(String content) {

    return content != null && content.startsWith(TRACK_DELIMITER);
  }

  /**
   * Determines whether the given content is a header line and not actual data
   * 
   * @param content
   *          content of one line in a bed file
   * @return true if content is a header line, false if content is a "normal" bed
   *         entry
   */
  public boolean isOtherHeader(String content) {
    boolean ret = content != null && content.matches(OTHER_HEADER_REGEX);
    return ret;
  }

  /**
   * Extracts track name information of a bed track line.
   * 
   * @param content
   *          track line content in the form
   *          ^track name=\S+ description=".*"$
   * @return track name
   */
  private String getTrackName(String content) {
    String trackName = trackNamePrefix.matcher(content).replaceAll("");
    trackName = trackNamePostfix.matcher(trackName).replaceAll("");
    return trackName;
  }

  /**
   * Extracts track description information of a bed track line.
   * 
   * @param content
   *          track line content in the form
   *          ^track name=\S+ description=".*"$
   * @return track description
   */
  private String getTrackDescription(String content) {
    String trackDescription = trackDscPrefix.matcher(content).replaceAll("");
    trackDescription = trackDscPostfix.matcher(trackDescription).replaceAll("");
    return trackDescription;
  }

  public boolean isParseTrackLine()
  {
    return parseTrackLine;
  }

  public void setParseTrackLine(boolean parseTrackLine)
  {
    this.parseTrackLine = parseTrackLine;
  }

  /**
   * Returns next bed file entry or throws an exception if track information
   * line occurs, track information is only read and updated in the local
   * fields.
   */
  public BedEntry parseEntry(String content) throws EntryCreationException {
    if (content == null || content.isEmpty()) return null;
    content = content.replaceAll("\n", "");
    if (isTrackLine(content)) {
      if(parseTrackLine){
        // set track information
        this.trackName = getTrackName(content);
        this.trackDescription = getTrackDescription(content);
      }
      return null;
    }
    if(isOtherHeader(content)){
      return null;
    }

    String[] cols = content.split(DEFAULT_COL_DELIMITER);
    int length = cols.length;
    if (length < 3) return null; // first three entries are required
    BedEntry bedEntry = null;
    try {
      bedEntry = new BedEntry(cols[0], Long.parseLong(cols[1]), Long.parseLong(cols[2]), this.trackName, this.trackDescription);
    } catch (NumberFormatException e) {
      e.printStackTrace();
    } catch (BedEntryException e) {
      e.printStackTrace();
    }
    for(int i = 3; i < length; i++){
      try {
        switch(i){
          case 3:
            bedEntry.setName(cols[i]);
            break;
          case 4:
            bedEntry.setScore(Long.parseLong(cols[i]));
            break;
          case 5:
            bedEntry.setStrand(cols[i].charAt(0));
            break;
          case 6:
            bedEntry.setThickStart(Long.parseLong(cols[i]));
            break;
          case 7:
            bedEntry.setThickEnd(Long.parseLong(cols[i]));
            break;
          case 8:
            bedEntry.setItemRgb(cols[i]);
            break;
          case 9:
            bedEntry.setBlockCount(Integer.parseInt(cols[i]));
            break;
          case 10:
            // set block sizes
            List<Integer> blockSizes = new ArrayList<Integer>();
            for (String blockSize : cols[i].split(",")) {
              blockSizes.add(Integer.parseInt(blockSize));
            }
            bedEntry.setBlockSizes(blockSizes);
            break;
          case 11:
            // set block starts
            List<Integer> blockStarts = new ArrayList<Integer>();
            for (String blockStart : cols[i].split(",")) {
              blockStarts.add(Integer.parseInt(blockStart));
            }
            bedEntry.setBlockStarts(blockStarts);
            break;
        }
      } catch (NumberFormatException e) {
        e.printStackTrace();
      } catch (BedEntryException e) {
        e.printStackTrace();
      }
    }

    return bedEntry;
  }

  public BedEntry parseEntry(RawEntry entry) throws EntryCreationException {
    if (entry == null)
      return null;
    return parseEntry(entry.getContent());
  }

}
