/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data.loader.variants;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;

import javax.swing.JOptionPane;

import at.icbi.gpviz.GPViz;
import at.icbi.gpviz.data.GPVDataContainer;
import at.icbi.gpviz.data.GPVOptions;
import at.icbi.gpviz.data.regions.GPVHighlightedRegionSet;
import at.icbi.gpviz.data.regions.GPVVariant;
import at.icbi.gpviz.swing.MainWindow;
import at.tugraz.genome.biojava.io.EntryReader;
import at.tugraz.genome.biojava.io.ParsedEntry;
import at.tugraz.genome.biojava.io.RawEntry;
import at.tugraz.genome.biojava.io.RawEntryParserInterface;
import at.tugraz.genome.util.FileUtils;

public abstract class AbstractVariantLoader<T extends ParsedEntry>
{

  public void loadVariants(final MainWindow parent, File[] selectedfiles){
    GPVDataContainer data = GPViz.getInstance().getDataContainer();

    int total_newvariants = 0;
    ArrayList<GPVHighlightedRegionSet> newsets = new ArrayList<>();
    for(File selectedfile:selectedfiles){
      GPVHighlightedRegionSet set = new GPVHighlightedRegionSet();
      set.setNeedsDynamicVariantScheme(needsDynamicVariantScheme());
      set.setSetName(FileUtils.removeFileExtension(selectedfile.getName()));

      GPViz.getInstance().getOptions().setLastInputDir(selectedfile.getParent());
      GPVOptions.save(GPViz.getInstance().getOptions());
      RawEntryParserInterface<T> parser = getParser();
            EntryReader<T> reader = null;
      try{
        reader = new EntryReader<T>(selectedfile,parser);        
        long file_size = selectedfile.length();
        long length_read = reader.getHeaderLength();
        
        int newvariants = 0;

        RawEntry raw_entry;
        while((raw_entry = reader.next())!=null){
          length_read += raw_entry.getContent().length();
          T entry = reader.getParser().parseEntry(raw_entry);
          GPVVariant variant = getVariant(entry);
          set.addRegion(variant);
          newvariants++;
          float percentRead = (float)length_read/(float)file_size*100f;
          parent.setProgress((int)percentRead);
          
          if(Thread.currentThread().isInterrupted()){
            return;
          }
        }

        set.updateChildren();
        newsets.add(set);
        total_newvariants+=newvariants;
      }catch (Exception e){
        String error = "Error loading "+getFormatName()+" input file. ";
        if(reader!=null)error+="(line "+reader.getCurrentLineNumber()+") ";
        JOptionPane.showMessageDialog(parent, error +e.getMessage());
        e.printStackTrace();
      }finally{
        
      }
    }
    data.addHighlightedRegions(newsets);
    if(total_newvariants>0){
      data.refreshRegionsPerChromosomeCache();
      JOptionPane.showMessageDialog(parent, "Successfully loaded "+total_newvariants+" variants from "+getFormatName()+" file.");
    }
  }
  
  public boolean validateFormat(File inputfile){
    RawEntryParserInterface<T> parser = getParser();
    EntryReader<T> reader =null;
    try
    {
      reader = new EntryReader<T>(inputfile, parser);
      T entry = reader.nextParsedEntry();
      return validateEntry(entry);
    }catch(Exception e){
      //not my file format
      return false;
    }finally{
      try{
        if(reader!=null) reader.close();
      }catch(IOException e){
        e.printStackTrace();
      }
    }
  }

  protected abstract GPVVariant getVariant(T entry);
  protected abstract RawEntryParserInterface<T> getParser(); 
  protected abstract String getFormatName();
  protected abstract boolean validateEntry(T entry);
  protected abstract boolean needsDynamicVariantScheme();
}