/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data.regions;

import at.icbi.gpviz.data.GPVChromosome;
import at.icbi.gpviz.data.GPVGene;
import at.tugraz.genome.biojava.seq.DNAStrand;


public class GPVHighlightedRegion implements Comparable<GPVHighlightedRegion>
{
  private GPVChromosome chromosome;
  private int start;
  private int stop;
  private DNAStrand strand;
  private GPVHighlightedRegionSet parentSet;
  
  private boolean inExon = false;
  
  public GPVHighlightedRegion(GPVChromosome chromosome, int start, int stop, DNAStrand strand){
    this.chromosome = chromosome;
    this.start = start;
    this.stop = stop;
    this.strand = strand;
  }
  
  public GPVHighlightedRegion(){
    
  }
  
  public GPVChromosome getChromosome()
  {
    return chromosome;
  }
  public void setChromosome(GPVChromosome chromosome)
  {
    this.chromosome = chromosome;
  }
  public int getStart()
  {
    return start;
  }
  public void setStart(int start)
  {
    this.start = start;
  }
  public int getStop()
  {
    return stop;
  }
  public void setStop(int stop)
  {
    this.stop = stop;
  }
  
  public boolean isInExon()
  {
    return inExon;
  }

  public void setInExon(boolean inExon)
  {
    this.inExon = inExon;
  }
  
  public int getLength(){
    return Math.abs(getStop()-getStart());
  }

  public DNAStrand getStrand()
  {
    return strand;
  }

  public void setStrand(DNAStrand strand)
  {
    this.strand = strand;
  }
  
  public GPVHighlightedRegionSet getParentSet()
  {
    return parentSet;
  }

  public void setParentSet(GPVHighlightedRegionSet parentSet)
  {
    this.parentSet = parentSet;
  }
  
  @Override
  public int compareTo(GPVHighlightedRegion o)
  {
    int ret = chromosome.compareTo(o.getChromosome());
    if(ret!=0)return ret;
    
    ret = start-o.getStart();
    if(ret!=0)return ret;
    
    ret = stop-o.getStop();
    if(ret!=0)return ret;
    
    ret = getStrand().compareTo(o.getStrand());
    if(ret!=0)return ret;

    //they are all the same, but are they also of the same set?
    if(o.getParentSet()!=this.getParentSet()){
      ret = this.getParentSet().getSetName().compareTo(o.getParentSet().getSetName());
      if(ret==0){
        // if the set is different, but the name of the set is the same, 
        // We have to find some way to create a consistent sorting.
        return System.identityHashCode(this.getParentSet())-System.identityHashCode(o.getParentSet()); 
      }else{
        return ret;
      }
    }
    
    return 0;
  }
  
  public boolean contains(GPVHighlightedRegion region){
    if(!region.getChromosome().equals(this.getChromosome())){
      return false;
    }
    if(Math.min(region.getStart(),region.getStop())>=Math.min(this.getStart(), this.getStop()) &&
        Math.max(region.getStart(),region.getStop())<=Math.max(this.getStart(), this.getStop())){
      return true;
    }
    return false;
  }
  
  public boolean isInGene(GPVGene gene){
    if(getStrand()==gene.getStrand()){
      int chrom_compare = getChromosome().compareTo(gene.getChromosome());

      if(chrom_compare==0){
        boolean inrange = getStart() >= gene.getFirstKnownPosition() && getStart()<=gene.getLastKnownPosition();
        inrange = inrange || (getStop()<gene.getFirstKnownPosition() && getStop()>gene.getLastKnownPosition());
        inrange = inrange || (getStart()<gene.getFirstKnownPosition() && getStop()>gene.getLastKnownPosition());
        if(inrange){
          return true;
        }
      }
    }
    return false;
  }
}
