/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.data.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import at.icbi.gpviz.data.GPVDomain;
import at.icbi.gpviz.data.GPVExon;
import at.icbi.gpviz.data.GPVGene;
import at.icbi.gpviz.data.GPVTranscript;
import at.icbi.gpviz.data.regions.GPVHighlightedRegion;
import at.icbi.gpviz.data.regions.GPVVariant;
import at.tugraz.genome.biojava.seq.DNAStrand;

public class GPVUtils
{
  private static Pattern chromosome_pattern = Pattern.compile("^[chr|chrom|chromsome]*(.*)");

  public TreeSet<GPVExon> getAllExonRegionsSeparately(Collection<GPVExon> exons)
  {
    return getAllExonRegionsSeparately(exons, Integer.MIN_VALUE + 6, Integer.MAX_VALUE - 6);
  }

  public TreeSet<GPVExon> getAllExonRegionsSeparately(Collection<GPVExon> exons, int minpos, int maxpos)
  {
    TreeSet<GPVExon> ret = new TreeSet<GPVExon>();
    for (GPVExon exon:exons)
    {
      int[] mergesites = exon.getMergeSites();
      if (mergesites.length > 0)
      {
        for (int i = 0; i <= mergesites.length; i++)
        {
          int start = (i == 0) ? exon.getStartPos() : mergesites[i - 1];
          int stop = (i == mergesites.length) ? exon.getEndPos() : mergesites[i];

          if (Math.abs(start - stop) >= 3)
          {
            GPVExon part = new GPVExon(start, stop, exon.getStrand());
            if (part.getStartPos() >= minpos & part.getEndPos() <= maxpos)
            {
              ret.add(part);
            }
          }
        }
      } else
      {
        if (exon.getStartPos() >= minpos && exon.getStartPos() <= maxpos || exon.getEndPos() >= minpos && exon.getEndPos() <= maxpos
            || exon.getStartPos() < minpos && exon.getEndPos() > maxpos)
        {
          int start = Math.max(exon.getStartPos(), minpos);
          int stop = Math.min(exon.getEndPos(), maxpos);
          if (Math.abs(start - stop) >= 3)
          {
            ret.add(new GPVExon(start, stop, exon.getStrand()));
          }
        }
      }
    }

    return ret;
  }

  public GPVExon[] getExonsInRange(int start, int stop, DNAStrand strand, TreeSet<GPVExon> exons)
  {
    ArrayList<GPVExon> retlist = new ArrayList<GPVExon>();
    for (GPVExon totest:exons)
    {
      if (strand == totest.getStrand())
      {
        if (start <= totest.getEndPos() && start >= totest.getStartPos() || stop <= totest.getEndPos() && stop >= totest.getStartPos()
            || start <= totest.getStartPos() && stop >= totest.getEndPos())
        {
          retlist.add(totest);
        }
        if(totest.getStartPos()>stop){
          break;
        }
      }
    }
    return retlist.toArray(new GPVExon[0]);
  }

  public static TreeSet<GPVHighlightedRegion> splitRegionsInExonRegions(GPVGene selectedGene, Collection<GPVHighlightedRegion> allregions)
  {
    return splitRegionsInExonRegions(selectedGene.getAllExonRegionsSeparately(true, false), allregions);
  }

  public static TreeSet<GPVHighlightedRegion> splitRegionsInCodingRegions(GPVGene selectedGene, Collection<GPVHighlightedRegion> allregions)
  {
    return splitRegionsInExonRegions(selectedGene.getAllExonRegionsSeparately(false, true), allregions);
  }

  public static TreeSet<GPVHighlightedRegion> splitRegionsInCodingRegions(GPVTranscript transcript, Collection<GPVHighlightedRegion> allregions)
  {
    return splitRegionsInExonRegions(transcript.getAllExonRegionsSeparately(false), allregions);
  }

  private static TreeSet<GPVHighlightedRegion> splitRegionsInExonRegions(Collection<GPVExon> exons, Collection<GPVHighlightedRegion> allregions)
  {
    TreeSet<GPVHighlightedRegion> ret = new TreeSet<GPVHighlightedRegion>();
    for (GPVHighlightedRegion region:allregions)
    {
      for (GPVExon exon:exons)
      {
        if ((region.getStart() >= exon.getStartPos() && region.getStart() <= exon.getEndPos()) || region.getStop() >= exon.getStartPos()
            && region.getStop() <= exon.getEndPos() || region.getStart() <= exon.getStartPos() && region.getStop() >= exon.getEndPos())
        {
          // the range overlaps with this exon
          int start = Math.max(exon.getStartPos(), region.getStart());
          int stop = Math.min(exon.getEndPos(), region.getStop());

          if (region instanceof GPVVariant)
          {
            ret.add(region);
          } else
          {
            GPVHighlightedRegion exon_subregion = new GPVHighlightedRegion(region.getChromosome(), start, stop, region.getStrand());
            if (exon_subregion.getLength() > 0 || region.getLength() <= 0)
            {
              ret.add(exon_subregion);
            }
          }
        }
      }
    }

    return ret;
  }
  
  public static TreeSet<GPVVariant> splitRegionsInVariants(Collection<GPVHighlightedRegion> allregions)
  {
    TreeSet<GPVVariant> ret = new TreeSet<GPVVariant>();
    for (GPVHighlightedRegion region:allregions)
    {
     if(region instanceof GPVVariant){
       ret.add((GPVVariant)region);
     }
    }
    return ret;
  }

  public static TreeSet<GPVHighlightedRegion> splitRegionsInIntronRegions(GPVGene selectedGene, Collection<GPVHighlightedRegion> allregions)
  {
    TreeSet<GPVHighlightedRegion> ret = new TreeSet<GPVHighlightedRegion>();
    for (GPVHighlightedRegion region:allregions)
    {
      GPVExon previous_exon = null;
      for (GPVExon exon:selectedGene.getAllExonRegionsSeparately(true, false))
      {
        if (previous_exon != null)
        {
          int intron_start = previous_exon.getEndPos() + 1;
          int intron_stop = exon.getStartPos() - 1;
          if ((region.getStart() >= intron_start && region.getStart() <= intron_stop) || region.getStop() >= intron_start && region.getStop() <= intron_stop
              || region.getStart() <= intron_start && region.getStop() >= intron_stop)
          {
            // the range overlaps with this exon
            int start = Math.max(intron_start, region.getStart());
            int stop = Math.min(intron_stop, region.getStop());

            if (region instanceof GPVVariant)
            {
              ret.add(region);
            } else
            {
              GPVHighlightedRegion exon_subregion = new GPVHighlightedRegion(region.getChromosome(), start, stop, region.getStrand());
              if (exon_subregion.getLength() > 0 || region.getLength() <= 0)
              {
                ret.add(exon_subregion);
              }
            }
          }
        }
        previous_exon = exon;
      }
    }

    return ret;
  }

  public static TreeSet<GPVHighlightedRegion> splitRegionsInProteinDomainRegions(GPVTranscript transcript, Collection<GPVDomain> domains,
      Collection<GPVHighlightedRegion> allregions)
  {
    TreeSet<GPVHighlightedRegion> ret = new TreeSet<GPVHighlightedRegion>();
    for (GPVHighlightedRegion region:allregions)
    {
      for (GPVDomain domain:domains)
      {

        Collection<GPVExon> exons = transcript.getAllExonRegionsSeparately(false);
        boolean rev = transcript.getStrand() == DNAStrand.REVERSE;
        int domainstart = transformPPToNT((rev ? domain.getProteinEndPos() : domain.getProteinStartPos()), exons, rev);
        int domainend = transformPPToNT((rev ? domain.getProteinStartPos() : domain.getProteinEndPos()), exons, rev);

        if ((region.getStart() >= domainstart && region.getStart() <= domainend) || region.getStop() >= domainstart && region.getStop() <= domainend
            || region.getStart() <= domainstart && region.getStop() >= domainend)
        {
          // the range overlaps with this exon

          int start = Math.max(domainstart, region.getStart());
          int stop = Math.min(domainend, region.getStop());

          if (region instanceof GPVVariant)
          {
            ret.add(region);
          } else
          {
            GPVHighlightedRegion domain_subregion = new GPVHighlightedRegion(region.getChromosome(), start, stop, region.getStrand());
            if (domain_subregion.getLength() > 0 || region.getLength() <= 0)
            {
              ret.add(domain_subregion);
            }
          }
        }
      }
    }

    return ret;
  }

  public static ArrayList<GPVDomain> mergeProteinDomainsThatOverlap(Collection<GPVDomain> domains)
  {
    ArrayList<GPVDomain> ret = new ArrayList<GPVDomain>();

    for (GPVDomain domain:domains)
    {
      boolean found_overlap = false;
      for (GPVDomain other:ret)
      {
        if (other.overlaps(domain))
        {
          other.setProteinStartPos(Math.min(domain.getProteinStartPos(), other.getProteinStartPos()));
          other.setProteinEndPos(Math.max(domain.getProteinEndPos(), other.getProteinEndPos()));
          found_overlap = true;
          break;
        }
      }
      if (!found_overlap)
        ret.add(domain.clone());
    }

    return ret;
  }

  public static boolean reagionOverlaps(int starta, int stopa, int startb, int stopb)
  {
    return (startb <= stopa && startb >= starta || startb <= stopa && startb >= starta || startb <= starta && stopb >= stopa);
  }

  /**
   * Just divides by 3 and rounds properly.
   */
  public static int nt2Prot(int nt)
  {
    return (int) Math.floor(nt / 3f);
  }

  /**
   * Just multiplies by 3 and rounds properly.
   */
  public static int prot2Nt(int pp)
  {
    return (int) (pp * 3f);
  }

  /**
   * debug print exon list
   */
  public static void printExonList(Collection<GPVExon> exons)
  {
    printExonList(exons.toArray(new GPVExon[0]));
  }

  /**
   * debug print exon list
   */
  public static void printExonList(GPVExon[] exons)
  {
    for (GPVExon exon:exons)
    {
      System.out.println(exon.getStartPos() + "-" + exon.getEndPos());
    }
  }

  /**
   * Calculates the protein position based on the nucleotide position You must make sure that the exons you provide are only exons fully within the coding
   * region!!
   */
  public static int transformNTToPP(int nt, Collection<GPVExon> exons, boolean reverse_strand)
  {

    int total_exon_len = 0;
    for (GPVExon exon:exons)
    {
      total_exon_len += Math.abs(exon.getEndPos() - exon.getStartPos()) + 1;
    }

    int cumulative_exon_length = 0;
    for (GPVExon exon:exons)
    {
      if (nt >= exon.getStartPos() - 3 && nt <= exon.getEndPos() + 3)
      {
        // we found our range.
        int in_range_position = nt - Math.min(exon.getStartPos(), exon.getEndPos());
        if (reverse_strand)
        {
          int ret = (int) GPVUtils.nt2Prot(total_exon_len - cumulative_exon_length - in_range_position);
          return ret;
        }
        return GPVUtils.nt2Prot(cumulative_exon_length + in_range_position);

      } else
      {
        cumulative_exon_length += Math.abs(exon.getEndPos() - exon.getStartPos()) + 1;
      }
    }
    // System.err.println("Unknown nucleotide position "+nt+". Not in a defined exon range!");
    return 0;
  }

  /**
   * Calculates the nucleotide position based on the protein position You must make sure that the exons you provide are all the exons fully within the coding
   * region!!
   */
  public static int transformPPToNT(int pp, Collection<GPVExon> exons, boolean reverse_strand)
  {
    for (GPVExon exon:exons)
    {
      int startpos = transformNTToPP((reverse_strand ? exon.getEndPos() : exon.getStartPos()), exons, reverse_strand);
      int endpos = transformNTToPP((reverse_strand ? exon.getStartPos() : exon.getEndPos()), exons, reverse_strand);
      if (pp >= startpos && pp <= endpos)
      {
        // we found our range.
        int pp_in_range_position = pp - startpos;
        int nt_in_range_position = GPVUtils.prot2Nt(pp_in_range_position);
        int ret = 0;
        if (reverse_strand)
        {
          ret = (exon.getEndPos() - nt_in_range_position);
        } else
        {
          ret = (exon.getStartPos() + nt_in_range_position);
        }

        return ret;

      }
    }
    System.err.println("Unknown protein position " + pp);
    return 0;
  }
  
  public static String getChromsomeName(String ori){
    Matcher mitch = chromosome_pattern.matcher(ori);
    if(mitch.find()){
      return mitch.group(1);
    }
    return ori;
  }

}
