/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.drawing;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;
import java.util.TreeSet;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.ImageOutputStream;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;

import org.apache.batik.dom.GenericDOMImplementation;
import org.apache.batik.svggen.SVGGeneratorContext;
import org.apache.batik.svggen.SVGGraphics2D;
import org.apache.fop.pdf.PDFXMode;
import org.apache.fop.svg.PDFDocumentGraphics2D;
import org.apache.xmlgraphics.java2d.GraphicContext;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;

import at.icbi.gpviz.GPViz;
import at.icbi.gpviz.data.GPVDataContainer;
import at.icbi.gpviz.data.GPVGene;
import at.icbi.gpviz.data.GPVOptions;
import at.icbi.gpviz.swing.ExtensionBasedFileFilter;
import at.icbi.gpviz.swing.canvas.CanvasProperties;
import at.icbi.gpviz.swing.canvas.GPVCanvas;
import at.tugraz.genome.util.FileUtils;
import at.tugraz.genome.util.ProgressListener;

import com.sun.media.imageio.plugins.tiff.BaselineTIFFTagSet;
import com.sun.media.imageio.plugins.tiff.TIFFCompressor;
import com.sun.media.imageio.plugins.tiff.TIFFImageWriteParam;
import com.sun.media.imageioimpl.plugins.tiff.TIFFImageWriter;
import com.sun.media.imageioimpl.plugins.tiff.TIFFImageWriterSpi;
import com.sun.media.imageioimpl.plugins.tiff.TIFFLZWCompressor;

public class GPVImageWriter
{
  private CanvasProperties properties;
  private Dimension size;
  private Dimension printSizeMM = new Dimension(297, 210);
  private Component parent;
  private ProgressListener listener;

  public GPVImageWriter(CanvasProperties props, Dimension size, Component parent, ProgressListener listener)
  {
    this.properties = props;
    this.size = size;
    this.parent = parent;
    this.listener = listener;
  }

  public GPVImageWriter(CanvasProperties props, Dimension size, Component parent)
  {
    this.properties = props;
    this.size = size;
    this.parent = parent;
  }

  public GPVImageWriter(CanvasProperties props, Component parent)
  {
    this.properties = props;
    this.parent = parent;
  }

  public void writeGene(GPVDataContainer data, GPVGene gene)
  {
    File selectedFile = chooseOutputFile(parent,false);
    try{
      if (selectedFile != null)
      {
        writeGene(data, selectedFile, gene);
      }
    }
    catch (Exception e)
    {
      JOptionPane.showMessageDialog(this.parent, "Could not write image " + selectedFile.getName() + ". Error: " + e.getMessage());
      e.printStackTrace();
    }
  }

  public static void addAllImageFilters(JFileChooser chooser, boolean multipage_only)
  {
    String[] suffixes;
    if(multipage_only){
      suffixes= getValidFileFormatsWithMultipageSupport();
    }else{
      suffixes= getValidFileFormats();
    }
    for (String suffix:suffixes)
    {
      FileFilter filter = new ExtensionBasedFileFilter(suffix);
      chooser.addChoosableFileFilter(filter);
    }
  }

  public static File chooseOutputFile(Component parent, boolean multipage_only)
  {
    String lastsavedir = GPViz.getInstance().getOptions().getLastImageSaveDir();
    if (lastsavedir == null)
      lastsavedir = System.getProperty("user.home");

    JFileChooser chooser = new JFileChooser(lastsavedir);
    addAllImageFilters(chooser,multipage_only);
    int ret = chooser.showDialog(parent, "Save");

    if (ret == JFileChooser.APPROVE_OPTION)
    {
      GPViz.getInstance().getOptions().setLastImageSaveDir(chooser.getSelectedFile().getParent());
      GPVOptions.save(GPViz.getInstance().getOptions());
      File selectedFile = chooser.getSelectedFile();
      if (chooser.getFileFilter() instanceof ExtensionBasedFileFilter)
      {
        ExtensionBasedFileFilter fileFilter = (ExtensionBasedFileFilter) chooser.getFileFilter();
        String filename = chooser.getSelectedFile().getName() + "." + fileFilter.getExtension();
        selectedFile = new File(selectedFile.getParent(), filename);
      }
      return selectedFile;
    }
    return null;
  }

  public void writeGenesToMultipageFile(GPVDataContainer data, Collection<GPVGene> genes, File file) throws IOException
  {
    String format = FileUtils.getExtension(file.getName());
    if (!validateFormat(format, getValidFileFormats()))
    {
      format = "pdf";
      file = new File(FileUtils.replaceFileExtension(file.getAbsolutePath(), "." + format));
    }
    if (format.toLowerCase().equals("pdf"))
    {
      writeMultipagePDF(prepareCanvas(data), genes, file);
    }else if(format.toLowerCase().equals("tif") || format.toLowerCase().equals("tiff")){
      writeMultipageTiff(prepareCanvas(data), genes, file);
    }

  }

  public void writeGenes(GPVDataContainer data, Collection<GPVGene> genes, File dir, String extension) throws IOException
  {
    if (!extension.startsWith("."))
      extension = "." + extension;
    int nr_genes = genes.size();
    int i = 0;
    for (GPVGene gene:genes)
    {
      File file = new File(dir, "GPV_" + gene.getGeneSymbol() + extension);
      writeGene(data, file, gene);
      i++;
      listener.set(100 / nr_genes * i);

      if (Thread.currentThread().isInterrupted())
      {
        break;
      }
    }
    listener.close();
  }

  public void writeGene(GPVDataContainer data, File outputfile, GPVGene gene) throws IOException
  {
    GPVCanvas canvas = prepareCanvas(data);

    canvas.setSelectedGene(gene);

    String set_extension = FileUtils.getExtension(outputfile.getName());
    if (!validateFormat(set_extension, getValidFileFormats()))
    {
      set_extension = "png";
      outputfile = new File(FileUtils.replaceFileExtension(outputfile.getAbsolutePath(), "." + set_extension));
    }

    writeImage(canvas, set_extension, outputfile);


  }

  private boolean validateFormat(String ext, String[] valids)
  {
    boolean extension_is_valid = false;
    for (String validformatname:valids)
    {
      if (ext.toLowerCase().equals(validformatname))
      {
        extension_is_valid = true;
      }
    }
    return extension_is_valid;
  }

  public void writeImage(GPVCanvas canvas, String extension, File outputfile) throws IOException
  {
    BufferedImage image = new BufferedImage((int) this.size.getWidth(), (int) this.size.getHeight(), BufferedImage.TYPE_INT_RGB);
    canvas.paintNotFromBuffer(image.getGraphics());

    if (extension.equals("pdf"))
    {
      writePDF(canvas, outputfile);
    } else if (extension.equals("svg"))
    {
      writeSVG(canvas, outputfile);
    } else
    {
      // WRITE USING IMAGEIO
      ImageTypeSpecifier spec = ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_INT_RGB);
      Iterator<ImageWriter> supported_writers = ImageIO.getImageWriters(spec, extension);
      if (!supported_writers.hasNext())
      {
        throw new IOException("Unsupported format \"" + extension + "\"");
      }
      ImageWriter writer = supported_writers.next();
      ImageWriteParam imageWriteParam = writer.getDefaultWriteParam();

      FileOutputStream os = new FileOutputStream(outputfile);
      ImageOutputStream ios = ImageIO.createImageOutputStream(os);

      writer.setOutput(ios);

      try
      {
        // WRITE TIFF
        if (extension.equals("tiff") || extension.equals("tif"))
        {
          IIOMetadata metadata = writer.getDefaultImageMetadata(spec, imageWriteParam);
          IIOImage iioImage = new IIOImage(image, null, metadata);

          TIFFImageWriteParam tiffParam = (TIFFImageWriteParam) imageWriteParam;
          initTiffParameters(tiffParam);

          writer.prepareWriteSequence(metadata);
          writer.writeToSequence(iioImage, tiffParam);
          writer.endWriteSequence();

        } else
        {
          // WRITE ALL THE OTHER BORING FORMATS

          writer.write(image);
        }
      }
      finally
      {
        ios.close();
        os.close();
      }
    }
  }

  private GPVCanvas prepareCanvas(GPVDataContainer data)
  {
    GPVCanvas canvas = new GPVCanvas(data);
    canvas.setSize(this.size);
    canvas.setProperties(this.properties);
    return canvas;
  }

  private void writeSVG(GPVCanvas canvas, File outputfile) throws IOException
  {
    // WRITE SVG
    FileWriter writer = new FileWriter(outputfile);
    try
    {
      getSVG(canvas).stream(writer);
    }
    finally
    {
      writer.close();
    }
  }

  private SVGGraphics2D getSVG(GPVCanvas canvas)
  {
    // determine appropriate canvas size. if it's too small, the svg will be sloppy
    Dimension canvasSize = new Dimension(size);
    double minsize = 2000;
    double ratio = canvasSize.getHeight() / canvasSize.getWidth();
    if (canvasSize.getHeight() < minsize)
    {
      double newwidth = minsize / ratio;
      canvasSize.setSize(newwidth, minsize);
    }
    if (canvasSize.getWidth() < minsize)
    {
      double newheight = minsize * ratio;
      canvasSize.setSize(minsize, newheight);
    }

    canvas.setSize(canvasSize);

    DOMImplementation domImpl = GenericDOMImplementation.getDOMImplementation();
    String svgNS = "http://www.w3.org/2000/svg";
    Document document = domImpl.createDocument(svgNS, "svg", null);

    SVGGeneratorContext ctx = SVGGeneratorContext.createDefault(document);
    ctx.setEmbeddedFontsOn(true);
    ctx.setPrecision(6);

    SVGGraphics2D svgGenerator = new SVGGraphics2D(ctx, false);
    svgGenerator.setSVGCanvasSize(canvasSize);
    canvas.paintNotFromBuffer(svgGenerator);
    canvas.setSize(size);

    return svgGenerator;
  }

  private void writePDF(GPVCanvas canvas, File outputfile) throws IOException
  {
    // WRITE PDF
    FileOutputStream fos = new FileOutputStream(outputfile);
    PDFDocumentGraphics2D g2 = new PDFDocumentGraphics2D(true, fos, (int) size.getWidth(), (int) size.getHeight());
    g2.getPDFDocument().getInfo().setTitle(GPViz.PROGRAM_NAME+" "+GPViz.VERSION+" export of 1 gene");
    g2.getPDFDocument().getProfile().setPDFXMode(PDFXMode.PDFX_3_2003);
    GraphicContext context = new GraphicContext();
    g2.setGraphicContext(context);
    g2.create();
    canvas.paintNotFromBuffer(g2);
    g2.finish();
    fos.close();
  }

  private void writeMultipagePDF(GPVCanvas canvas, Collection<GPVGene> selectedGenes, File outputfile) throws IOException
  {
    // WRITE PDF
    int nr_genes = selectedGenes.size();
    FileOutputStream fos = new FileOutputStream(outputfile);
    PDFDocumentGraphics2D g2 = new PDFDocumentGraphics2D(true, fos, (int) size.getWidth(), (int) size.getHeight());
    g2.getPDFDocument().getProfile().setPDFXMode(PDFXMode.PDFX_3_2003);
    g2.getPDFDocument().getInfo().setTitle(GPViz.PROGRAM_NAME+" "+GPViz.VERSION+" export of "+nr_genes+" genes");
    
    g2.setGraphicContext(new GraphicContext());
    
    g2.create();

    Iterator<GPVGene> it = selectedGenes.iterator();
    
    int i = 0;
    while (it.hasNext())
    {
      canvas.setSelectedGene(it.next());
      canvas.paintNotFromBuffer(g2);
      if(it.hasNext())g2.nextPage();
      i++;
      listener.set(100 / nr_genes * i);

      fos.flush();
      if (Thread.currentThread().isInterrupted())
      {
        fos.close(); return;
      }
    }
    g2.finish();
    fos.close();
    listener.close();
  }
  
  private void initTiffParameters(TIFFImageWriteParam tiffParam){
    TIFFCompressor compressor = new TIFFLZWCompressor(BaselineTIFFTagSet.PREDICTOR_NONE);
    tiffParam.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
    tiffParam.setTIFFCompressor(compressor);
    tiffParam.setCompressionType(compressor.getCompressionType());
  }
  
  private void writeMultipageTiff(GPVCanvas canvas, Collection<GPVGene> selectedGenes, File outputfile) throws IOException{
    // WRITE USING IMAGEIO
    int nr_genes = selectedGenes.size();
    
    ImageTypeSpecifier spec = ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_INT_RGB);
    
    TIFFImageWriter writer = new TIFFImageWriter(new TIFFImageWriterSpi());
    
    ImageWriteParam imageWriteParam = writer.getDefaultWriteParam();
    FileOutputStream os = new FileOutputStream(outputfile);
    ImageOutputStream ios = ImageIO.createImageOutputStream(os);
    writer.setOutput(ios);
    
    IIOMetadata metadata = writer.getDefaultImageMetadata(spec, imageWriteParam);
    TIFFImageWriteParam tiffParam = (TIFFImageWriteParam) imageWriteParam;
    initTiffParameters(tiffParam);
    writer.prepareWriteSequence(metadata);
    Iterator<GPVGene> it = selectedGenes.iterator();
    int i = 0;
    while (it.hasNext())
    {
      canvas.setSelectedGene(it.next());
      BufferedImage image = new BufferedImage((int) this.size.getWidth(), (int) this.size.getHeight(), BufferedImage.TYPE_INT_RGB);
      canvas.paintNotFromBuffer(image.getGraphics());
      IIOImage iioImage = new IIOImage(image, null, metadata);
      writer.writeToSequence(iioImage, tiffParam);
      i++;
      listener.set(100 / nr_genes * i);
      if (Thread.currentThread().isInterrupted())
      {
        os.close(); return;
      }
    }
    writer.endWriteSequence();
    ios.close();
    os.close();
    listener.close();
  }

  public static String[] getValidFileFormats()
  {
    String[] common = ImageIO.getWriterFileSuffixes();

    TreeSet<String> ret = new TreeSet<String>();
    for (String toadd:common)
    {
      if (toadd.trim().length() > 0)
      {
        // check if there really is an imagewriter for this - imageio tends to lie about the supported formats..
        ImageTypeSpecifier spec = ImageTypeSpecifier.createFromRenderedImage(new BufferedImage(1, 1, BufferedImage.TYPE_INT_RGB));
        if (ImageIO.getImageWriters(spec, toadd).hasNext())
        {
          ret.add(toadd);
        }
      }
    }
    ret.add("pdf");
    ret.add("svg");
    return ret.toArray(new String[0]);
  }

  public static String[] getValidFileFormatsWithMultipageSupport()
  {
    TreeSet<String> ret = new TreeSet<String>();
    ret.add("pdf");
    ret.add("tif");
    ret.add("tiff");
    return ret.toArray(new String[0]);
  }

}
