/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.text.NumberFormat;

import javax.swing.DefaultComboBoxModel;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import com.jgoodies.forms.factories.FormFactory;
import com.jgoodies.forms.layout.ColumnSpec;
import com.jgoodies.forms.layout.FormLayout;
import com.jgoodies.forms.layout.RowSpec;

public class ImageSizePanel extends JPanel implements ItemListener, DocumentListener
{
  private static final long serialVersionUID = -2972357663687587755L;
  private JTextField txtWidth;
  private JTextField txtHeight;
  private JTextField txtResolution;
  private JCheckBox chckbxRatio = new JCheckBox("Lock ratio");
  private double ratio = 1f;
  
  private double width;
  private double height;
  private int resolution;

  private enum UNIT {pixels,mm,inch};
  
  private static final double INCH_TO_MM = 0.0393701d;
  
  private JComboBox<UNIT> comboBoxUnit = new JComboBox<UNIT> ();
  
  private UNIT unit;
  
  private JLabel lblError = new JLabel("");
  private boolean ignoreDocumentListenerEvents = false;

  public ImageSizePanel(){
    this(1920,1200);
  }
  
  /**
   * Create the panel.
   */
  public ImageSizePanel(int initial_width, int initial_height)
  {
    this.width = initial_width;
    this.height = initial_height;
    this.resolution = 300;
    this.unit = UNIT.pixels;
    this.ratio = this.height/this.width;
    
    setLayout(new FormLayout(new ColumnSpec[] {
        FormFactory.RELATED_GAP_COLSPEC,
        FormFactory.DEFAULT_COLSPEC,
        FormFactory.RELATED_GAP_COLSPEC,
        FormFactory.DEFAULT_COLSPEC,
        FormFactory.RELATED_GAP_COLSPEC,
        FormFactory.DEFAULT_COLSPEC,},
      new RowSpec[] {
        FormFactory.RELATED_GAP_ROWSPEC,
        FormFactory.DEFAULT_ROWSPEC,
        FormFactory.RELATED_GAP_ROWSPEC,
        FormFactory.DEFAULT_ROWSPEC,
        FormFactory.RELATED_GAP_ROWSPEC,
        FormFactory.DEFAULT_ROWSPEC,
        FormFactory.RELATED_GAP_ROWSPEC,}));
    
    JLabel lblWidth = new JLabel("Width:");
    add(lblWidth, "2, 2, right, default");
    
    txtWidth = new JTextField();
    add(txtWidth, "4, 2, left, default");
    txtWidth.setColumns(10);
    comboBoxUnit.setModel(new DefaultComboBoxModel(UNIT.values()));
    comboBoxUnit.addItemListener(this);
    
    add(comboBoxUnit, "6, 2, left, default");
    
    JLabel lblHeight = new JLabel("Height:");
    add(lblHeight, "2, 4, right, default");
    
    txtHeight = new JTextField();
    add(txtHeight, "4, 4, fill, default");
    txtHeight.setColumns(10);
    chckbxRatio.setSelected(true);
    
    
    add(chckbxRatio, "6, 4");
    chckbxRatio.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if(chckbxRatio.isSelected()){
          recalculateRatio();
        }
      }
    });
    
    JLabel lblres = new JLabel("Resolution:");
    add(lblres, "2, 6, right, default");
    
    txtResolution = new JTextField();
    txtResolution.setText("300");
    txtResolution.setEnabled(false);
    add(txtResolution, "4, 6, fill, default");
    txtResolution.setColumns(10);
    
    txtHeight.getDocument().addDocumentListener(this);
    txtWidth.getDocument().addDocumentListener(this);
    txtResolution.getDocument().addDocumentListener(this);
    
    JLabel lbldpi = new JLabel("dpi");
    add(lbldpi, "6, 6");
    
    
    lblError.setForeground(Color.red);
    
    updateTextFromValues();
  }

  private int parseDPI(){
    return Integer.parseInt(txtResolution.getText());
  }
  
  private double parseWidth(){
    if(this.unit == UNIT.pixels){
      return Integer.parseInt(txtWidth.getText());  
    }
    return Double.parseDouble(txtWidth.getText());
  }
  
  private double parseHeight(){
    if(this.unit == UNIT.pixels){
      return Integer.parseInt(txtHeight.getText());  
    }
    return Double.parseDouble(txtHeight.getText());
  }
  
  private boolean updateValues(){
    String base_error = "Invalid ";
    try{
      this.width = parseWidth();
    }catch(NumberFormatException nfe){
      this.lblError.setText(base_error+"width");
      return false;
    }
    try{
      this.height= parseHeight();
    }catch(NumberFormatException nfe){
      this.lblError.setText(base_error+"height");
      return false;
    }
    try{
      this.resolution= parseDPI();
      this.lblError.setText(base_error+"dpi");
    }catch(NumberFormatException nfe){
      return false;
    }
    this.lblError.setText("");
    return true;
  }
  
  private void updateTextFromValues(){
    updateTextFromValues(true,true);
  }
  private void updateTextFromValues(boolean dowidth, boolean doheight){
    ignoreDocumentListenerEvents = true;
    String height;
    String width;
    
    if(unit == UNIT.pixels){
      height = ""+((int)this.height);
      width = ""+((int)this.width);
    }else{
      height = formatDouble(this.height);
      width = formatDouble(this.width);
    }

    if(doheight)txtHeight.setText(height);
    if(dowidth)txtWidth.setText(width);
    
    txtResolution.setText(""+this.resolution);
    ignoreDocumentListenerEvents = false;
  }
  
  public Dimension getSizeInPixels(){
    if(unit == UNIT.pixels) return new Dimension((int)this.width,(int)this.height);
    double height_inch;
    double width_inch;
    
    if(unit == UNIT.inch){
      height_inch = this.height;
      width_inch = this.width;
    }else{
      //MM
      height_inch = this.height*INCH_TO_MM;
      width_inch = this.width*INCH_TO_MM;
    }
    
    return new Dimension((int)(width_inch*this.resolution), (int)(height_inch*this.resolution));
  }
  
  private String formatDouble(double d){
    NumberFormat f = NumberFormat.getInstance();
    f.setMinimumFractionDigits(2);
    f.setGroupingUsed(false);
    return f.format(d);
  }
  
  private void setSizeFromPixels(Dimension d){
    if(unit == UNIT.pixels){
      this.width = d.getWidth();
      this.height = d.getHeight();
    }else{
      double height_inch = d.getHeight()/this.resolution;
      double width_inch = d.getWidth()/this.resolution;
      
      if(unit == UNIT.inch){
        this.width = width_inch;
        this.height = height_inch;
      }else{
        //MM
        this.width = width_inch/INCH_TO_MM;
        this.height = height_inch/INCH_TO_MM;
      }
    }
    updateTextFromValues();
  }
  
  
  @Override
  public void itemStateChanged(ItemEvent e)
  {
    if(e.getStateChange()==ItemEvent.SELECTED){
      UNIT selectedunit = (UNIT) comboBoxUnit.getSelectedItem();
      Dimension pixelsize = getSizeInPixels();
      this.unit = selectedunit;
      setSizeFromPixels(pixelsize);
      
      this.txtResolution.setEnabled(this.unit!=UNIT.pixels);
      
      updateValues();
    }
  }
  @Override
  public void setEnabled(boolean enabled){
    this.txtHeight.setEnabled(enabled);
    this.txtWidth.setEnabled(enabled);
    this.chckbxRatio.setEnabled(enabled);
    this.txtResolution.setEnabled(enabled);
    this.comboBoxUnit.setEnabled(enabled);
    super.setEnabled(enabled);
  }
  
  protected void recalculateRatio(){
    if(updateValues()){
      ratio = this.height/this.width;
    }
  }

  @Override
  public void insertUpdate(DocumentEvent e)
  {
    changedUpdate(e);    
  }

  @Override
  public void removeUpdate(DocumentEvent e)
  {
    changedUpdate(e);
  }

  public void setErrorLabel(JLabel error){
    lblError = error;
  }
  
  @Override
  public void changedUpdate(DocumentEvent e)
  {
    if(!ignoreDocumentListenerEvents){
      if(updateValues()){
        if(chckbxRatio.isSelected()){
          if(e.getDocument() == txtHeight.getDocument()){
            this.width = height/ratio;
            updateTextFromValues(true,false);
          }
          if(e.getDocument() == txtWidth.getDocument()){
            this.height = width*ratio;
            updateTextFromValues(false,true);
          }
        }
      }    
    }
  }
}
