/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing;

import java.awt.Adjustable;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowStateListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Box;
import javax.swing.ButtonGroup;
import javax.swing.DefaultListModel;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JLayeredPane;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.JSlider;
import javax.swing.JSplitPane;
import javax.swing.JTextField;
import javax.swing.JToggleButton;
import javax.swing.JToolBar;
import javax.swing.ListModel;
import javax.swing.ScrollPaneConstants;
import javax.swing.SpringLayout;
import javax.swing.SwingWorker;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import at.icbi.gpviz.GPViz;
import at.icbi.gpviz.data.DataChangedEvent;
import at.icbi.gpviz.data.DataChangedEvent.CHANGEACTION;
import at.icbi.gpviz.data.DataChangedListener;
import at.icbi.gpviz.data.GPVGene;
import at.icbi.gpviz.data.GPVGene.DISPLAYNAME;
import at.icbi.gpviz.data.GPVOptions;
import at.icbi.gpviz.data.filter.DomainsAffectedGeneFilter;
import at.icbi.gpviz.data.filter.GeneFilter;
import at.icbi.gpviz.data.filter.GeneFilterCache;
import at.icbi.gpviz.data.filter.GlobBasedGeneSymbolFilter;
import at.icbi.gpviz.data.filter.HighlightedRegionsGeneFilter;
import at.icbi.gpviz.data.filter.MinNumberOfCodingTranscriptsGeneFilter;
import at.icbi.gpviz.data.filter.VariantsIntersectingRegionsGeneFilter;
import at.icbi.gpviz.data.sorting.AlphabeticalGeneSorter;
import at.icbi.gpviz.data.sorting.GeneSorter;
import at.icbi.gpviz.data.sorting.GenomicOrderGeneSorter;
import at.icbi.gpviz.swing.canvas.CanvasPopUpMenu;
import at.icbi.gpviz.swing.canvas.CanvasPropertiesMenu;
import at.icbi.gpviz.swing.canvas.GPVCanvas;
import at.icbi.gpviz.swing.canvas.GPVCanvasDecoLayer;
import at.icbi.gpviz.swing.detailpane.DetailsTablePanel;
import at.icbi.gpviz.swing.options.ProgramOptionsDialog;
import at.icbi.gpviz.swing.sidebar.DisplayPropertiesPanel;
import at.icbi.gpviz.swing.sidebar.Sidebar;

public class MainWindow extends JFrame implements ActionListener, MouseListener, KeyListener, ListSelectionListener, MouseWheelListener, MouseMotionListener, ChangeListener, DataChangedListener
{
  private static long memory_warning_upper_threshold = GPViz.MEMORY_WARNING_UPPER_THRESHOLD;
  private static long memory_warning_lower_threshold = GPViz.MEMORY_WARNING_LOWER_THRESHOLD;
  
  private static final long serialVersionUID = -3166890028027844135L;

  private GPVCanvas canvas;
  
  private GPVCanvasDecoLayer canvasDecoLayer;

  protected ListModel<GPVGene> fullGeneListModel;
  
  private final int displayPanelWidth=175;

  // GUI elements
  protected JList<GPVGene> geneList;
  private JMenuItem mntmExit = new JMenuItem("Exit");
  private JButton btnLoadGenome = new JButton("Load Gene Models");
  private JButton btnLoadBEDRegions = new JButton("Load Regions");
  private JButton btnLoadVariants = new JButton("Load Variants");
  private JButton btnLoadDomains = new JButton("Load Domains");
  private JToggleButton btnSetSize = new JToggleButton("");
  private JToggleButton btnShowHelperLines = new JToggleButton("");
  private JLabel btnClearSearch = new JLabel("");
  protected JTextField searchField = new JTextField();
  private JProgressBar progressBar = new JProgressBar();
  private JProgressBar memoryBar = new JProgressBar();
  private JToggleButton btnTrackLabels = new JToggleButton("");
  protected Sidebar sidebar = new Sidebar(this); 

  private JButton btnCancelAction ;
  
  private JMenuItem toolsOptionsMenuItem = new JMenuItem("Options");
  private JCheckBoxMenuItem mntmFilterHighlight = new JCheckBoxMenuItem("Genes with regions/variants");
  private JCheckBoxMenuItem mntmFilterRVIntersect = new JCheckBoxMenuItem("Genes with regions intersecting variants");
  private JCheckBoxMenuItem mntmFilterDomainsEffected = new JCheckBoxMenuItem("Genes with regions/variants on domains");
  private JCheckBoxMenuItem mntmFilterMinCodingTranscript = new JCheckBoxMenuItem("Genes with >1 coding transcript");
  private JMenuItem mntmClearFilter = new JMenuItem("All Genes");
  private JMenuItem mntmUnloadFiltered = new JMenuItem("Unload filtered genes...");
  private JButton btnSaveImage = new JButton("");
  private JToggleButton btnShowLegend = new JToggleButton("");
  private JMenuItem mntmSaveCurr = new JMenuItem("Save current image...");
  private JMenuItem mntmSaveSelected = new JMenuItem("Save images for selected genes...");
  private JMenuItem mntmSaveAll = new JMenuItem("Save images for all filtered genes...");
  private JToggleButton btnHighlight = new JToggleButton("");
  private JMenuItem mntmLoadRegions = new JMenuItem("Load regions to highlight from file...");
  private JMenuItem mntmLoadVariants = new JMenuItem("Load variants to highlight from file...");
  private JMenuItem mntmLoadGenome = new JMenuItem("Load gene models from file...");
  private JMenuItem mntmLoadDomains = new JMenuItem("Load domain information from File...");
  private JMenuItem mntmAbout = new JMenuItem("About");
  private JMenu menuSetGeneDisplayedAs = new JMenu("Display genes as:");
  private JRadioButtonMenuItem mntmSetGeneDisplayedAsSymbol = new JRadioButtonMenuItem("Gene symbol");
  private JRadioButtonMenuItem mntmSetGeneDisplayedAsID = new JRadioButtonMenuItem("Gene ID");
  private JRadioButtonMenuItem mntmSetGeneDisplayedAsName = new JRadioButtonMenuItem("Gene name");
  private JMenu menuSortGenesBy = new JMenu("Sort genes by:");
  private JRadioButtonMenuItem mntmSortGenesBySymbol = new JRadioButtonMenuItem("Gene symbol");
  private JRadioButtonMenuItem mntmSortGenesByID = new JRadioButtonMenuItem("Gene ID");
  private JRadioButtonMenuItem mntmSortGenesByName = new JRadioButtonMenuItem("Gene name");
  private JRadioButtonMenuItem mntmSortGenesByChromosome = new JRadioButtonMenuItem("Chromosome");
  private JRadioButtonMenuItem mntmSortAscending = new JRadioButtonMenuItem("Ascending");
  private JRadioButtonMenuItem mntmSortDescending = new JRadioButtonMenuItem("Descending");
  private JCheckBoxMenuItem showMemory = new JCheckBoxMenuItem("Show memory usage");
  private JSlider sliderProteinTrackHeight = new JSlider();
  private JSlider sliderExonSpace = new JSlider();
  private final JToggleButton btnExonsToScale = new JToggleButton("");
  private final JToggleButton btnShowSidebar = new JToggleButton("");
  private JSplitPane imageSplitPane = new JSplitPane();
  private final Component horizontalStrut = Box.createHorizontalStrut(100);
  private FocusTrappingCancelListener asynchronousActionListener;
  private JScrollPane canvasScrollPane = new JScrollPane();
  
  protected GeneFilterCache geneFilterCache = new GeneFilterCache();
  
  //got to reserve the memory while we can:
  private LowMemoryDialog lowMemoryDialog = new LowMemoryDialog();
  private boolean ignoredLowMemory = false;
  
  private DetailsTablePanel statsTabbedPane = new DetailsTablePanel();
  
  private GeneSorter geneListSortOrder = new AlphabeticalGeneSorter(GPVGene.DISPLAYNAME.SYMBOL,false);
  
  /**
   * Create the application.
   */
  public MainWindow()
  {
    setIconImage(Toolkit.getDefaultToolkit().getImage(MainWindow.class.getResource("/at/icbi/gpviz/icons/GPVLogo64.png")));
    initialize();
    GPViz.getInstance().getDataContainer().addDataChangeListener(this);
  }

  public GPVCanvas getCanvas()
  {
    return this.canvas;
  }

  /**
   * Initialize the contents of the frame.
   */
  private void initialize()
  {
    this.setTitle("GeneProteinViz");
    Dimension screensize = Toolkit.getDefaultToolkit().getScreenSize();
    this.setBounds(100, 100, (int) Math.min(1024,screensize.getWidth()-100), (int) Math.min(768,screensize.getHeight()-100));
    this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    
    
    canvas = new GPVCanvas(GPViz.getInstance().getDataContainer());
    
    JMenuBar menuBar = new JMenuBar();
    this.setJMenuBar(menuBar);

    JMenu mnFile = new JMenu("File");
    menuBar.add(mnFile);
    
    mntmExit.addActionListener(this);
    mnFile.add(mntmLoadGenome);
    mntmLoadGenome.addActionListener(this);

    mnFile.add(mntmLoadRegions);
    mnFile.add(mntmLoadVariants);
    mnFile.add(mntmLoadDomains);
    mntmLoadRegions.addActionListener(this);
    mntmLoadDomains.addActionListener(this);
    mntmLoadVariants.addActionListener(this);

    JSeparator separator_1 = new JSeparator();
    mnFile.add(separator_1);

    mnFile.add(mntmSaveCurr);
    mnFile.add(mntmSaveSelected);
    mnFile.add(mntmSaveAll);
    mntmSaveSelected.addActionListener(this);
    mntmSaveCurr.addActionListener(this);
    mntmSaveAll.addActionListener(this);

    JSeparator separator_2 = new JSeparator();
    mnFile.add(separator_2);
    mnFile.add(mntmExit);
    
    JMenu mnFilter = new JMenu("Filter");
    menuBar.add(mnFilter);
    
    mntmFilterHighlight.addActionListener(this);
    mnFilter.add(mntmFilterHighlight);
    
    mntmFilterDomainsEffected.addActionListener(this);
    mnFilter.add(mntmFilterDomainsEffected);
    
    mntmFilterRVIntersect.addActionListener(this);
    mnFilter.add(mntmFilterRVIntersect);
    
    mntmFilterMinCodingTranscript.addActionListener(this);
    mnFilter.add(mntmFilterMinCodingTranscript);
    mnFilter.add(new JSeparator());
    mnFilter.add(mntmClearFilter);
    mntmClearFilter.addActionListener(this);
    mnFilter.add(new JSeparator());
    mntmUnloadFiltered.addActionListener(this);
    mnFilter.add(mntmUnloadFiltered);
    
    CanvasPropertiesMenu propsmenu = new CanvasPropertiesMenu(canvas);
    menuBar.add(propsmenu);

    JMenu mnTools = new JMenu("Tools");
    menuBar.add(mnTools);
    
    toolsOptionsMenuItem.addActionListener(this);
    mnTools.add(toolsOptionsMenuItem);

    JMenu mnHelp = new JMenu("Help");
    menuBar.add(mnHelp);
    mnHelp.add(showMemory);
    mnHelp.add(mntmAbout);
    mntmAbout.addActionListener(this);
    showMemory.addActionListener(this);
    this.getContentPane().setLayout(new BorderLayout(0, 0));

    JSplitPane mainSplitPane = new JSplitPane();
    mainSplitPane.setOneTouchExpandable(true);
    this.getContentPane().add(mainSplitPane);
    mainSplitPane.setDividerLocation(180);

    JPanel leftPanel = new JPanel();
    leftPanel.setMinimumSize(new Dimension(180, 500));
    mainSplitPane.setLeftComponent(leftPanel);
    SpringLayout sl_leftPanel = new SpringLayout();
    leftPanel.setLayout(sl_leftPanel);

    btnLoadGenome.addActionListener(this);
    btnLoadGenome.setPreferredSize(new Dimension(160, 20));
    sl_leftPanel.putConstraint(SpringLayout.WEST, btnLoadGenome, 10, SpringLayout.WEST, leftPanel);
    leftPanel.add(btnLoadGenome);

    btnLoadBEDRegions.addActionListener(this);
    btnLoadBEDRegions.setPreferredSize(new Dimension(160, 20));
    sl_leftPanel.putConstraint(SpringLayout.NORTH, btnLoadBEDRegions, 6, SpringLayout.SOUTH, btnLoadGenome);
    sl_leftPanel.putConstraint(SpringLayout.WEST, btnLoadBEDRegions, 10, SpringLayout.WEST, leftPanel);
    leftPanel.add(btnLoadBEDRegions);
    
    btnLoadVariants.addActionListener(this);
    btnLoadVariants.setPreferredSize(new Dimension(160, 20));
    sl_leftPanel.putConstraint(SpringLayout.NORTH, btnLoadVariants, 6, SpringLayout.SOUTH, btnLoadBEDRegions);
    sl_leftPanel.putConstraint(SpringLayout.WEST, btnLoadVariants, 10, SpringLayout.WEST, leftPanel);
    leftPanel.add(btnLoadVariants);

    btnLoadDomains.setPreferredSize(new Dimension(160, 20));
    btnLoadDomains.addActionListener(this);
    sl_leftPanel.putConstraint(SpringLayout.NORTH, btnLoadDomains, 6, SpringLayout.SOUTH, btnLoadVariants);
    sl_leftPanel.putConstraint(SpringLayout.WEST, btnLoadDomains, 0, SpringLayout.WEST, btnLoadGenome);
    leftPanel.add(btnLoadDomains);

    JSeparator leftSeparator = new JSeparator();
    sl_leftPanel.putConstraint(SpringLayout.NORTH, leftSeparator, 10, SpringLayout.SOUTH, btnLoadDomains);
    leftSeparator.setPreferredSize(new Dimension(170, 1));
    sl_leftPanel.putConstraint(SpringLayout.WEST, leftSeparator, -5, SpringLayout.WEST, btnLoadDomains);
    leftPanel.add(leftSeparator);

    JLabel lblGenelist = new JLabel("Select Gene:");
    sl_leftPanel.putConstraint(SpringLayout.NORTH, searchField, 5, SpringLayout.SOUTH, lblGenelist);
    sl_leftPanel.putConstraint(SpringLayout.NORTH, lblGenelist, 6, SpringLayout.SOUTH, leftSeparator);
    leftPanel.add(lblGenelist);
    leftPanel.add(searchField);
    searchField.setColumns(10);
    searchField.addKeyListener(this);
    this.addKeyListener(this);
    this.setFocusable(true);
    
    JMenuBar bar = new JMenuBar();
    bar.setOpaque(false);
    JMenu genelistmenu = new JMenu();
    genelistmenu.setOpaque(false);
    bar.add(genelistmenu);
    genelistmenu.setIcon(new ImageIcon(MainWindow.class.getResource("/javax/swing/plaf/metal/icons/sortDown.png")));
    sl_leftPanel.putConstraint(SpringLayout.NORTH, bar, 0, SpringLayout.NORTH, lblGenelist);
    sl_leftPanel.putConstraint(SpringLayout.EAST, bar, -5, SpringLayout.EAST, leftPanel);
    leftPanel.add(bar);
    
    genelistmenu.add(menuSetGeneDisplayedAs);
    menuSetGeneDisplayedAs.add(mntmSetGeneDisplayedAsSymbol);
    menuSetGeneDisplayedAs.add(mntmSetGeneDisplayedAsID);
    menuSetGeneDisplayedAs.add(mntmSetGeneDisplayedAsName);
    ButtonGroup group = new ButtonGroup();
    group.add(mntmSetGeneDisplayedAsSymbol);
    group.add(mntmSetGeneDisplayedAsName);
    group.add(mntmSetGeneDisplayedAsID);
    mntmSetGeneDisplayedAsSymbol.setSelected(true);
    mntmSetGeneDisplayedAsSymbol.addActionListener(this);
    mntmSetGeneDisplayedAsName.addActionListener(this);
    mntmSetGeneDisplayedAsID.addActionListener(this);
    
    genelistmenu.add(menuSortGenesBy);
    menuSortGenesBy.add(mntmSortGenesBySymbol);
    menuSortGenesBy.add(mntmSortGenesByID);
    menuSortGenesBy.add(mntmSortGenesByName);
    menuSortGenesBy.add(mntmSortGenesByChromosome);
    menuSortGenesBy.addSeparator();
    menuSortGenesBy.add(mntmSortAscending);
    menuSortGenesBy.add(mntmSortDescending);
    group = new ButtonGroup();
    group.add(mntmSortGenesBySymbol);
    group.add(mntmSortGenesByID);
    group.add(mntmSortGenesByName);
    group.add(mntmSortGenesByChromosome);
    mntmSortGenesBySymbol.setSelected(true);
    mntmSortGenesBySymbol.addActionListener(this);
    mntmSortGenesByID.addActionListener(this);
    mntmSortGenesByName.addActionListener(this);
    mntmSortGenesByChromosome.addActionListener(this);
    
    group = new ButtonGroup();
    group.add(mntmSortAscending);
    group.add(mntmSortDescending);
    mntmSortAscending.setSelected(true);
    mntmSortAscending.addActionListener(this);
    mntmSortDescending.addActionListener(this);
    
    JScrollPane scrlPnGenelist = new JScrollPane();
    sl_leftPanel.putConstraint(SpringLayout.WEST, scrlPnGenelist, 5, SpringLayout.WEST, leftPanel);
    sl_leftPanel.putConstraint(SpringLayout.EAST, scrlPnGenelist, -5, SpringLayout.EAST, leftPanel);
    sl_leftPanel.putConstraint(SpringLayout.EAST, searchField, -25, SpringLayout.EAST, scrlPnGenelist);
    sl_leftPanel.putConstraint(SpringLayout.NORTH, scrlPnGenelist, 5, SpringLayout.SOUTH, searchField);
    sl_leftPanel.putConstraint(SpringLayout.WEST, searchField, 0, SpringLayout.WEST, scrlPnGenelist);
    sl_leftPanel.putConstraint(SpringLayout.SOUTH, scrlPnGenelist, -10, SpringLayout.SOUTH, leftPanel);
    scrlPnGenelist.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
    leftPanel.add(scrlPnGenelist);

    geneList = new JList<GPVGene>();
    geneList.addListSelectionListener(this);
    scrlPnGenelist.setViewportView(geneList);

    JLabel lblLoadData = new JLabel("Load Data:");
    sl_leftPanel.putConstraint(SpringLayout.WEST, lblGenelist, 0, SpringLayout.WEST, lblLoadData);
    sl_leftPanel.putConstraint(SpringLayout.NORTH, btnLoadGenome, 5, SpringLayout.SOUTH, lblLoadData);
    sl_leftPanel.putConstraint(SpringLayout.NORTH, lblLoadData, 5, SpringLayout.NORTH, leftPanel);
    sl_leftPanel.putConstraint(SpringLayout.WEST, lblLoadData, 10, SpringLayout.WEST, leftPanel);
    leftPanel.add(lblLoadData);

    btnClearSearch.addMouseListener(this);
    sl_leftPanel.putConstraint(SpringLayout.NORTH, btnClearSearch, 0, SpringLayout.NORTH, searchField);
    btnClearSearch.setIcon(new ImageIcon(MainWindow.class.getResource("/at/icbi/gpviz/icons/kurumizawa/action_delete.png")));
    sl_leftPanel.putConstraint(SpringLayout.WEST, btnClearSearch, 5, SpringLayout.EAST, searchField);
    leftPanel.add(btnClearSearch);

    JSplitPane rightSplitPane = new JSplitPane();
    rightSplitPane.setResizeWeight(1d);
    rightSplitPane.setOneTouchExpandable(true);
    rightSplitPane.setOrientation(JSplitPane.VERTICAL_SPLIT);
    mainSplitPane.setRightComponent(rightSplitPane);

    JPanel imagePane = new JPanel();
    rightSplitPane.setLeftComponent(imagePane);
    imagePane.setLayout(new BorderLayout(0, 0));
    
    imageSplitPane.setResizeWeight(1d);
    imagePane.add(imageSplitPane, BorderLayout.CENTER);

    
    imageSplitPane.setLeftComponent(canvasScrollPane);
    canvasScrollPane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);
    canvasScrollPane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
    
    
        JLayeredPane canvasPanel = new JLayeredPane();
        canvasScrollPane.setViewportView(canvasPanel);
        
            canvas.addMouseMotionListener(this);
            canvas.addMouseListener(this);
            canvas.setFont(new Font("SansSerif", Font.PLAIN, 12));
            canvas.setForeground(Color.BLACK);
            canvas.setBackground(Color.WHITE);
            canvasDecoLayer = new GPVCanvasDecoLayer(canvas, this);
            canvas.addMouseListener(new CanvasPopUpMenu(canvas, canvasDecoLayer,this));
            canvas.addMouseWheelListener(this);
            
            canvasPanel.setLayout(new FullSizeLayout());
            
            canvasPanel.add(canvas, new Integer(1));
            canvasPanel.add(canvasDecoLayer,new Integer(2));
            
            canvas.getProperties().addCanvasPropertiesChangeListener(propsmenu);
    
    sliderProteinTrackHeight.setValue(canvas.getProperties().getProteinTrackHeightMultiplierAsPercent());
    sliderExonSpace.setValue(canvas.getProperties().getExonSpaceAsPercent());
    
        JToolBar imageToolBar = new JToolBar();
        imagePane.add(imageToolBar, BorderLayout.NORTH);
        imageToolBar.setFloatable(false);
        
            btnSaveImage.addActionListener(this);
            btnSaveImage.setToolTipText("Save Image...");
            btnSaveImage.setIcon(new ImageIcon(MainWindow.class.getResource("/com/sun/java/swing/plaf/windows/icons/FloppyDrive.gif")));
            imageToolBar.add(btnSaveImage);
                btnSetSize.setToolTipText("Set custom image size");
                btnSetSize.setIcon(new ImageIcon(MainWindow.class.getResource("/at/icbi/gpviz/icons/kurumizawa/search.png")));
                
                    btnSetSize.addActionListener(this);
                    
                    imageToolBar.add(horizontalStrut);
                    
                    JLabel lblTrackHeight = new JLabel(" Protein height ");
                    imageToolBar.add(lblTrackHeight);
                    
                    sliderProteinTrackHeight.setSnapToTicks(true);
                    sliderProteinTrackHeight.setPaintTicks(true);
                    sliderProteinTrackHeight.setMinorTickSpacing(5);
                    sliderProteinTrackHeight.setMajorTickSpacing(10);
                    imageToolBar.add(sliderProteinTrackHeight);
                    sliderProteinTrackHeight.addChangeListener(this);
                    
                    
                    JLabel lblExonSpace = new JLabel(" Exon space ");
                    imageToolBar.add(lblExonSpace);
                    

                    sliderExonSpace.setSnapToTicks(true);
                    sliderExonSpace.setPaintTicks(true);
                    sliderExonSpace.setMinorTickSpacing(5);
                    sliderExonSpace.setMajorTickSpacing(10);
                    sliderExonSpace.addChangeListener(this);
                    imageToolBar.add(sliderExonSpace);
                    
                    imageToolBar.add(btnSetSize);
                    btnShowSidebar.setSelected(true);
                    btnShowSidebar.setIcon(DisplayPropertiesPanel.EXPAND_ICON);
                    btnShowSidebar.addActionListener(this);
                    
                    imageToolBar.add(btnShowSidebar);
    btnExonsToScale.setToolTipText("Exons to scale. When this is selected exons will be of the correct scale to each other.");
    btnExonsToScale.setIcon(new ImageIcon(MainWindow.class.getResource("/at/icbi/gpviz/icons/scale.png")));
    btnExonsToScale.addActionListener(this);

    btnShowHelperLines.addActionListener(this);
    btnTrackLabels.setToolTipText("Show/hide track labels");
    btnTrackLabels.setSelected(true);
    
    
    btnTrackLabels.setIcon(new ImageIcon(MainWindow.class.getResource("/at/icbi/gpviz/icons/kurumizawa/comments.png")));
    btnTrackLabels.addActionListener(this);
    btnHighlight.setToolTipText("Show/hide region highlights");
    
    btnHighlight.addActionListener(this);
    btnHighlight.setBackground(Color.WHITE);
    btnHighlight.setIcon(new ImageIcon(MainWindow.class.getResource("/at/icbi/gpviz/icons/highlighter.png")));
    btnHighlight.setSelected(true);
    
    
    btnShowLegend.addActionListener(this);
    btnShowLegend.setSelected(true);
    btnShowLegend.setToolTipText("Show protein domain legend");
    btnShowLegend.setIcon(new ImageIcon(MainWindow.class.getResource("/at/icbi/gpviz/icons/legend.png")));
    btnTrackLabels.setSelectedIcon(null);
    
    
    btnShowHelperLines.setSelected(true);
    btnShowHelperLines.setIcon(new ImageIcon(MainWindow.class.getResource("/at/icbi/gpviz/icons/showlines.png")));
    btnShowHelperLines.setToolTipText("Show/hide helper lines");
    
    rightSplitPane.setRightComponent(statsTabbedPane);
    
    imageSplitPane.setRightComponent(sidebar);
    

    
    rightSplitPane.setDividerLocation(getHeight()-300);
    
    JPanel statusBar = new JPanel();
    getContentPane().add(statusBar, BorderLayout.SOUTH);
    statusBar.setLayout(new BorderLayout(2,2));
    
    memoryBar.setPreferredSize(new Dimension(185, 15));
    memoryBar.setMaximumSize(new Dimension(185, 15));
    memoryBar.setStringPainted(true);
    memoryBar.setVisible(false);
    
    menuBar.add(Box.createHorizontalGlue());
    menuBar.add(memoryBar);
    
    statusBar.add(progressBar,BorderLayout.CENTER);
    
    btnCancelAction = new JButton();
    btnCancelAction.setPreferredSize(new Dimension(30,20));
    btnCancelAction.setIcon(new ImageIcon(MainWindow.class.getResource("/at/icbi/gpviz/icons/kurumizawa/action_delete.png")));
    statusBar.add(btnCancelAction,BorderLayout.EAST);

    imageSplitPane.setDividerSize(0);
    imageSplitPane.setDividerLocation(getWidth()-382);
    
    progressBar.setMaximum(100);
    progressBar.setVisible(false);
    btnCancelAction.setVisible(false);
    
    asynchronousActionListener = new FocusTrappingCancelListener(btnCancelAction);
    btnCancelAction.addFocusListener(asynchronousActionListener);
    btnCancelAction.addKeyListener(asynchronousActionListener);
    btnCancelAction.addActionListener(asynchronousActionListener);
    
    this.setGlassPane(new BlockingGlassPane(btnCancelAction));
    
    this.addWindowStateListener(new WindowStateListener()
    {
      @Override
      public void windowStateChanged(WindowEvent e)
      {
        //Repaint after workspace switching on linux, or when minimized
        repaint();
      }
    });
    
    statsTabbedPane.refreshEverything(null);
    new CheckFreeMemoryThread(this).start();
  }
  
  public JScrollPane getCanvasScrollPanel(){
    return canvasScrollPane;
  }
  
  public void clearFilters(){
    geneFilterCache.clearFilters();
    if(fullGeneListModel!=null){
      setGeneListModel(fullGeneListModel);
    }
    mntmFilterDomainsEffected.setSelected(false);
    mntmFilterHighlight.setSelected(false);
    mntmFilterRVIntersect.setSelected(false);
    mntmFilterMinCodingTranscript.setSelected(false);
  }
  
  public void reapplyTextFilter(){
    if (fullGeneListModel != null)
    {
      GlobBasedGeneSymbolFilter filter = new GlobBasedGeneSymbolFilter(searchField.getText());
      SwingActions.filterGenes(this,filter);
    }
  }
  
  public void reapplyFilterCache(){
    GPVGene selected = canvas.getSelectedGene();
    boolean [] filter_result= geneFilterCache.getResultsFromCache();
    int accepted_nr = 0;
    for(int i = 0; i < filter_result.length; i++){
      if(filter_result[i])accepted_nr++;
    }
    DefaultListModel<GPVGene> filteredmodel = new DefaultListModel<GPVGene>();
    filteredmodel.setSize(accepted_nr);
    int j = 0;
    for(int i = 0; i < filter_result.length; i++){
      if(filter_result[i]){
        filteredmodel.set(j, fullGeneListModel.getElementAt(i));
        j++;
      }
    }
    setGeneListModel(filteredmodel);
    refreshEverything();
    selectGene(selected);
  }
  
  public void refreshCanvas()
  {
    GPVOptions opt = GPViz.getInstance().getOptions();
    if(opt!=null){
      if(opt.getPreferredColorScheme()!=null){
        canvas.getProperties().setColorScheme(opt.getPreferredColorScheme());
      }
    }
    canvas.refresh();
    canvasDecoLayer.clearStatus();
  }
  
  private List<GPVGene> getAllFilteredGenes(){
    ArrayList<GPVGene> ret = new ArrayList<GPVGene>();
    for(int i = 0; i < geneList.getModel().getSize(); i++){
      ret.add(geneList.getModel().getElementAt(i));
    }
    return ret;
  }
  
  public void refreshEverything(){
    refreshCanvas();
    statsTabbedPane.refreshEverything(canvas.getSelectedGene());
    sidebar.getDisplayPropertiesPanel().refreshTranscriptChooserPanel();
    sidebar.getDisplayPropertiesPanel().refreshRegionSetsPanel(this);
  }
  
  public void setGeneListModel(ListModel<GPVGene> model){
    GPVGene selected = geneList.getSelectedValue();
    geneList.setModel(model);
    selectGene(selected);
  }
  
  private void selectGene(GPVGene selected){
    if(selected==null){
      geneList.setSelectedIndex(0);
    }else{
      int toselect = 0;
      ListModel<GPVGene> model = geneList.getModel();
      for(int i = 0; i < model.getSize(); i++){
        if(model.getElementAt(i) == selected){
          toselect = i;
          break;
        }
      }
      geneList.setSelectedIndex(toselect);
    }
    geneList.ensureIndexIsVisible(geneList.getSelectedIndex());
  }
  
  public void refreshGeneListSorting(){
    GPVGene selected = geneList.getSelectedValue();
    geneList.setListData(GPViz.getInstance().getDataContainer().getSortedGeneVector(geneListSortOrder));
    selectGene(selected);
    fullGeneListModel = geneList.getModel();
  }
  
  public void setProgress(int progress){
    progressBar.setValue(progress);
    if(!ignoredLowMemory){
      long memory = CheckFreeMemoryThread.getFreeMemory();
      //less than XXmb left?
      if(memory < memory_warning_lower_threshold){
        Runtime.getRuntime().gc();
        memory = CheckFreeMemoryThread.getFreeMemory();
        if(memory < memory_warning_upper_threshold){
          //garbage collection barely freed anything. better warn the user
          boolean cancel = lowMemoryDialog.activate();
          if(cancel){
            asynchronousActionListener.cancel();
          }else{
            ignoredLowMemory = true;
          }
        }
      }
    }
  }
  
  private void setStatus(String status){
//    statuslabel.setText(status);
    canvasDecoLayer.setStatus(status);
    canvasDecoLayer.repaint();
  }
  private void clearStatus(){
    canvasDecoLayer.clearStatus();
    canvasDecoLayer.repaint();
  }
  
  protected void setMemoryUsage(int percentfull, String text){
    memoryBar.setValue(percentfull);
    memoryBar.setString(text);
    memoryBar.repaint();
  }
  
  private void setSidebarVisible(final boolean vis){
    
    //make a sliding animation
    new Thread(){
      public void run(){
        int steps = 10;
        int animation_time = 100;
        int steptime = animation_time/steps;
        int dest = imageSplitPane.getMaximumDividerLocation()-displayPanelWidth;
        
        if(!vis){ 
          dest = imageSplitPane.getMaximumDividerLocation();
        }
        
        int loc = imageSplitPane.getDividerLocation();
        int incr = (loc-dest)/steps;
        for(int i = 0; i< steps; i++){
          loc = loc-incr;
          imageSplitPane.setDividerLocation(loc);
//          imageSplitPane.revalidate();
          try{
            Thread.sleep(steptime);
          }catch (InterruptedException e){}
          imageSplitPane.setDividerLocation(loc);
        }
        imageSplitPane.setDividerLocation(dest);
      }
    }.start();
  }
  
  private int jobs_running = 0;
  public synchronized void startAsynchronousAction(final SwingWorker<?,?> interruptable,final String message){
    jobs_running++;
    /* This is kind of a queue, which allows one asynchrounous action to start another one
     * For example, when the region loading thread fires the datachangedlistener and that one
     * then re-applies the filters
     */
    new Thread(){
      public void run(){
        while(jobs_running>1){
          try{
            sleep(100);
          }catch (InterruptedException e){
            return;
          }
        }    
        progressBar.setVisible(true);
        progressBar.setStringPainted(true);
        getGlassPane().setVisible(true);
        btnCancelAction.setVisible(true);
        btnCancelAction.requestFocus();
        asynchronousActionListener.setWorker(interruptable);
        progressBar.setValue(0);
        progressBar.setString(message);
        interruptable.execute();
      }
    }.start();
  }
  
  public synchronized void finishedAsynchronousAction(){
    jobs_running--;
    if(jobs_running<=0){
      asynchronousActionListener.setWorker(null);
      progressBar.setString("");
      progressBar.setVisible(false);
      progressBar.setStringPainted(false);
      btnCancelAction.setVisible(false);
      this.getGlassPane().setVisible(false);
    }
  }
  
  public DisplayPropertiesPanel getDisplayPropertiesPanel(){
    return sidebar.getDisplayPropertiesPanel();
  }
  
  /*******************************
   * LISTENER METHODS
   ******************************/

  @Override
  public void actionPerformed(ActionEvent e)
  {
    if (e.getSource() == mntmExit)
      System.exit(0);
    if (e.getSource() == btnLoadGenome || e.getSource() == mntmLoadGenome)
    {
      SwingActions.loadGenomeFromFile(this);
    }else if(e.getSource()==btnLoadBEDRegions  || e.getSource() == mntmLoadRegions){
      SwingActions.loadRegionsFromFile(this);
    }else if(e.getSource()==btnLoadVariants  || e.getSource() == mntmLoadVariants){
      SwingActions.loadVariantsFromFile(this);
    }else if(e.getSource()==btnLoadDomains || e.getSource() == mntmLoadDomains){
      SwingActions.loadProteinDomainsFromFile(this);
    }else if (e.getSource() == btnSetSize) {
      if (btnSetSize.isSelected())
      {
        SwingActions.setImageDimensions(canvas.getProperties(), this, canvas.getSize());
        refreshCanvas();
      } else
      {
        canvas.getProperties().setUseCustomSize(false);
        refreshCanvas();
      }
    } else if (e.getSource() == btnShowHelperLines)
    {
      canvas.getProperties().setDrawHelperLines(btnShowHelperLines.isSelected());
      refreshCanvas();
    }else if (e.getSource() == btnTrackLabels){
      canvas.getProperties().setDrawGeneAndTranscriptLabels(btnTrackLabels.isSelected());
      refreshCanvas();
    }else if(e.getSource() == toolsOptionsMenuItem){
      ProgramOptionsDialog options = new ProgramOptionsDialog(this);
      options.setVisible(true);
      refreshEverything();
    }else if(e.getSource() == mntmFilterHighlight){
      if(mntmFilterHighlight.isSelected()){
        SwingActions.filterGenesAsynchronously(this,new HighlightedRegionsGeneFilter());
      }else{
        geneFilterCache.removeFilter(HighlightedRegionsGeneFilter.FILTER_NAME);
        reapplyFilterCache();
      }
      refreshCanvas();
    }else if(e.getSource() == mntmFilterRVIntersect){
        if(mntmFilterRVIntersect.isSelected()){
          SwingActions.filterGenesAsynchronously(this,new VariantsIntersectingRegionsGeneFilter());
        }else{
          geneFilterCache.removeFilter(VariantsIntersectingRegionsGeneFilter.FILTER_NAME);
          reapplyFilterCache();
        }
        refreshCanvas();
      
    }else if(e.getSource() == mntmFilterDomainsEffected){
      if(mntmFilterDomainsEffected.isSelected()){
        SwingActions.filterGenesAsynchronously(this, new DomainsAffectedGeneFilter());
      }else{
        geneFilterCache.removeFilter(DomainsAffectedGeneFilter.FILTER_NAME);
        reapplyFilterCache();
      }
      refreshCanvas();
    }else if(e.getSource()==mntmFilterMinCodingTranscript){
      if(mntmFilterMinCodingTranscript.isSelected()){
        SwingActions.filterGenesAsynchronously(this, new MinNumberOfCodingTranscriptsGeneFilter(2));
      }else{
        geneFilterCache.removeFilter(MinNumberOfCodingTranscriptsGeneFilter.createFilterName(2));
        reapplyFilterCache();
      }
      refreshCanvas();
    }else if(e.getSource() == btnSaveImage || e.getSource() == mntmSaveCurr){
      boolean shift = e.getModifiers()-InputEvent.BUTTON1_MASK == InputEvent.SHIFT_MASK;
      SwingActions.saveCurrentImage(this, shift);
    }else if(e.getSource() == mntmSaveSelected){
      BatchImageSaveDialog batch = new BatchImageSaveDialog(this, geneList.getSelectedValuesList(), canvas.getProperties());
      batch.setModal(true);
      batch.setVisible(true);
    }else if(e.getSource() == mntmSaveAll){
      BatchImageSaveDialog batch = new BatchImageSaveDialog(this, getAllFilteredGenes(), canvas.getProperties());
      batch.setModal(true);
      batch.setVisible(true);
    }else if(e.getSource() == btnHighlight){
      canvas.getProperties().setDrawHighlighting(btnHighlight.isSelected());
      refreshCanvas();
    }else if(e.getSource() == btnShowLegend){
      canvas.getProperties().setDrawDomainRegionLegend(btnShowLegend.isSelected());
      refreshCanvas();
    }else if(e.getSource() == mntmAbout){
      new AboutDialog(this).setVisible(true);
    }else if(e.getSource() == btnExonsToScale){
      canvas.getProperties().setExonsToScale(btnExonsToScale.isSelected());
      refreshCanvas();
    }else if(e.getSource() == btnShowSidebar){
      setSidebarVisible(btnShowSidebar.isSelected());
    }else if(e.getSource()== mntmClearFilter){
      clearFilters();
      refreshEverything();
    }else if(e.getSource() == mntmUnloadFiltered){
      int select = JOptionPane.showConfirmDialog(this, "This will unload all genes that are currently filtered. To undo this you will need to reload the original files. Proceed?",
          "Unload filtered genes?", JOptionPane.OK_CANCEL_OPTION);
      if(select==JOptionPane.OK_OPTION){
        SwingActions.unloadFilteredGenes(this);
      }
    }else if(e.getSource() == showMemory){
      memoryBar.setVisible(showMemory.isSelected());
    }else if (e.getSource() == mntmSetGeneDisplayedAsSymbol){
      GPViz.getInstance().getOptions().setGeneNameDisplayed(DISPLAYNAME.SYMBOL);
      geneList.repaint();
    }else if (e.getSource() == mntmSetGeneDisplayedAsID){
      GPViz.getInstance().getOptions().setGeneNameDisplayed(DISPLAYNAME.ID);
      geneList.repaint();
    }else if (e.getSource() == mntmSetGeneDisplayedAsName){
      GPViz.getInstance().getOptions().setGeneNameDisplayed(DISPLAYNAME.LONGNAME);
      geneList.repaint();
    }else if (e.getSource() == mntmSortGenesBySymbol){
      this.geneListSortOrder = new AlphabeticalGeneSorter(DISPLAYNAME.SYMBOL, mntmSortDescending.isSelected());
      refreshGeneListSorting();
      refreshAllFilters();
    }else if (e.getSource() == mntmSortGenesByID){
      this.geneListSortOrder = new AlphabeticalGeneSorter(DISPLAYNAME.ID, mntmSortDescending.isSelected());
      refreshGeneListSorting();
      refreshAllFilters();
    }else if (e.getSource() == mntmSortGenesByName){
      this.geneListSortOrder = new AlphabeticalGeneSorter(DISPLAYNAME.LONGNAME, mntmSortDescending.isSelected());
      refreshGeneListSorting();
      refreshAllFilters();
    }else if (e.getSource() == mntmSortGenesByChromosome){
      this.geneListSortOrder = new GenomicOrderGeneSorter(mntmSortDescending.isSelected());
      refreshGeneListSorting();
      refreshAllFilters();
    }else if(e.getSource() == mntmSortAscending || e.getSource() == mntmSortDescending){
      this.geneListSortOrder.setDescending(mntmSortDescending.isSelected());
      refreshGeneListSorting();
      refreshAllFilters();
    }
  }
  
  

  @Override
  public void mouseClicked(MouseEvent e)
  {
    if (e.getSource() == btnClearSearch)
    {
      searchField.setText("");
      reapplyTextFilter();
    }
  }

  @Override
  public void mousePressed(MouseEvent e)
  {
  }

  @Override
  public void mouseReleased(MouseEvent e)
  {
  }

  @Override
  public void mouseEntered(MouseEvent e)
  {
  }

  @Override
  public void mouseExited(MouseEvent e)
  {
    if(e.getSource() == canvas){
      clearStatus();
    }
  }

  @Override
  public void keyTyped(KeyEvent e)
  {
  }

  private int aimonk=0;
  private int[] aimonk_doce =new int[]{38,38,40,40,37,39,37,39,66,65,10};
  @Override
  public void keyPressed(KeyEvent e)
  {
    if(e.getSource()==this){
      if(aimonk<aimonk_doce.length){
        if(aimonk_doce[aimonk]==e.getKeyCode()){
          aimonk++;
        }else{
          aimonk=0;
        }
      }
      if(aimonk==aimonk_doce.length){
        GPViz.AIMONK=true; // yeah
        searchField.setBackground(Color.red);
        this.removeKeyListener(this);
      }
    }
  }

  @Override
  public void keyReleased(KeyEvent e)
  {
    if (e.getSource() == searchField)
    {
      reapplyTextFilter();
    }
  }

  @Override
  public void valueChanged(ListSelectionEvent e)
  {
    if (e.getSource() == geneList)
    {
      GPVGene selected_gene = geneList.getSelectedValue();
      sidebar.getDisplayPropertiesPanel().rebuildTranscriptChooserPanel(selected_gene);
      sidebar.getDisplayPropertiesPanel().revalidate();
      sidebar.getDisplayPropertiesPanel().repaint();
      canvas.setSelectedGene(selected_gene);
      
      refreshEverything();
    }
  }

  @Override
  public void mouseDragged(MouseEvent e)
  {
    
  }

  @Override
  public void mouseMoved(MouseEvent e)
  {
    if(e.getSource()==canvas){
      if(canvas.getSelectedGene()!=null){
        int nucleotide = canvas.getNucleotidePositionOn(e.getX(), e.getY());
        if(nucleotide!=0){
          setStatus("Hovering over nucleotide position: " +nucleotide);
        }else{
          int proteinPosition=canvas.getProteinPositionOn(e.getX(), e.getY());
          if(proteinPosition!=0){
            setStatus("Hovering over protein on position: " +proteinPosition);  
          }else{
            clearStatus();
          }
        }
      }
    }
  }

  @Override
  public void stateChanged(ChangeEvent e)
  {
    if(e.getSource() == sliderProteinTrackHeight){
      canvas.getProperties().setProteinTrackHeightMultiplierAsPercent(sliderProteinTrackHeight.getValue());
      refreshCanvas();
    }else if(e.getSource() == sliderExonSpace){
      canvas.getProperties().setExonSpaceAsPercent(sliderExonSpace.getValue());
      refreshCanvas();
    }
    
  }

  @Override
  public void dataChanged(DataChangedEvent event)
  {
    switch(event.getWhatChanged()){
      case CHROMOSOMES: break;
      case DOMAINS:
        refreshFilter(DomainsAffectedGeneFilter.FILTER_NAME);
        break;
      case EXONS:
        refreshGeneListSorting();
        clearFilters();
      case REGIONS:
        if(event.getChangeAction()==CHANGEACTION.LOADED || event.getChangeAction()==CHANGEACTION.PURGED){
          refreshFilter(HighlightedRegionsGeneFilter.FILTER_NAME);
          refreshFilter(VariantsIntersectingRegionsGeneFilter.FILTER_NAME);
          refreshFilter(DomainsAffectedGeneFilter.FILTER_NAME);
          sidebar.getDisplayPropertiesPanel().refreshRegionSetsPanel(this);
        }
    }
    refreshEverything();
  }
  
  private void refreshAllFilters(){
    for(GeneFilter filter:geneFilterCache.getFilters()){
      refreshFilter(filter.getFilterDescription());
    }
  }
  private void refreshFilter(String filtername){
    GeneFilter filter = geneFilterCache.getFilter(filtername);
    if(filter!=null) {
      SwingActions.filterGenesAsynchronously(this, filter);
    }
  }

  @Override
  public void mouseWheelMoved(MouseWheelEvent e)
  {
    if(e.isControlDown()){
      int units = e.getUnitsToScroll()*-1; //negative when scrolled down
      int custheight = canvas.getProperties().getCustomHeight();
      int custwidth = canvas.getProperties().getCustomWidth();

      if(custheight <= 0 )custheight = canvas.getHeight();
      if(custwidth <= 0 )custwidth = canvas.getWidth();

      float zoomfactor = 0.02f*units;
      float destfactor = 1f+zoomfactor;

      //dont do anything if we are at maximum zoom
      if(zoomfactor<0 || (custheight<=GPViz.MAX_DRAW_HEIGHT_IN_GUI && custwidth<=GPViz.MAX_DRAW_WIDTH_IN_GUI)){
        custheight = (int)(custheight*destfactor);
        custwidth = (int)(custwidth*destfactor);

        if(custheight > canvasScrollPane.getHeight() && custwidth > canvasScrollPane.getWidth()){
          canvas.getProperties().setCustomHeight(custheight);
          canvas.getProperties().setCustomWidth(custwidth);
          canvas.getProperties().setUseCustomSize(true);
          btnSetSize.setSelected(true);
          refreshCanvas();

          Point pos = canvasScrollPane.getViewport().getViewPosition();
          Point point = new Point(e.getPoint().x-pos.x,e.getPoint().y-pos.y);
          int newX = (int)(point.x*(zoomfactor) + destfactor*pos.x);
          int newY = (int)(point.y*(zoomfactor) + destfactor*pos.y);
          canvasScrollPane.getViewport().setViewPosition(new Point(newX, newY));

        }else{
          canvas.getProperties().setCustomHeight(0);
          canvas.getProperties().setCustomWidth(0);
          canvas.getProperties().setUseCustomSize(false);
          btnSetSize.setSelected(false);
          refreshCanvas();
        }
      }
    } else if (e.isShiftDown()) {
      // Horizontal scrolling
      Adjustable adj = canvasScrollPane.getHorizontalScrollBar();
      int scroll = e.getUnitsToScroll() * adj.getBlockIncrement();
      adj.setValue(adj.getValue() + scroll);
    } else {
      // Vertical scrolling
      Adjustable adj = canvasScrollPane.getVerticalScrollBar();
      int scroll = e.getUnitsToScroll() * adj.getBlockIncrement();
      adj.setValue(adj.getValue() + scroll);
    }

  }

}
