/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing;

import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.io.File;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.ListModel;
import javax.swing.SwingWorker;

import at.icbi.gpviz.GPViz;
import at.icbi.gpviz.data.GPVDataContainer;
import at.icbi.gpviz.data.GPVGene;
import at.icbi.gpviz.data.filter.GeneFilter;
import at.icbi.gpviz.data.filter.GeneFilterCache;
import at.icbi.gpviz.data.loader.BedFileRegionLoader;
import at.icbi.gpviz.data.loader.GTFExonLoader;
import at.icbi.gpviz.data.loader.domains.AbstractProteinDomainLoader;
import at.icbi.gpviz.data.loader.domains.ProteinDomainLoaderHelper;
import at.icbi.gpviz.data.loader.domains.parser.ProteinDomainEntry;
import at.icbi.gpviz.data.loader.variants.AbstractVariantLoader;
import at.icbi.gpviz.data.loader.variants.VariantLoaderHelper;
import at.icbi.gpviz.drawing.GPVImageWriter;
import at.icbi.gpviz.swing.canvas.CanvasProperties;
import at.icbi.gpviz.swing.canvas.GPVCanvas;
import at.tugraz.genome.biojava.io.ParsedEntry;

public class SwingActions
{

  public static void loadVariantsFromFile(final MainWindow parent){
    SwingWorker<Void,Void> worker = new SwingWorker<Void,Void>(){
      @Override
      protected Void doInBackground() throws Exception
      {
        File[] selectedFiles = chooseInputFiles(parent);
        if(selectedFiles!=null && selectedFiles.length>0){
          AbstractVariantLoader<? extends ParsedEntry> loader = VariantLoaderHelper.chooseFormat(selectedFiles[0]);
          if(loader==null){
            JOptionPane.showMessageDialog(parent, "Could not determine format for variants file. Please check the format.");
          }else{
            loader.loadVariants(parent, selectedFiles);
            parent.getDisplayPropertiesPanel().refreshRegionSetsPanel(parent);
          }
        }
        return null;
      }
      
      @Override
      protected void done() {
        parent.finishedAsynchronousAction();
      } 
    };
    parent.startAsynchronousAction(worker, "Loading variants from file...");
  }
  
  public static void loadRegionsFromFile(final MainWindow parent){
    SwingWorker<Void,Void> worker = new SwingWorker<Void,Void>(){
      @Override
      protected Void doInBackground() throws Exception
      {
        new BedFileRegionLoader().loadRegions(parent);
        parent.getDisplayPropertiesPanel().refreshRegionSetsPanel(parent);
        return null;
      }
      @Override
      protected void done() {
        parent.finishedAsynchronousAction();
      } 
    };
    parent.startAsynchronousAction(worker,"Loading regions from file...");
  }
  
  public static void loadGenomeFromFile(final MainWindow parent){
    SwingWorker<Void,Void> worker = new SwingWorker<Void,Void>(){
      @Override
      protected Void doInBackground() throws Exception
      {
        new GTFExonLoader().loadGenome(parent);
        return null;
      }
      @Override
      protected void done() {
        parent.finishedAsynchronousAction();
      } 
    };
    parent.startAsynchronousAction(worker,"Loading exons from file...");
  }
  
  public static void loadProteinDomainsFromFile(final MainWindow parent){
    SwingWorker<Void,Void> worker = new SwingWorker<Void,Void>(){
      @Override
      protected Void doInBackground() throws Exception
      {

        File selectedFile = chooseInputFile(parent);
        if(selectedFile!=null){
          AbstractProteinDomainLoader<? extends ProteinDomainEntry> loader = ProteinDomainLoaderHelper.chooseFormat(selectedFile);
          if(loader==null){
            JOptionPane.showMessageDialog(parent, "Could not determine format for protein domains file. Please check the format.");
          }else{
            loader.loadDomains(parent, selectedFile);
            parent.getDisplayPropertiesPanel().refreshDomainSetsPanel(parent);
          }
        }
        return null;
      }
      @Override
      protected void done() {
        parent.finishedAsynchronousAction();
      } 
    };
    parent.startAsynchronousAction(worker,"Loading protein domains from file...");
  }
  
  public static void setImageDimensions(CanvasProperties props, JFrame parent, Dimension currentSize){
    ImageSizeDialog dialog = new ImageSizeDialog((int)currentSize.getWidth(), (int)currentSize.getHeight(), parent);
    dialog.setLocationRelativeTo(parent);
    dialog.setModal(true);
    dialog.setVisible(true);
    int ret = dialog.getReturnVal();
    if(ret == JOptionPane.OK_OPTION){
      props.setCustomHeight(Integer.parseInt(dialog.getTextHeight().getText()));
      props.setCustomWidth(Integer.parseInt(dialog.getTextWidth().getText()));
      props.setUseCustomSize(true);
    }
  }
  
  public static void saveCurrentImage(MainWindow parent, boolean shift){
    GPVCanvas canvas = parent.getCanvas();
    if(shift){
      GPVImageWriter writer = new GPVImageWriter(canvas.getProperties(), canvas.getSize() ,parent);
      writer.writeGene(GPViz.getInstance().getDataContainer(), canvas.getSelectedGene());
    }else{
      BatchImageSaveDialog batch = new BatchImageSaveDialog(parent,canvas.getSelectedGene(), canvas.getProperties());
      batch.setModal(true);
      batch.setVisible(true);  
    }
  }

  public static void copyCurrentImageToClipboard(MainWindow parent){
    GPVCanvas canvas = parent.getCanvas();
    BufferedImage image = canvas.getBuffer();
    TransferableImage transferable = new TransferableImage( image );
    Toolkit.getDefaultToolkit().getSystemClipboard().setContents(transferable, null);
  }
  
  public static void filterGenesAsynchronously(final MainWindow parent, final GeneFilter filter){
    SwingWorker<Void,Void> worker = new SwingWorker<Void,Void>(){
      @Override
      protected Void doInBackground() throws Exception
      {
        filterGenes(parent,filter,true);
        return null;
      }
      @Override
      protected void done() {
        parent.finishedAsynchronousAction();
      } 
    };
    parent.startAsynchronousAction(worker, "Applying filter to "+parent.fullGeneListModel.getSize()+" genes...");
  }

  public static void filterGenes(final MainWindow parent, final GeneFilter filter){
    filterGenes(parent,filter,false);
  }
  
  private static void filterGenes(final MainWindow parent, final GeneFilter filter, boolean fill_progressbar){
    ListModel<GPVGene> model = parent.fullGeneListModel;
    if (model != null)
    {
      boolean[] result = new boolean[model.getSize()]; 
      for (int i = 0; i < model.getSize(); i++)
      {
        if(Thread.currentThread().isInterrupted())return;
        parent.setProgress((int)(100f/model.getSize()*i));
        boolean accept = filter.accept(model.getElementAt(i));
        result[i] = accept;
      }
      parent.geneFilterCache.addFilter(filter, result);
      parent.reapplyFilterCache();
    }
  }
  
  public static void unloadFilteredGenes(final MainWindow parent){
    GPVDataContainer data = GPViz.getInstance().getDataContainer();
    GeneFilterCache cache = parent.geneFilterCache;
    boolean[] results = cache.getResultsFromCache();
    for(int i = 0; i < results.length; i++){
      if(!results[i]){
        GPVGene todelete = parent.fullGeneListModel.getElementAt(i);
        data.getGenes().remove(todelete.getGeneID());
      }
    }
    //free some memory
    Runtime.getRuntime().gc();
    //sneaky way to fire the data change listeners
    data.setGenes(data.getGenes());
    parent.clearFilters();
    parent.refreshEverything();
  }
  
  public static File chooseInputFile(MainWindow parent){
    String lastinputdir = GPViz.getInstance().getOptions().getLastInputDir();
    if(lastinputdir==null)lastinputdir = System.getProperty("user.home");
    
    JFileChooser chooser = new JFileChooser(lastinputdir);
    chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
    int result = chooser.showDialog(parent, "Open");
    if(result!=JFileChooser.APPROVE_OPTION) return null;
    return chooser.getSelectedFile();
  }
  public static File[] chooseInputFiles(MainWindow parent){
    String lastinputdir = GPViz.getInstance().getOptions().getLastInputDir();
    if(lastinputdir==null)lastinputdir = System.getProperty("user.home");
    
    JFileChooser chooser = new JFileChooser(lastinputdir);
    chooser.setMultiSelectionEnabled(true);
    chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
    int result = chooser.showDialog(parent, "Open");
    if(result!=JFileChooser.APPROVE_OPTION) return null;
    return chooser.getSelectedFiles();
  }
}
