/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing.canvas;

import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JSeparator;

import at.icbi.gpviz.GPViz;
import at.icbi.gpviz.data.GPVExon;
import at.icbi.gpviz.data.GPVTranscript;
import at.icbi.gpviz.data.regions.GPVHighlightedRegion;
import at.icbi.gpviz.swing.MainWindow;
import at.icbi.gpviz.swing.SwingActions;
import at.icbi.gpviz.weblinks.WebLink;


public class CanvasPopUpMenu extends JPopupMenu implements MouseListener, ActionListener
{
  private static final long serialVersionUID = 5270481311576401429L;
  private JMenuItem savePicture = new JMenuItem("Save current image...");
  private JMenuItem clipboardImage = new JMenuItem("Copy current image to clipboard");
  private JMenuItem createRegion = new JMenuItem("Highlight this region");
  private JMenuItem selectExon = new JMenuItem("Highlight this exon");
  private JMenuItem removeRegion = new JMenuItem("Remove this highlighting");
  private JMenuItem insertPointHighlighter = new JMenuItem("Insert SNP");
  private JMenuItem hideTranscript = new JMenuItem("Hide this transcript");
  private JMenuItem unhideTranscripts = new JMenuItem("Unhide all transcripts");
  
  
  private JMenuItem openGeneInGenomeBrowser = new JMenuItem("Open Gene in Genome Browser");
  
  private GPVCanvas canvas;
  private GPVCanvasDecoLayer deco;
  private MainWindow mainwindow;
  
  private Point lastClickedAt;
  
  
  public CanvasPopUpMenu(GPVCanvas canvas, GPVCanvasDecoLayer deco, MainWindow mainwindow){
    super();
    
    this.canvas = canvas;
    this.deco = deco;
    this.mainwindow = mainwindow;
    
    savePicture.addActionListener(this);
    clipboardImage.addActionListener(this);
    createRegion.addActionListener(this);
    selectExon.addActionListener(this);
    insertPointHighlighter.addActionListener(this);
    removeRegion.addActionListener(this);
    

    hideTranscript.addActionListener(this);
    unhideTranscripts.addActionListener(this);
    
    openGeneInGenomeBrowser.addActionListener(this);

    //set defaults
    
    //add menu items
    add(savePicture);
    add(clipboardImage);
    add(new JSeparator());
    add(createRegion);
    add(selectExon);
    add(insertPointHighlighter);
    add(removeRegion);
    add(new JSeparator());
    add(hideTranscript);
    add(unhideTranscripts);
    add(new JSeparator());
    add(new CanvasPropertiesMenu(canvas));
    add(new JSeparator());
    add(openGeneInGenomeBrowser);
    
  }
  
  private void showMe(Point loc){
    createRegion.setEnabled(deco.lassoRectangle!=null && deco.lassoRectangle.contains(loc));
    savePicture.setEnabled(canvas.selectedGene!=null);
    clipboardImage.setEnabled(canvas.selectedGene!=null);
    
    //false until proven true
    unhideTranscripts.setEnabled(false);
    hideTranscript.setEnabled(false);
    selectExon.setEnabled(false);
    insertPointHighlighter.setEnabled(false);
    removeRegion.setEnabled(false);
    openGeneInGenomeBrowser.setEnabled(false);
    
    if(canvas.map!=null && canvas.selectedGene!=null){
      if(canvas.hasTranscriptSelected){ 
        removeRegion.setEnabled(canvas.getHighlightedRegionAt((int)loc.getX(), (int)loc.getY())!=null);
        GPVExon exon = canvas .map.getExonAt((int)loc.getX(), (int)loc.getY());
        selectExon.setEnabled(exon!=null);
        int transcript = canvas.map.getTranscriptThatPixelIsIn((int)loc.getX(), (int)loc.getY());
        boolean inregionblock =  canvas.map.getContinuousRegionHighlighterBlock().contains(loc);
        boolean invariantblock =  canvas.map.getPointRegionHighlighterBlock().contains(loc);
        insertPointHighlighter.setEnabled(deco.lassoRectangle==null && (transcript>=0 ||invariantblock || inregionblock));
        hideTranscript.setEnabled(transcript>0);
      }
      openGeneInGenomeBrowser.setEnabled(true);
      unhideTranscripts.setEnabled(true);
    }
    this.show(canvas, (int)loc.getX(), (int)loc.getY());
  }
  
  
  @Override
  public void mouseClicked(MouseEvent e)
  {
    if(e.getButton()==MouseEvent.BUTTON3){
      lastClickedAt=e.getPoint();
      showMe(e.getPoint());
    }
  }


  @Override
  public void mousePressed(MouseEvent e)
  {
    
  }


  @Override
  public void mouseReleased(MouseEvent e)
  {
    
  }


  @Override
  public void mouseEntered(MouseEvent e)
  {
    
  }


  @Override
  public void mouseExited(MouseEvent e)
  {
    
  }

  @Override
  public void actionPerformed(ActionEvent e)
  {
    if(e.getSource()==savePicture){
      boolean shift = e.getModifiers()-InputEvent.BUTTON1_MASK == InputEvent.SHIFT_MASK;
      SwingActions.saveCurrentImage(mainwindow, shift);
    }else if(e.getSource()==clipboardImage){
      SwingActions.copyCurrentImageToClipboard(mainwindow);
    }else if(e.getSource()==createRegion){
      deco.transformLassoRegionIntoHighlightedRegion();
    }else if(e.getSource()==selectExon){
      deco.selectExonIfHoveringOverOne(lastClickedAt);
      deco.transformLassoRegionIntoHighlightedRegion();
      deco.repaint();
    }else if(e.getSource()==removeRegion){
      GPVHighlightedRegion region_to_remove = canvas.getHighlightedRegionAt((int)lastClickedAt.getX(), (int)lastClickedAt.getY());    
      GPViz.getInstance().getDataContainer().removeHighlightedRegionThatContains(region_to_remove);
      mainwindow.getDisplayPropertiesPanel().refreshRegionSetsPanel(mainwindow);
    }else if(e.getSource() == insertPointHighlighter){
      deco.selectAsPointHighlighter(lastClickedAt);
      deco.transformLassoRegionIntoHighlightedRegion();
      deco.repaint();
    }else if(e.getSource() == hideTranscript){
      int transcript_nr = canvas.map.getTranscriptThatPixelIsIn((int)lastClickedAt.getX(), (int)lastClickedAt.getY());
      GPVTranscript transcript = canvas.selectedGene.getTranscriptFromEnabledTranscripts(transcript_nr);
      if(transcript!=null)transcript.setDisplayed(false);
      canvas.refresh();
      mainwindow.getDisplayPropertiesPanel().refreshTranscriptChooserPanel();
    }else if(e.getSource() == unhideTranscripts){
      for(GPVTranscript t:canvas.selectedGene.getTranscripts().values())t.setDisplayed(true);
      canvas.refresh();
      mainwindow.getDisplayPropertiesPanel().refreshTranscriptChooserPanel();
    }else if(e.getSource() == openGeneInGenomeBrowser){
      String baselink = GPViz.getInstance().getOptions().getRegionLinkBase();
      String link = WebLink.resolveRegionLink(baselink, canvas.getSelectedGene());
      WebLink.visitLink(link);
    }
    
    
  }
  
}
