/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing.canvas;

import java.util.ArrayList;

import at.icbi.gpviz.GPViz;
import at.icbi.gpviz.data.GPVOptions;
import at.icbi.gpviz.drawing.colorscheme.DefaultGPVColorScheme;
import at.icbi.gpviz.drawing.colorscheme.GPVColorScheme;
import at.icbi.gpviz.drawing.variantshapes.DefaultGPVVariantShapeScheme;
import at.icbi.gpviz.drawing.variantshapes.GPVVariantShapeScheme;

public class CanvasProperties
{
  public static int REGION_HIGHLIGHT_MODE_ALL = 0;
  public static int REGION_HIGHLIGHT_MODE_DOMAIN = 1;
  
  public static final float MINIMUM_EXON_SPACE = 0.01f;
  public static final float MAXIMUM_EXON_SPACE = 1f;
  public static final float MINIMUM_PROT_TRACK_HEIGHT_M = 0.05f;
  public static final float MAXIMUM_PROT_TRACK_HEIGHT_M = 0.80f;

  
  private GPVColorScheme colorScheme = new DefaultGPVColorScheme();
  private GPVVariantShapeScheme preferredVariantShapeScheme = new DefaultGPVVariantShapeScheme();
  
  private boolean drawHighlighting = true;
  private boolean drawHelperLines = true;
  private boolean drawStartStopCodonLines = false;
  private int customWidth = 0;
  private int customHeight = 0;
  private boolean useCustomSize = false;
  
  private float imageWidthPercent = 0.95f;
  private float headerHeightPercent =0.05f;
  private float highlighterBarHeightPercent = 0.07f; 
  private float axisHeightPercent =0.04f;
  private float legendHeightPercent =0.13f;
  private float imageTotalGeneAndProteinSpaceHightPercent = 0.68f;
  
  private float maximumTrackHeight = 0.10f;
  private float exonSpaceSharingCutoff = 0.2f;
  private float strokeWithRatio = 0.0006f;
  private float pointHighlighterTriangleWidth = 0.003f;
  
  private int regionHighlightMode = REGION_HIGHLIGHT_MODE_ALL;
  private boolean drawGeneAndTranscriptLabels = true;
  private float percentTrackGapRelativeToTrack = 0.05f;
  private float minimumTrackGapRelativeToImageHeight = 0.006f;
  private float exonSpace = 0.7f;
  private boolean useDynamicDomainFontSize = false;
  private float minimumDomainNameVisibleFraction = 1f;//0.75f;
  private boolean showProteinXAxis = true;
  private float proteinXAxisPercentOfProteinBlock = 0.15f;
  
  private boolean drawDomainRegionLegend = true;
  private boolean drawFootnotesLegend = true;
  private boolean drawVariantTypesLegend = true;
  
  
  private float untranslatedTranscriptRegionNarrowing = 0.4f;
  
  private float geneProteinTrackRatio = 0.3f;
  
  private boolean regionSelectingModeEnabled = true;
  
  private ArrayList<CanvasPropertiesChangedListener> changeListeners = new ArrayList<CanvasPropertiesChangedListener>();
  
  public CanvasProperties(){
    GPVOptions opt = GPViz.getInstance().getOptions();
    if(opt!=null){
      if(opt.getPreferredColorScheme()!=null){
        this.colorScheme = opt.getPreferredColorScheme();
      }
    }
  }
  
  public boolean isDrawHelperLines()
  {
    return drawHelperLines;
  }
  public void setDrawHelperLines(boolean drawHelperLines)
  {
    this.drawHelperLines = drawHelperLines;
    fireChangeListeners();
  }
  public int getCustomWidth()
  {
    return customWidth;
  }
  public void setCustomWidth(int customWidth)
  {
    this.customWidth = customWidth;
    fireChangeListeners();
  }
  public int getCustomHeight()
  {
    return customHeight;
  }
  public void setCustomHeight(int customHeight)
  {
    this.customHeight = customHeight;
    fireChangeListeners();
  }
  public boolean isUseCustomSize()
  {
    return useCustomSize;
  }
  public void setUseCustomSize(boolean useCustomHeight)
  {
    this.useCustomSize = useCustomHeight;
    fireChangeListeners();
  }
  public GPVColorScheme getColorScheme()
  {
    return colorScheme;
  }
  public void setColorScheme(GPVColorScheme colorScheme)
  {
    this.colorScheme = colorScheme;
    fireChangeListeners();
  }
  public float getExonSpace()
  {
    return exonSpace;
  }
  public void setExonSpace(float exonSpace)
  {
    this.exonSpace = exonSpace;
    fireChangeListeners();
  }
  public void setExonSpaceAsPercent(int percent){
    float len = MAXIMUM_EXON_SPACE-MINIMUM_EXON_SPACE;
    this.exonSpace = MINIMUM_EXON_SPACE + (percent/100f)*len;
  }
  public int getExonSpaceAsPercent(){
    float len = MAXIMUM_EXON_SPACE-MINIMUM_EXON_SPACE;
    return (int)(((this.exonSpace-MINIMUM_EXON_SPACE)/len)*100);
  }
  public void setProteinTrackHeightMultiplierAsPercent(int percent){
    float len = MAXIMUM_PROT_TRACK_HEIGHT_M-MINIMUM_PROT_TRACK_HEIGHT_M;
    this.geneProteinTrackRatio = MINIMUM_PROT_TRACK_HEIGHT_M + (percent/100f)*len;
  }
  public int getProteinTrackHeightMultiplierAsPercent(){
    float len = MAXIMUM_PROT_TRACK_HEIGHT_M-MINIMUM_PROT_TRACK_HEIGHT_M;
    return (int)(((this.geneProteinTrackRatio-MINIMUM_PROT_TRACK_HEIGHT_M)/len)*100);
  }
  public float getImageWidthPercent()
  {
    return imageWidthPercent;
  }
  public void setImageWidthPercent(float imageWidthPercent)
  {
    this.imageWidthPercent = imageWidthPercent;
    fireChangeListeners();
  }
  public float getImageTotalGeneAndProteinSpaceHightPercent()
  {
    float ret = imageTotalGeneAndProteinSpaceHightPercent;
    if(!isDrawAnyLegend()) ret+=legendHeightPercent;
    if(!isDrawHighlighting()) ret+=highlighterBarHeightPercent;
    return ret;
  }
  public void setImageTotalGeneAndProteinSpaceHightPercent(float imageTotalGeneAndProteinSpaceHightPercent)
  {
    this.imageTotalGeneAndProteinSpaceHightPercent = imageTotalGeneAndProteinSpaceHightPercent;
    fireChangeListeners();
  }
  public float getMaximumTrackHeight()
  {
    return maximumTrackHeight;
  }
  public void setMaximumTrackHeight(float maximumTrackHeight)
  {
    this.maximumTrackHeight = maximumTrackHeight;
    fireChangeListeners();
  }
  public float getExonSpaceSharingCutoff()
  {
    return exonSpaceSharingCutoff;
  }
  public void setExonSpaceSharingCutoff(float exonSpaceSharingCutoff)
  {
    this.exonSpaceSharingCutoff = exonSpaceSharingCutoff;
    fireChangeListeners();
  }
  public float getPercentTrackGapRelativeToTrack()
  {
    return percentTrackGapRelativeToTrack;
  }
  public void setPercentTrackGapRelativeToTrack(float percentTrackGap)
  {
    this.percentTrackGapRelativeToTrack = percentTrackGap;
    fireChangeListeners();
  }
  public float getMinimumTrackGapRelativeToImageHeight()
  {
    return minimumTrackGapRelativeToImageHeight;
  }
  public void setMinimumTrackGapRelativeToImageHeight(float minimumTrackGapRelativeToImageHeight)
  {
    this.minimumTrackGapRelativeToImageHeight = minimumTrackGapRelativeToImageHeight;
    fireChangeListeners();
  }
  public boolean isDrawGeneAndTranscriptLabels()
  {
    return drawGeneAndTranscriptLabels;
  }
  public void setDrawGeneAndTranscriptLabels(boolean drawGeneAndTranscriptLabels)
  {
    this.drawGeneAndTranscriptLabels = drawGeneAndTranscriptLabels;
    fireChangeListeners();
  }
  public float getHeaderHeightPercent()
  {
    return headerHeightPercent;
  }
  public void setHeaderHeightPercent(float headerHeightPercent)
  {
    this.headerHeightPercent = headerHeightPercent;
    fireChangeListeners();
  }
  public int getRegionHighlightMode()
  {
    return regionHighlightMode;
  }
  public void setRegionHighlightMode(int regionHighlightMode)
  {
    this.regionHighlightMode = regionHighlightMode;
    fireChangeListeners();
  }
  public boolean isDrawHighlighting()
  {
    return drawHighlighting;
  }
  public void setDrawHighlighting(boolean drawHighlighting)
  {
    this.drawHighlighting = drawHighlighting;
    fireChangeListeners();
  }
  public float getStrokeWithRatio()
  {
    return strokeWithRatio;
  }
  public void setStrokeWithRatio(float strokeWithRatio)
  {
    this.strokeWithRatio = strokeWithRatio;
    fireChangeListeners();
  }
  public boolean isDrawDomainRegionLegend()
  {
    return drawDomainRegionLegend;
  }
  public void setDrawDomainRegionLegend(boolean drawDomainRegionLegend)
  {
    this.drawDomainRegionLegend = drawDomainRegionLegend;
    fireChangeListeners();
  }
  public float getLegendHeightPercent()
  {
    float ret = legendHeightPercent;
    if(!isDrawAnyLegend()) ret = 0f;
    return ret;
  }
  public void setLegendHeightPercent(float legendHeightPercent)
  {
    this.legendHeightPercent = legendHeightPercent;
    fireChangeListeners();
  }
  public float getGeneProteinTrackRatio()
  {
    return geneProteinTrackRatio;
  }
  public void setGeneProteinTrackRatio(float geneProteinTrackRatio)
  {
    this.geneProteinTrackRatio = geneProteinTrackRatio;
    fireChangeListeners();
  }
  public boolean isUseDynamicDomainFontSize()
  {
    return useDynamicDomainFontSize;
  }
  public void setUseDynamicDomainFontSize(boolean useDynamicDomainFontSize)
  {
    this.useDynamicDomainFontSize = useDynamicDomainFontSize;
    fireChangeListeners();
  }
  public float getMinimumDomainNameVisibleFraction()
  {
    return minimumDomainNameVisibleFraction;
  }
  public void setMinimumDomainNameVisibleFraction(float minimumDomainNameVisibleFraction)
  {
    this.minimumDomainNameVisibleFraction = minimumDomainNameVisibleFraction;
    fireChangeListeners();
  }
  public float getAxisHeightPercent()
  {
    return axisHeightPercent;
  }
  public void setAxisHeightPercent(float axisHeightPercent)
  {
    this.axisHeightPercent = axisHeightPercent;
    fireChangeListeners();
  }
  public boolean isRegionSelectingModeEnabled()
  {
    return regionSelectingModeEnabled;
  }
  public void setRegionSelectingModeEnabled(boolean regionSelectingModeEnabled)
  {
    this.regionSelectingModeEnabled = regionSelectingModeEnabled;
    fireChangeListeners();
  }
  public float getHighlighterBarHeightPercent()
  {
    float ret = highlighterBarHeightPercent;
    if(!drawHighlighting) ret = 0f;;
    return ret;
  }
  public void setHighlighterBarHeightPercent(float highlighterBarHeightPercent)
  {
    this.highlighterBarHeightPercent = highlighterBarHeightPercent;
    fireChangeListeners();
  }
  public float getPointHighlighterTriangleWidth()
  {
    return pointHighlighterTriangleWidth;
  }
  public void setPointHighlighterTriangleWidth(float pointHighlighterTriangleWidth)
  {
    this.pointHighlighterTriangleWidth = pointHighlighterTriangleWidth;
    fireChangeListeners();
  }
  public boolean isShowProteinXAxis()
  {
    return showProteinXAxis;
  }
  public void setShowProteinXAxis(boolean showProteinXAxis)
  {
    this.showProteinXAxis = showProteinXAxis;
    fireChangeListeners();
  }
  public float getProteinXAxisPercentOfProteinBlock()
  {
    return proteinXAxisPercentOfProteinBlock;
  }
  public void setProteinXAxisPercentOfProteinBlock(float proteinXAxisPercentOfProteinBlock)
  {
    this.proteinXAxisPercentOfProteinBlock = proteinXAxisPercentOfProteinBlock;
    fireChangeListeners();
  }
  
  public float getUntranslatedTranscriptRegionNarrowing()
  {
    return untranslatedTranscriptRegionNarrowing;
  }

  public void setUntranslatedTranscriptRegionNarrowing(float untranslatedTranscriptRegionNarrowing)
  {
    this.untranslatedTranscriptRegionNarrowing = untranslatedTranscriptRegionNarrowing;
    fireChangeListeners();
  }
  
  public boolean isDrawStartStopCodonLines()
  {
    return drawStartStopCodonLines;
  }

  public void setDrawStartStopCodonLines(boolean drawStartStopCodonLines)
  {
    this.drawStartStopCodonLines = drawStartStopCodonLines;
    fireChangeListeners();
  }
  
  public boolean isDrawFootnotesLegend()
  {
    return drawFootnotesLegend;
  }

  public void setDrawFootnotesLegend(boolean drawFootnotesLegend)
  {
    this.drawFootnotesLegend = drawFootnotesLegend;
    fireChangeListeners();
  }

  public boolean isDrawVariantTypesLegend()
  {
    return drawVariantTypesLegend;
  }

  public void setDrawVariantTypesLegend(boolean drawVariantTypesLegend)
  {
    this.drawVariantTypesLegend = drawVariantTypesLegend;
    fireChangeListeners();
  }
  
  public GPVVariantShapeScheme getPreferredVariantShapeScheme()
  {
    return preferredVariantShapeScheme;
  }

  public void setPreferredVariantShapeScheme(GPVVariantShapeScheme variantShapeScheme)
  {
    this.preferredVariantShapeScheme = variantShapeScheme;
    fireChangeListeners();
  }

  public void setExonsToScale(boolean toscale){
    if(toscale){
      setExonSpaceSharingCutoff(1f);
    }else{
      setExonSpaceSharingCutoff(0.2f);
    }
    fireChangeListeners();
  }
  public boolean isExonsSetToScale(){
    return getExonSpaceSharingCutoff()==1f;
  }
  
  public boolean isDrawAnyLegend(){
    return isDrawDomainRegionLegend() || isDrawFootnotesLegend() || isDrawVariantTypesLegend();
  }
  
  public void setDrawAnyLegend(boolean draw){
    drawDomainRegionLegend=draw;
    drawFootnotesLegend=draw;
    drawVariantTypesLegend=draw;
    fireChangeListeners();
  }
  
  public void removeAllChangeListeners(){
    this.changeListeners = new ArrayList<CanvasPropertiesChangedListener>();
  }
  
  public void addCanvasPropertiesChangeListener(CanvasPropertiesChangedListener listener){
    this.changeListeners.add(listener);
  }
  
  protected void fireChangeListeners(){
    for(CanvasPropertiesChangedListener listener:this.changeListeners)listener.canvasPropertiesChanged();
  }
  
}
