/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing.canvas;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;
import java.util.TreeSet;

import javax.swing.JPanel;


import at.icbi.gpviz.data.GPVDataContainer;
import at.icbi.gpviz.data.GPVDomain;
import at.icbi.gpviz.data.GPVDomainSet;
import at.icbi.gpviz.data.GPVExon;
import at.icbi.gpviz.data.GPVGene;
import at.icbi.gpviz.data.GPVProtein;
import at.icbi.gpviz.data.GPVTranscript;
import at.icbi.gpviz.data.regions.GPVHighlightedRegion;
import at.icbi.gpviz.data.regions.GPVVariant;
import at.icbi.gpviz.data.util.GPVUtils;
import at.icbi.gpviz.drawing.colorscheme.ColorProvider;
import at.icbi.gpviz.drawing.colorscheme.GPVColorScheme;
import at.icbi.gpviz.drawing.variantshapes.DynamicGPVVariantShapeScheme;
import at.icbi.gpviz.drawing.variantshapes.GPVVariantShape;
import at.icbi.gpviz.drawing.variantshapes.GPVVariantShapeScheme;
import at.icbi.gpviz.exceptions.NeedMoreMemoryException;
import at.icbi.gpviz.swing.CheckFreeMemoryThread;
import at.tugraz.genome.biojava.seq.DNAStrand;

public class GPVCanvas extends JPanel
{
  private static final long serialVersionUID = 5342052565313366750L;
  
  protected CanvasProperties properties = new CanvasProperties();
  protected GPVColorScheme colors;
  protected GPVVariantShapeScheme variantShapes;

  protected GPVGene selectedGene;

  protected GPVCanvasMap map;

  protected BasicStroke defaultStroke;
  
  protected List<GPVCanvasListener> canvasListeners;

  protected boolean hasStartStopCodon = true;
  protected boolean hasAtLeastOneCodingTranscript = true;
  protected boolean hasTranscriptSelected = true;
  protected boolean hasDomainSets = true;
  protected boolean hasVariants = true;
  protected boolean hasSomethingInTranscriptionRange = true;

  protected ArrayList<String> errors = new ArrayList<String>();
  
  protected Footnotes footnotes;
  protected Integer lastFootnote = 0;

  private HashMap<String,Color> domainSetColors;
  
  private BufferedImage buffer = null;
  
  private boolean bufferInvalidated = true;
  
  private GPVDataContainer data;
  
  public GPVCanvas(GPVDataContainer data)
  {
    super();
    this.data = data;
    canvasListeners = new ArrayList<GPVCanvasListener>();
    setPreferredSize(null);
  }
  
  public void addCanvasListener(GPVCanvasListener listener){
    this.canvasListeners.add(listener);
  }

  public void refresh()
  {
    
    if (properties.isUseCustomSize())
    {
      int height = properties.getCustomHeight();
      int width = properties.getCustomWidth();
      setPreferredSize(new Dimension(width, height));
    } else
    {
      setPreferredSize(null);
    }
    bufferInvalidated = true;
    revalidate();
    repaint();
  }

  public void paintToBuffer() throws NeedMoreMemoryException{
    if(getWidth()<=0 || getHeight()<=0)return;

    long free = CheckFreeMemoryThread.getFreeMemory();
    long needed = (long)((getWidth()*getHeight()*4l)*1.1d);
    if(needed > free){
      throw new NeedMoreMemoryException(free,needed);
    }

    this.buffer = new BufferedImage(getWidth(), getHeight(), BufferedImage.TYPE_INT_RGB);
    bufferInvalidated = false;

    paintNotFromBuffer(this.buffer.getGraphics());
  }

  
  public void paintNotFromBuffer(Graphics g)
  {
    Graphics2D g2 =  (Graphics2D) g;
    
    this.errors = new ArrayList<String>();
    this.footnotes = new Footnotes();
    this.lastFootnote = 0;
    
    int height = getSize().height;
    int width = getSize().width;

    double diag = Math.sqrt(Math.pow(height, 2) + Math.pow(width, 2));
    float stroke_with = (float) (properties.getStrokeWithRatio() * diag);

    this.defaultStroke = new BasicStroke(stroke_with);
    g2.setStroke(this.defaultStroke);

    this.colors = properties.getColorScheme();
    if(data.getUnknownVariantTypes().size()>0){
      //there are unknown variant types (maybe from mut or VCF files)
      //i can't use the preferred shape scheme, i have to use the dynamic one
      this.variantShapes = new DynamicGPVVariantShapeScheme(data.getUnknownVariantTypes(),properties.getPreferredVariantShapeScheme());
    }else{
      this.variantShapes = properties.getPreferredVariantShapeScheme();
    }
    
    g2.setColor(colors.getCanvasBackground());
    g2.fillRect(0, 0, width, height);

    // Enable antialiasing for shapes
    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    // Enable antialiasing for text
    g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);

    if (selectedGene != null)
    {
      validateAll();
      map = new GPVCanvasMap(selectedGene, properties, height, width, selectedGene.getStrand() == DNAStrand.REVERSE, hasStartStopCodon
          && hasSomethingInTranscriptionRange, needsAnyLegend());
      map.setVariantShapes(this.variantShapes);
      if (hasTranscriptSelected)
      {
        
        drawTitle(g2);
        drawGene(g2);
        drawTranscripts(g2);
        drawXAxis(g2);
          
        if (hasStartStopCodon)
        {
          if (hasSomethingInTranscriptionRange)
          {
            drawUnionTranscript(g2);
            if (hasAtLeastOneCodingTranscript)
            {
              drawUnionProtein(g2);
              drawProteinVariations(g2);
              drawProteinDomains(g2);
              if (properties.isDrawStartStopCodonLines()){
                drawStartAndStopCodons(g2);
              }
              if(properties.isDrawAnyLegend()){
                drawLegends(g2);
              }
            }
            if (properties.isDrawHighlighting()){
              drawRegionHighlights(g2);
            }
          }
        }
        
        if (properties.isDrawGeneAndTranscriptLabels())
        {
          drawAllTrackLabels(g2);
        }
      }
      validateAllAndDrawErrors(g2);
    }
    
    for(GPVCanvasListener listener:canvasListeners){
      listener.canvasRefreshed();
    }
    
  }
  
  @Override
  public void paint(Graphics g)
  {
    try{
      if(bufferInvalidated || buffer==null || this.getWidth()!=buffer.getWidth() || this.getHeight()!=buffer.getHeight()){

        paintToBuffer();
      }
      g.drawImage(buffer,0,0,null);
    }catch(NeedMoreMemoryException mem){
      //not enough memory to buffer, draw directly.
      paintNotFromBuffer(g);
    }
  }
  
  private void drawRegionHighlights(Graphics2D g2)
  {
    g2.setColor(colors.getBlockBorder());
    Rectangle2D highlighter_block_cont = map.getContinuousRegionHighlighterBlock();
    g2.draw(highlighter_block_cont);
    
    Rectangle2D highlighter_block_variants = map.getPointRegionHighlighterBlock();
    g2.drawLine((int)highlighter_block_variants.getX(),(int)highlighter_block_variants.getMaxY(),(int)highlighter_block_variants.getMaxX(),(int)highlighter_block_variants.getMaxY());

    Collection<GPVHighlightedRegion> allregions = data.getHighlightedRegionsForGene(selectedGene,true);
    Collection<GPVHighlightedRegion> intronregions = GPVUtils.splitRegionsInIntronRegions(selectedGene, allregions);
    Collection<GPVHighlightedRegion> exonregions = GPVUtils.splitRegionsInExonRegions(selectedGene, allregions);
    TreeSet<GPVVariant> variants_only = GPVUtils.splitRegionsInVariants(allregions);
    
    drawRegionHighlightersOnTop(g2,intronregions, colors.getTopIntronRegionHighlighter());
    drawRegionHighlightersOnTop(g2,exonregions, colors.getTopExonRegionHighlighter());
    
    map.initVariantShapeDistribution(variants_only);
//    drawVariantsOnTop(g2,intronregions, colors.getTopIntronRegionHighlighter());
    drawVariantsOnTop(g2,allregions, colors.getTopExonRegionHighlighter());
    
    drawAllHighligtersOnTracks(g2,allregions);
  }
  
  
  private void drawRegionHighlightersOnTop(Graphics2D g2,Collection<GPVHighlightedRegion> regions, Color color){
    g2.setColor(color);
    //draw block marker on the top
    for (GPVHighlightedRegion region:regions)
    {
      if(!(region instanceof GPVVariant)){
        g2.fill(map.getContinuousRegionHighlighter(region.getStart(), region.getStop()));
      }
    }
  }
  
  private void drawVariantsOnTop(Graphics2D g2, Collection<GPVHighlightedRegion> regions, Color color){
    //draw variant shapes on top
    g2.setStroke(new BasicStroke(defaultStroke.getLineWidth() / 2f));
    g2.setColor(color);
    for (GPVHighlightedRegion region:regions)
    {
      if(region instanceof GPVVariant){
        GPVVariant variant = (GPVVariant)region;
        GPVVariantShape shape = map.getVariantHighlighter(variant);
        Point2D line_destination = map.getVariantHighlighterLineDock(variant);
        if(shape.isSupposedToBeFilled()){
          g2.fill(shape);
        }else{
          g2.draw(shape);
        }
        g2.drawLine((int)shape.getLineAnchor().getX(), (int)shape.getLineAnchor().getY(), (int)line_destination.getX(), (int)line_destination.getY());
        
      }
    }
    g2.setStroke(defaultStroke);
  }
  
  private void drawAllHighligtersOnTracks(Graphics2D g2, Collection<GPVHighlightedRegion> allregions){
    TreeSet<GPVHighlightedRegion> regions;
    
    if (properties.getRegionHighlightMode() == CanvasProperties.REGION_HIGHLIGHT_MODE_ALL)
    {
      regions = GPVUtils.splitRegionsInCodingRegions(selectedGene, allregions);
      drawRegionHighlightersOnTracks(regions,g2, colors.getHighlighterColor(), colors.getSNPHighlighterColor());
    }

    HashMap<String,ArrayList<GPVDomain>> domains = data.getProteinDomainsForGene(selectedGene);

    if(hasSomethingInTranscriptionRange && hasAtLeastOneCodingTranscript && hasStartStopCodon){
      for (GPVTranscript transcript:selectedGene.getEnabledTranscriptsSorted())
      {
        if (transcript.isCoding())
        {
          ArrayList<GPVDomain> prot_domains = domains.get(transcript.getResultingProtein().getIDFrom(data.getDomainSource()));
          if (prot_domains != null)
          {
            regions = GPVUtils.splitRegionsInCodingRegions(transcript, allregions);
            regions = GPVUtils.splitRegionsInProteinDomainRegions(transcript, prot_domains, regions);
            drawRegionHighlightersOnTracks(regions,g2, colors.getDomainOverlapHighlighterFill(), colors.getSNPDomainOverlapHighlighterFill());
          }
        }
      }
    }else{
      regions =  GPVUtils.splitRegionsInExonRegions(selectedGene, allregions);
      drawRegionHighlightersOnTracks(regions,g2, colors.getHighlighterColor(), colors.getSNPHighlighterColor());
    }
  }
  
  private void drawRegionHighlightersOnTracks(Collection<GPVHighlightedRegion> regions, Graphics2D g2, Color color_region, Color color_point){
    //do it once for snps and once for regions
    //this is to make sure that snps are not covered up by regions, by drawing the regions first and then the snps/points
    for(int i = 0; i<2; i++){
      for (GPVHighlightedRegion region:regions)
      {
        boolean is_region = !(region instanceof GPVVariant); 
        if((is_region && i==0) || (!is_region && i==1)){
          if(is_region){
            g2.setColor(color_region);
          }else{
            g2.setColor(color_point);
          }
          
          Rectangle2D t_high_block = ensureMinimumWidth(map.getUnionTranscriptMarkerBlock(region.getStart(), region.getStop(), defaultStroke.getLineWidth()));
          Rectangle2D p_high_block = null;
          if(hasSomethingInTranscriptionRange && hasAtLeastOneCodingTranscript && hasStartStopCodon){
            p_high_block = ensureMinimumWidth(map.getUnionProteinHighlighterBlock(region.getStart(), region.getStop(),defaultStroke.getLineWidth()));
          }

          // on the union transcript
          g2.fill(t_high_block);
          // on the protein
          if(p_high_block!=null)g2.fill(p_high_block);

          if (properties.isDrawHelperLines() && p_high_block!=null)
          {
            if (is_region){
              g2.drawLine((int)t_high_block.getX(), (int)t_high_block.getMaxY(), (int)p_high_block.getX(), (int)p_high_block.getY());
              g2.drawLine((int)t_high_block.getMaxX(), (int)t_high_block.getMaxY(), (int)p_high_block.getMaxX(), (int)p_high_block.getY());  
            }else{
              g2.drawLine((int)t_high_block.getCenterX(), (int)t_high_block.getMaxY(), (int)p_high_block.getCenterX(), (int)p_high_block.getY());
            }
          }
        }
      }
    }
  }

  private void drawTitle(Graphics2D g2)
  {
    g2.setColor(properties.getColorScheme().getTitleColor());
    int fontsize = (int) map.hscale(properties.getHeaderHeightPercent() * 0.8f);
    g2.setFont(new Font("Arial", Font.PLAIN, fontsize));
    String gname = selectedGene.getDisplayName();
    String chr = selectedGene.getChromosome().getName();
    String from = selectedGene.getFirstKnownPosition() + "";
    String to = selectedGene.getLastKnownPosition() + "";
    String strand = selectedGene.getStrand()==DNAStrand.REVERSE?"-":"+";
    
    String title = "Gene \"" + gname + "\" on " + chr + ":" + from + "-" + to +" ( "+strand+" strand)";
    int y = (int)map.hscale(map.getBorderNorth()+properties.getHeaderHeightPercent() * 0.8f);
    int x = (int) map.wscale(0.01f);
    g2.drawString(title, x, y);
  }

  private void drawGene(Graphics2D g2)
  {

    Rectangle2D gene = map.getGeneBlock();
    g2.setColor(colors.getIntronFill());
    g2.fillRect((int) gene.getX(), (int) gene.getY(), (int) gene.getWidth(), (int) gene.getHeight());
    g2.setColor(colors.getBlockBorder());
    g2.drawRect((int) gene.getX(), (int) gene.getY(), (int) gene.getWidth(), (int) gene.getHeight());

    // draw direction
    int fontsize = (int) map.wscale(0.02f);
    g2.setFont(new Font("Arial", Font.PLAIN, fontsize));
    g2.setColor(Color.black);
    g2.drawString("5'", (int) (gene.getX() - map.wscale(0.02f)), (int) gene.getY() + fontsize);
    g2.drawString("3'", (int) (gene.getMaxX() + map.wscale(0.005f)), (int) gene.getY() + fontsize);

    // draw exons
    drawExons(g2, 0, selectedGene.getExons(), colors.getExonFill(), colors.getBlockBorder());
  }

  private void drawUnionTranscript(Graphics2D g2)
  {
    int transcript_index = selectedGene.getEnabledTranscriptsSorted().size() + 1;
    Line2D rail = map.getUnionTranscriptRail();
    g2.setColor(colors.getTranscriptRails());
    g2.drawLine((int) rail.getX1(), (int) rail.getY2(), (int) rail.getX2(), (int) rail.getY2());

    TreeSet<GPVExon> unionTranscript = selectedGene.getAllExonRegionsSeparately(true, true);
    drawExons(g2, transcript_index, unionTranscript, colors.getUnionTranscriptFill(), colors.getBlockBorder());

  }

  private void drawStartAndStopCodons(Graphics2D g2)
  {
    g2.setStroke(new BasicStroke(defaultStroke.getLineWidth() * 2f));

    if (selectedGene.getEarliestStartCodon() > 0 && selectedGene.getLatestStopCodon() > 0)
    {
      // first for the gene
      Rectangle2D block = map.getGeneBlock();
      int last_y1 = (int) ((map.getPointRegionHighlighterBlock().getMaxY() +block.getY())/2f);
      int last_y2 = (int) block.getMaxY();
      int gene_x1 = map.transformNTToAbsoluteHorizontalPosition(selectedGene.getEarliestStartCodon());
      int gene_x2 = map.transformNTToAbsoluteHorizontalPosition(selectedGene.getLatestStopCodon());
      g2.setColor(colors.getStartCodon());
      g2.drawLine(gene_x1, last_y1, gene_x1, last_y2); // start
      g2.setColor(colors.getStopCodon());
      g2.drawLine(gene_x2, last_y1, gene_x2, last_y2); // stop

      int i = 1;
      for (GPVTranscript transcript:selectedGene.getEnabledTranscriptsSorted())
      {
        if (transcript.isCoding())
        {
          block = map.getTranscriptBlock(i);
          int y1 = (int) block.getY();
          int y2 = (int) block.getMaxY();
          int x1 = map.transformNTToAbsoluteHorizontalPosition(transcript.getStartCodon());
          int x2 = map.transformNTToAbsoluteHorizontalPosition(transcript.getStopCodon());

          // first draw the connection between the tracks
          g2.setColor(colors.getStartCodon());
          // g2.drawLine(last_x1, last_y2, x1, y1);
          // now draw the line on the track
          g2.drawLine(x1, y1, x1, y2);

          // and now the same for the stop codon
          g2.setColor(colors.getStopCodon());
          // g2.drawLine(last_x2, last_y2, x2, y1);
          g2.drawLine(x2, y1, x2, y2);

          last_y1 = y1;
          last_y2 = y2;
        }
        i++;
      }

      // and now for the union transcript
      block = map.getTranscriptBlock(selectedGene.getEnabledTranscriptsSorted().size() + 1);
      int y1 = (int) block.getY();
      int y2 = (int) block.getMaxY();
      g2.setColor(colors.getStartCodon());
      // first draw the connection between the tracks
      // g2.drawLine(last_x1, last_y2, gene_x1, y1);
      // now draw the line on the track
      g2.drawLine(gene_x1, y1, gene_x1, y2);

      // and now the same for the stop codon
      g2.setColor(colors.getStopCodon());
      // g2.drawLine(last_x2, last_y2, gene_x2, y1);
      g2.drawLine(gene_x2, y1, gene_x2, y2);

      // last, connect it to the union protein
      g2.setColor(colors.getStartCodon());
      block = map.getProteinVariantBlock(0);

    }
    g2.setStroke(defaultStroke);
  }

  private void drawUnionProtein(Graphics2D g2)
  {
    for (GPVExon exon:selectedGene.getAllExonRegionsSeparately(false, true))
    {
      Rectangle2D prot = map.getProteinRegionBlockForExon(exon, 0);
      if (prot != null)
      {
        g2.setColor(colors.getUnionProteinFill());
        g2.fill(prot);
        g2.setColor(colors.getBlockBorder());
        g2.draw(prot);

        // draw the connections between the union transcript and the union protein
        if (properties.isDrawHelperLines())
        {
          g2.setColor(colors.getGuidelines());
          Shape corresponding_transcript_block = map.getExon(exon, selectedGene.getEnabledTranscriptsSorted().size() + 1);
          int x1 = (int) corresponding_transcript_block.getBounds().getX();
          int x2 = (int) prot.getX();
          int y1 = (int) corresponding_transcript_block.getBounds().getMaxY();
          int y2 = (int) prot.getY();

          g2.drawLine(x1, y1, x2, y2);

          x1 = (int) corresponding_transcript_block.getBounds().getMaxX();
          x2 = (int) prot.getMaxX();
          g2.setColor(colors.getGuidelines());
          g2.drawLine(x1, y1, x2, y2);
        }
      }
    }
  }

  private void drawProteinVariations(Graphics2D g2)
  {
    int i = 1;
    for (GPVTranscript transcript:selectedGene.getEnabledTranscriptsSorted())
    {
      if (transcript.isCoding())
      {
        Line2D rail = map.getProteinVariantRail(i);
        g2.setColor(colors.getTranscriptRails());
        g2.draw(rail);

        for (GPVExon exon:transcript.getAllExonRegionsSeparately())
        {
          Rectangle2D protein_block = map.getProteinRegionBlockForExon(exon, i);

          if (protein_block != null)
          {
            g2.setColor(colors.getProteinFill());
            g2.fill(protein_block);
            g2.setColor(colors.getBlockBorder());
            g2.draw(protein_block);
          }
        }
        
        if(properties.isShowProteinXAxis()){
          //draw axis
          Rectangle2D axis_block = map.getProteinVariantAxisBlock(i);
          g2.setFont(new Font("Arial", Font.PLAIN, (int)axis_block.getHeight()));
          GPVProtein prot = transcript.getResultingProtein();
          g2.setColor(colors.getBlockBorder());
          int interval = (int)(Math.ceil(prot.getProteinLength()/100d)*10);
          int prev_x =0;
          for(int j = interval; j< prot.getProteinLength(); j+=interval){
            int x = map.transformPPToAbsoluteHorizontalPosition(j, i);
            if(x-prev_x>30){
              g2.drawLine(x,(int)axis_block.getY(), x, (int)axis_block.getMaxY());
              Rectangle2D stringbounds = g2.getFontMetrics().getStringBounds(""+j, g2);
              g2.drawString(""+j, (int)(x-stringbounds.getWidth()-2), (int)(axis_block.getMaxY()-g2.getFontMetrics().getDescent()));
              prev_x = x;
            }
          }
        }

        i++;
      }
    }
  }

  private void drawTranscripts(Graphics2D g2)
  {
    int i = 1;
    for (GPVTranscript transcript:selectedGene.getEnabledTranscriptsSorted())
    {
      Line2D rail = map.getTranscriptRail(i);
      g2.setColor(colors.getTranscriptRails());
      g2.drawLine((int) rail.getX1(), (int) rail.getY2(), (int) rail.getX2(), (int) rail.getY2());

      drawExons(g2, i, transcript.getExons(), colors.getTranscriptFill(), colors.getBlockBorder());
      i++;
    }
  }

  private void drawAllTrackLabels(Graphics2D g2)
  {

    // draw gene name
    Point2D loc = map.getGeneLabelLocation();
    drawTrackLabel(selectedGene.getDisplayName(), loc, g2, false);

    if (hasStartStopCodon)
    {
      // draw union transcript name
      loc = map.getTranscriptLabelLocation(selectedGene.getEnabledTranscriptsSorted().size() + 1);
      drawTrackLabel("Union transcript", loc, g2, false);

      if (hasAtLeastOneCodingTranscript)
      {
        // draw union protein
        loc = map.getProteinLabelLocation(0);
        drawTrackLabel("Union protein", loc, g2, false);
      }
    }
    // draw transcript ids
    int i = 1;
    int prot_i = 1;
    for (GPVTranscript transcript:selectedGene.getEnabledTranscriptsSorted())
    {
      // draw transcript name
      loc = map.getTranscriptLabelLocation(i);
      drawTrackLabel(transcript.getTranscriptId(), loc, g2, false);

      // draw protein label
      if (transcript.isCoding() && transcript.getResultingProtein() != null && hasStartStopCodon)
      {
        String id = transcript.getResultingProtein().getProteinID();
        if (id != null)
        {
          loc = map.getProteinLabelLocation(prot_i);
          drawTrackLabel(id, loc, g2, true);
        }
        prot_i++;
      }

      i++;
    }
    //header area labels
    
    Rectangle2D highlighter_block_cont = map.getContinuousRegionHighlighterBlock();
    g2.setFont(new Font("Arial", Font.PLAIN, (int)(highlighter_block_cont.getHeight()*0.7f)));
    
    int label_offset = (int)((highlighter_block_cont.getHeight()-g2.getFont().getSize())/2);
    
    int x = (int)map.wscale(map.getBorderEastWest());
    String label = "Regions";
    loc = new Point2D.Double(x, highlighter_block_cont.getMaxY() - label_offset);
    drawLabelWithShadow(label,loc,g2);
    
    label="Variants";
    Rectangle2D highlighter_block_variants = map.getPointRegionHighlighterBlock();
    loc = new Point2D.Double(x, highlighter_block_variants.getMaxY()-label_offset);
    drawLabelWithShadow(label,loc,g2);
    
    label="Scale";
    Rectangle2D xaxis_block = map.getXAxisBlock();
    loc = new Point2D.Double(x, (int)xaxis_block.getMaxY()-label_offset);
    drawLabelWithShadow(label,loc,g2);
    
  }
  
  public Font getTrackLabelFont(boolean protein){
    int fontsize = protein ? getProteinLabelFontSize() : getGeneLabelFontSize();
    return(new Font("Arial", Font.PLAIN, fontsize));
  }
  
  private void drawTrackLabel(String name, Point2D loc, Graphics2D g2, boolean protein)
  {
    g2.setFont(getTrackLabelFont(protein));
    drawLabelWithShadow(name,loc,g2);
  }
  
  private void drawLabelWithShadow(String name, Point2D loc, Graphics2D g2)
  {
    Rectangle2D shadow = getTrackLabelBlock(loc,name,g2.getFont());
    g2.setColor(Color.white);
    g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.75f));
    g2.fill(shadow);
    g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1f));
    g2.setColor(Color.black);
    g2.drawString(name, (int) loc.getX(), (int) loc.getY());
  }
  
  public Rectangle2D getTrackLabelBlock(Point2D labelLocation, String label, Font font){
    if(label==null)return null;
    Graphics g = getGraphics();
    if(g==null){
      g = new BufferedImage(1, 1, BufferedImage.TYPE_3BYTE_BGR).getGraphics();
    }
    FontMetrics metrics = g.getFontMetrics(font);
    int width = metrics.stringWidth(label);
    Rectangle2D shadow = new Rectangle2D.Double(labelLocation.getX() - map.wscale(0.005f), labelLocation.getY() - font.getSize(), width + map.wscale(0.01f), font.getSize()+2);
    return shadow;
  }
  
  public Rectangle2D getGeneTrackLabelBlock(){
    return getTrackLabelBlock(map.getGeneLabelLocation(), selectedGene.getDisplayName(), getTrackLabelFont(false));
  }
  public Rectangle2D getTranscriptTrackLabelBlock(int transcript){
    String name = null;
    int i =1;
    for(GPVTranscript tr:selectedGene.getEnabledTranscriptsSorted()){
      if(i++==transcript){
        name = tr.getTranscriptId();
      }
    }
    return getTrackLabelBlock(map.getTranscriptLabelLocation(transcript), name, getTrackLabelFont(false));
  }
  public Rectangle2D getProteinTrackLabelBlock(int variation){
    String name = null;
    int i =1;
    for(GPVTranscript tr:selectedGene.getEnabledTranscriptsSorted()){
      if(tr.isCoding()){
        if(i++==variation){
          name = tr.getResultingProtein().getProteinID();
        }
      }
    }
    return getTrackLabelBlock(map.getProteinLabelLocation(variation), name, getTrackLabelFont(true));
  }

  private void drawXAxis(Graphics2D g2)
  {
    int n_start = selectedGene.getFirstKnownPosition();
    int n_end = selectedGene.getLastKnownPosition();
    
    Rectangle2D axisBlock = map.getXAxisBlock();

    int axis_y = (int)((axisBlock.getMaxY()+axisBlock.getY())/2f);
    int tick_y1 = (int) axisBlock.getY();
    int tick_y2 = (int) axisBlock.getMaxY();

    int axis_x1 = map.transformNTToAbsoluteHorizontalPosition(n_start);
    int axis_x2 = map.transformNTToAbsoluteHorizontalPosition(n_end);

    g2.setColor(Color.black);
    g2.drawLine(axis_x1, axis_y, axis_x2, axis_y);

    for (int i = n_start; i < n_end; i += 100)
    {
      int tick_x = map.transformNTToAbsoluteHorizontalPosition(i);
      g2.drawLine(tick_x, tick_y1, tick_x, tick_y2);
    }

    // draw arrow
    int border = (int) map.wscale(map.getBorderEastWest());
    int[] arrow_x = new int[] { getWidth()-border / 2, getWidth()-border, getWidth()-border };
    int[] arrow_y = new int[] { axis_y, tick_y1, tick_y2 };
    g2.fill(new Polygon(arrow_x, arrow_y, 3));

  }

  /**
   * Draw a list of exons in a specific area. If transcript object is provided we assume we draw exons for a transcript This is important for calculating the
   * necessary gaps.
   * 
   * @param g2
   * @param ypos
   * @param transcript_index: set to 0 for the gene
   * @param transcript: set to null for the gene
   */
  private void drawExons(Graphics2D g2, int transcript_index, Collection<GPVExon> exons, Color fill_color, Color border_color)
  {
    boolean is_transcript = transcript_index != 0;
    // determine total amount of space required for exons:
    int i = 1;
    for (GPVExon exon:exons)
    {
      Shape exon_block;
      if (is_transcript)
      {
        exon_block = map.getExon(exon, transcript_index);
      } else
      {
        exon_block = map.getExonOnGene(i);
      }

      g2.setColor(fill_color);
      g2.fill(exon_block);
      g2.setColor(border_color);
      g2.draw(exon_block);

      // draw lines for merge sites
      g2.setColor(colors.getMergeSiteMarkerColor());
      float[] dash = new float[] { (float) map.getGeneBlock().getHeight() * 0.1f };
      g2.setStroke(new BasicStroke(defaultStroke.getLineWidth(), BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 1f, dash, 0.0f));
      for (Line2D mergeSiteLine:map.getMergeSiteLines(exon))
      {
        g2.drawLine((int) mergeSiteLine.getX1(), (int) mergeSiteLine.getY1(), (int) mergeSiteLine.getX2(), (int) mergeSiteLine.getY2());
      }
      g2.setStroke(defaultStroke);

      i++;
    }
  }

  private void drawProteinDomains(Graphics2D g2)
  {
    HashMap<String,ArrayList<GPVDomain>> protein_domains = data.getProteinDomainsForGene(selectedGene);
    int sets = data.getEnabledDomainSets().size();

    int height = (int) ((float) (map.getProteinTrackHeight() - 1) / (float) sets) - 2;
    HashMap<String,Integer> set_y = new HashMap<String,Integer>();
    domainSetColors = new HashMap<String,Color>();
    int i = 0;
    ColorProvider colorprovider = colors.getDomainSetColorProvider();
    for (GPVDomainSet set:data.getDomainSets())
    {
      if (set.isEnabled())
      {
        set_y.put(set.getSetName(), (i * height) + (i * 2)) /* margin */;
        i++;
      }
      Color next = colorprovider.next();
      domainSetColors.put(set.getSetName(), next);
    }

    int variation_nr = 1;
    for (GPVTranscript transcript:selectedGene.getEnabledTranscriptsSorted())
    {
      if (transcript.isCoding())
      {
        ArrayList<GPVDomain> domains = protein_domains.get(transcript.getResultingProtein().getIDFrom(data.getDomainSource()));
        if (domains != null)
        {
          domains = GPVUtils.mergeProteinDomainsThatOverlap(domains);
          int base_y = (int) map.getProteinVariantBlock(variation_nr).getY() + 2;
          for (GPVDomain domain:domains)
          {
            Rectangle2D[] full_blocks_split = map.getProteinRangeAsSeparateBlock(domain.getProteinStartPos(), domain.getProteinEndPos(), variation_nr);
            int y1 = base_y + set_y.get(domain.getParentSet().getSetName());
            Color color = domainSetColors.get(domain.getParentSet().getSetName());
            g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.7f));
            g2.setColor(color);

            for (Rectangle2D full_block:full_blocks_split)
            {
              int x1 = (int) full_block.getX();
              int width = (int) full_block.getWidth();
              g2.fillRect(x1, y1, width, height);
            }
            g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1f));

            int fontsize = (int)Math.ceil((float) height * 0.8f);
            fontsize = (int)Math.min(fontsize, map.hscale(0.02f)); // set max font size
            
            g2.setFont(new Font("Arial", Font.PLAIN, fontsize));
            int minfontsize = (int)map.hscale(0.015f);
            
            if (height > minfontsize || !properties.isUseDynamicDomainFontSize())
            {
              // write name
              // determine font size
              int first_x = map.transformPPToAbsoluteHorizontalPosition(domain.getProteinStartPos(), variation_nr);
              int last_x = map.transformPPToAbsoluteHorizontalPosition(domain.getProteinEndPos(), variation_nr);
              int available_width = last_x - first_x;
              if(properties.isUseDynamicDomainFontSize()){
                setFontSizeToFitWith(g2, domain.getDescription(), fontsize, minfontsize, available_width, "Arial");
              }
              String shortened_domainname = shortenTextToFitWith(g2, domain.getDescription(), available_width);
              //if we cannot fit at least a certain amount of the domain name in there, use a footnote
              if(shortened_domainname.length()<Math.min(domain.getDescription().length(),domain.getDescription().length()*properties.getMinimumDomainNameVisibleFraction())){
                this.footnotes.addFootnote(domain.getDescription());
                Integer footnote = this.footnotes.getIndexOf(domain.getDescription());
                shortened_domainname ="*"+footnote;
              }
              g2.setColor(Color.black);
              g2.drawString(shortened_domainname, first_x + 1, y1 + height - g2.getFontMetrics().getDescent());
            }
          }
        }
        variation_nr++;
      }
    }
  }
  
  private boolean[] needsLegends(){
    boolean prot_legend = properties.isDrawDomainRegionLegend() && hasDomainSets;
    boolean footnotes = properties.isDrawFootnotesLegend() && hasDomainSets && this.footnotes.size()>0;
    boolean variant_legend = properties.isDrawVariantTypesLegend() && hasVariants;
    return new boolean[]{prot_legend,variant_legend,footnotes};
  }
  
  private boolean needsAnyLegend(){
    boolean[] legends = needsLegends();
    boolean ret = false;
    for(boolean legend:legends)ret=ret||legend;
    return ret;
  }

  public void drawLegends(Graphics2D g2){
    Rectangle2D space = map.getAvailableSpaceForLegend();
    boolean[] legends = needsLegends();
    boolean prot_legend = legends[0];
    boolean variant_legend = legends[1];
    boolean footnotes = legends[2];

    int segments = (prot_legend?1:0)+(footnotes?1:0)+(variant_legend?1:0);
    
    String prot_header="Protein domain sets";
    String footnotes_header="Domain footnotes";
    String variant_header="Variant types";
    
    int fontsize = (int) (space.getHeight()*0.18f);
    int fontpadding = (int)(fontsize*0.15f);

    if(segments>0){
      double x_margin = space.getWidth()*0.01d;
      double x_padding = space.getWidth()*0.01d;
      double y_padding_bottom = space.getHeight()*0.02d;
      double y_padding_top = y_padding_bottom+fontsize+fontpadding*2;
      double subspace_width = space.getWidth()/segments;

      double x = space.getX();

      Rectangle2D subspace;
      Rectangle2D innerspace;
      
      for(int i = 0; i<3; i++){
        boolean show = false;
        switch(i){
          case 0: show = prot_legend; break;
          case 1: show = variant_legend; break;
          case 2: show = footnotes; break;
        }
        
        if(show){
          subspace = new Rectangle2D.Double(x,space.getY(),subspace_width-x_margin,space.getHeight());
          innerspace = applyPadding(subspace,x_padding,y_padding_top, y_padding_bottom);
          g2.setColor(colors.getBlockBorder());
          g2.draw(subspace);
          String header="";
          switch(i){
            case 0:  
              header = prot_header;
              drawDomainSetLegend(g2, innerspace);
              break;
            case 1: 
              header=variant_header;
              drawVariantLegend(g2,innerspace);
              break;
            case 2: 
              header=footnotes_header;
              drawFootnotes(g2, innerspace);
              break;
          }
          g2.setFont(new Font("Arial",Font.PLAIN,fontsize));
          g2.setColor(colors.getTitleColor());
          g2.drawString(header, (int)innerspace.getX(), (int)subspace.getY()+fontsize+fontpadding);
          x+=subspace_width+x_margin;
        }
      }
    }
  }
  
  public void drawDomainSetLegend(Graphics2D g2, Rectangle2D space)
  {
    ArrayList<GPVDomainSet> domainsets = data.getEnabledDomainSets();
    int n = domainsets.size();
    int rows = (int)Math.ceil(Math.sqrt(n)*0.95f);
//    int rows = 2;// (int)Math.ceil(Math.sqrt(n));
    int cols = (int) Math.ceil((float) n / (float) rows);
    int i = 0;
    int y_offset = (int) Math.max(1, space.getHeight() * 0.15f / rows);
    int x_offset = (int) Math.max(3, space.getWidth() * 0.03f / cols);
    int boxheight = (int) ((space.getHeight()) / rows - y_offset);
    boxheight = (int) Math.min(boxheight, map.hscale(0.025f));
    
    g2.setFont(new Font("Arial", Font.PLAIN, boxheight - 2));
    for (GPVDomainSet set:domainsets)
    {
      int col = i % cols;
      int row = i / cols;
      int y = (int)space.getY() + (boxheight + y_offset) * row;
      int x = (int) (space.getX() + space.getWidth() / cols * col + x_offset * (col));

      g2.setColor(domainSetColors.get(set.getSetName()));
      g2.fillRect(x, y, boxheight, boxheight);
      g2.setColor(Color.black);
      g2.drawRect(x, y, boxheight, boxheight);
      g2.drawString(set.getSetName(), x + boxheight + 5, y + boxheight);
      i++;
    }
  }
  
  public void drawVariantLegend(Graphics2D g2, Rectangle2D space)
  {
    HashMap<String, GPVVariantShape> variant_classifications = new HashMap<String,GPVVariantShape>();
    TreeSet<GPVHighlightedRegion> regions = data.getHighlightedRegionsForGene(selectedGene, true);
    
    for(GPVHighlightedRegion region:regions){
      if(region instanceof GPVVariant){
        GPVVariant variant = (GPVVariant)region;
        GPVVariantShape shape = variantShapes.getShapeFor(variant);
        variant_classifications.put(variant.getVariantTypeString(false), shape);
      }
    }
    
    int n = variant_classifications.size();
    int rows = (int)Math.ceil(Math.sqrt(n)*0.90f);
//    int rows = 2;// (int)Math.ceil(Math.sqrt(n));
    int cols = (int) Math.ceil((float) n / (float) rows);
    int i = 0;
    int y_offset = (int) Math.max(1, space.getHeight() * 0.1f / rows);
    int x_offset = (int) Math.max(3, space.getWidth() * 0.03f / cols);
    int boxheight = (int) ((space.getHeight()) / rows - y_offset * rows);
    boxheight = (int) Math.min(boxheight, map.hscale(0.025f));

    g2.setFont(new Font("Arial", Font.PLAIN, boxheight - 2));
    for (Entry<String,GPVVariantShape> entry:variant_classifications.entrySet())
    {
      int col = i % cols;
      int row = i / cols;
      int y = (int)space.getY() + (boxheight + y_offset) * row;
      int x = (int) (space.getX() + space.getWidth() / cols * col + x_offset * (col));

      g2.setColor(Color.black);
      GPVVariantShape shape = entry.getValue();
      shape.init(new Point2D.Double(x+boxheight/2d,y),boxheight,boxheight);
      
      if(shape.isSupposedToBeFilled()){
        g2.fill(shape);
      }else{
        g2.draw(shape);
      }
      g2.drawString(entry.getKey(), x + boxheight + 5, y + boxheight);
      i++;
    }
  }
  
  
  public void drawFootnotes(Graphics2D g2, Rectangle2D space)
  {
    int n = footnotes.size();

    if(n>0){
      int cols = (int)Math.ceil(Math.sqrt(n)*0.40f);
      int rows = (int) Math.ceil((float) n / (float) cols);
      int i = 0;
      int y_offset = (int) Math.max(1, space.getHeight() * 0.1f / rows);
      int x_offset = (int) Math.max(3, space.getWidth() * 0.1f / cols);
      int fontheight = (int) ( space.getHeight() / (float)rows - (float)y_offset);
      fontheight = (int) Math.min(fontheight, map.hscale(0.02f));
      int width_per_column =  (int)space.getWidth() / cols;
      
      //determine perfect font size:
      int minfontsize = (int)map.hscale(0.008f);
      for (Entry<Integer,String> footnote:footnotes.getFootnotes().entrySet())
      {
        String text = Footnotes.footNoteToString(footnote.getKey(),footnote.getValue());
        setFontSizeToFitWith(g2, text, fontheight, minfontsize, width_per_column, "Arial");
        fontheight = g2.getFont().getSize();
      }

      g2.setFont(new Font("Arial", Font.PLAIN, fontheight));
      
      for (Entry<Integer,String> footnote:footnotes.getFootnotes().entrySet())
      {
        int col = i % cols;
        int row = i / cols;
        int y = (int)space.getY() + fontheight + (fontheight + y_offset) * row;
        int x = (int) (space.getX() + width_per_column * col + x_offset * col);
        g2.drawString(Footnotes.footNoteToString(footnote.getKey(),footnote.getValue()), x, y);
        i++;
      }
    }
  }

  private void setFontSizeToFitWith(Graphics2D g2, String text_to_fit, int init_fontsize, int min_font_size, int available_width, String font_face)
  {
    int curr_fontsize = init_fontsize;

    int required_width;
    do
    {
      g2.setFont(new Font(font_face, Font.PLAIN, curr_fontsize));
      required_width = (int) g2.getFontMetrics().getStringBounds(text_to_fit, g2).getWidth();
      curr_fontsize--;
    } while (required_width > available_width && curr_fontsize > min_font_size);

    curr_fontsize = g2.getFont().getSize();
    g2.setFont(new Font("Arial", Font.PLAIN, Math.max(curr_fontsize, min_font_size)));
  }
  
  private Rectangle2D ensureMinimumWidth(Rectangle2D rect){
    if(rect.getWidth()>defaultStroke.getLineWidth())return rect;
    return new Rectangle2D.Float((float)rect.getCenterX()-(defaultStroke.getLineWidth()/2f),(float)rect.getY(),defaultStroke.getLineWidth(),(float)rect.getHeight());
  }

  private String shortenTextToFitWith(Graphics2D g2, String text_to_fit, int available_width)
  {
    String ret = text_to_fit;
    if (text_to_fit.length() <= 3)
      return text_to_fit;
    int required_width = Integer.MAX_VALUE;
    while (required_width > available_width && ret.length() > 4)
    {
      required_width = (int) g2.getFontMetrics().getStringBounds(ret, g2).getWidth();
      if (required_width > available_width)
      {
        ret = ret.substring(0, ret.length() - 4) + "...";
      }
    }
    if (ret.length() <= 4 && text_to_fit.length() > 4)
    {
      ret = text_to_fit.substring(0, 3);
    }
    return ret;
  }
  
  private Rectangle2D applyPadding(Rectangle2D rect, double x_padding, double y_padding_top, double y_padding_bottom){
    return new Rectangle2D.Double(rect.getX()+x_padding, rect.getY()+y_padding_top, rect.getWidth()-x_padding*2d, rect.getHeight()-y_padding_top-y_padding_bottom);
  }

  private void validateAllAndDrawErrors(Graphics2D g2)
  {
    validateAll();
    int fontsize = (int) map.wscale(0.02f);
    g2.setColor(Color.red);
    g2.setFont(new Font("Arial", Font.PLAIN, fontsize));
    int error_offset = 0;
    for (String error:errors)
    {
      error_offset += (int) ((float) fontsize * 1.5f);
      g2.drawString(error, 10, getHeight() - error_offset);
    }
  }

  private void validateAll()
  {
    this.errors = new ArrayList<String>();
    hasTranscriptSelected = true;
    hasStartStopCodon = true;
    hasSomethingInTranscriptionRange = true;
    hasAtLeastOneCodingTranscript = true;

    if(selectedGene!=null){
      // check if there is a transcript selected
      if (selectedGene.getEnabledTranscriptsSorted().size() == 0)
      {
        errors.add("No transcripts selected.");
        hasTranscriptSelected = false;
        return; // no further checks needed
      }
      // check start stop codon
      int coding = 0;
      for (GPVTranscript tr:selectedGene.getEnabledTranscriptsSorted())
      {
        if (tr.isCoding())
        {
          coding++;
          if (tr.getStartCodon() == 0 || tr.getStopCodon() == 0)
          {
            errors.add("Transcript " + tr.getTranscriptId() + " is missing start/stop codon information.");
            hasStartStopCodon = false;
          } else
          {
            if (tr.getAllExonRegionsSeparately(false).size() == 0)
            {
              errors.add("Transcript " + tr.getTranscriptId() + " has no exon within the transcription range.");
              hasSomethingInTranscriptionRange = false;
            }
          }
        }
      }
      hasDomainSets = data.getEnabledDomainSets() != null && data.getEnabledDomainSets().size() > 0;

      hasVariants = data.isGeneHasHighlightedRegions(selectedGene, true,false, true);

      if (coding == 0)
        hasAtLeastOneCodingTranscript = false;
    }
  }

  @Override
  public Dimension getMinimumSize()
  {
    return new Dimension(100, 100);
  }

  public CanvasProperties getProperties()
  {
    return properties;
  }

  public void setProperties(CanvasProperties properties)
  {
    this.properties = properties;
  }

  public GPVGene getSelectedGene()
  {
    return selectedGene;
  }

  public void setSelectedGene(GPVGene selectedGene)
  {
    this.selectedGene = selectedGene;
  }

  public int getNucleotidePositionOn(int x, int y)
  {
    if (errors.size() > 0)
      return -1;
    return map.getNucleotidePositionOn(x, y);
  }

  public int getProteinPositionOn(int x, int y)
  {
    if (errors.size() > 0)
      return -1;
    return map.getProteinPositionOn(x, y);
  }

  public int getTranscriptThatPixelIsIn(int x, int y)
  {
    if (errors.size() > 0)
      return -1;
    return map.getTranscriptThatPixelIsIn(x, y);
  }

  public GPVExon getExonAt(int x, int y)
  {
    return map.getExonAt(x, y);
  }

  private int getGeneLabelFontSize()
  {
    int fontsize = (int) (map.getGeneTrackHeight() * 0.3f);
    fontsize = (int) Math.max(map.hscale(0.01f), fontsize);
    fontsize = (int) Math.min(map.hscale(0.02f), fontsize);
    return fontsize;
  }

  private int getProteinLabelFontSize()
  {
    int fontsize = (int) (map.getProteinTrackHeight() * 0.3f);
    fontsize = (int) Math.max(map.hscale(0.01f), fontsize);
    fontsize = (int) Math.min(map.hscale(0.02f), fontsize);
    return fontsize;
  }
  
  public GPVHighlightedRegion getHighlightedRegionAt(int pixel_x, int pixel_y){
    Point loc = new Point(pixel_x,pixel_y);
    TreeSet<GPVHighlightedRegion> highlightedRegions = data.getHighlightedRegionsForGene(selectedGene,true);
    
    if(highlightedRegions!=null && highlightedRegions.size()>0){
      for(GPVHighlightedRegion region:highlightedRegions){
        //is it in the region on the top highlighter area?
        if(map.getContinuousRegionHighlighter(region.getStart(), region.getStop()).contains(pixel_x, pixel_y)){
          return region;    
        }  
        
        //is it in the region on the union transcript?
        Rectangle2D t_high_block = map.getUnionTranscriptMarkerBlock(region.getStart(), region.getStop(),defaultStroke.getLineWidth());
        if(t_high_block.contains(loc)) return region;
      }
      for(GPVTranscript transcript:selectedGene.getEnabledTranscriptsSorted()){
        //is it in the region on the union protein?
        if(transcript.isCoding()){
          highlightedRegions = GPVUtils.splitRegionsInCodingRegions(transcript, highlightedRegions);
          for(GPVHighlightedRegion region:highlightedRegions){
            Rectangle2D p_high_block = map.getUnionProteinHighlighterBlock(Math.min(region.getStart(), region.getStop()),Math.max(region.getStart(), region.getStop()),defaultStroke.getLineWidth());
            if(p_high_block.contains(loc)) return region;
          }
        }
      }
    }
    return null;
  }
  public BufferedImage getBuffer(){
    return this.buffer;
  }
}