/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing.canvas;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Shape;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.Collection;

import javax.swing.JPanel;
import javax.swing.JViewport;

import at.icbi.gpviz.GPViz;
import at.icbi.gpviz.data.GPVDataContainer;
import at.icbi.gpviz.data.GPVExon;
import at.icbi.gpviz.data.GPVOptions;
import at.icbi.gpviz.data.GPVTranscript;
import at.icbi.gpviz.data.regions.GPVHighlightedRegion;
import at.icbi.gpviz.drawing.colorscheme.GPVColorScheme;
import at.icbi.gpviz.swing.MainWindow;
import at.icbi.gpviz.weblinks.WebLink;
import at.tugraz.genome.biojava.seq.DNAStrand;

public class GPVCanvasDecoLayer extends JPanel implements MouseListener, MouseMotionListener, GPVCanvasListener
{
  private static final long serialVersionUID = 1L;

  private Point2D lassoStart = null;
  private Point2D lassoEnd = null;

  private boolean draggingWithButton1 = false;

  Rectangle2D lassoRectangle = null;
  int[] lassoNucleotides = null;

  private GPVCanvas canvas;
  private MainWindow mainwindow;

  private Shape debugShape;
  
  private String statuslabel = null;

  public GPVCanvasDecoLayer(GPVCanvas canvas, MainWindow mainwindow)
  {
    super();

    this.canvas = canvas;
    this.mainwindow = mainwindow;
    this.setOpaque(false);
    this.canvas.addMouseListener(this);
    this.canvas.addMouseMotionListener(this);
    this.canvas.addCanvasListener(this);
  }

  @Override
  public void canvasRefreshed()
  {
    clearAll();
  }

  public void clearAll()
  {
    this.lassoStart = null;
    this.lassoRectangle = null;
    this.lassoNucleotides = null;
    this.debugShape = null;
    this.setCursor(Cursor.getDefaultCursor());
  }

  @Override
  public void paint(Graphics g)
  {
    Graphics2D g2 = (Graphics2D) g;
    GPVColorScheme colors = canvas.getProperties().getColorScheme();

    if (debugShape != null)
    {
      g2.setColor(Color.red);
      g2.draw(debugShape);
    }

    if (lassoRectangle != null)
    {
      // draw highlighter on gene
      setOpacity(g2, 0.75f);
      g2.setColor(colors.getLassoHighlighterColor());
      g2.fill(lassoRectangle);
      g2.setColor(colors.getBlockBorder());
      g2.draw(lassoRectangle);
      setOpacity(g2, 1f);

      // draw highlighter on protein
      if (lassoNucleotides != null)
      {

        // draw text above highlighter
        int fontsize = 10;
        g2.setFont(new Font("Arial", Font.PLAIN, fontsize));
        String marker_label = lassoNucleotides[0] + "-" + lassoNucleotides[1] + " (" + Math.abs(lassoNucleotides[1] - lassoNucleotides[0]) + " bases)";

        double label_y2 = lassoRectangle.getY();
        double label_y = label_y2 - (fontsize + 5);
        double label_x = lassoRectangle.getX();
        double label_width = g2.getFontMetrics().getStringBounds(marker_label, g2).getWidth() + 4;

        Rectangle2D rect = new Rectangle2D.Double(label_x, label_y, label_width, label_y2 - label_y);
        g2.setColor(new Color(223,223,223));
        setOpacity(g2, 0.90f);
        g2.fill(rect);
        setOpacity(g2, 1f);
        g2.setColor(Color.black);
        g2.draw(rect);
        g2.drawString(marker_label, (int) (label_x + 2), (int) (label_y2 - 3));
      }
    }
    
    if(statuslabel!=null){
      g2.setFont(new Font("Arial", Font.PLAIN, 11));
      Rectangle2D stringbounds = g2.getFontMetrics().getStringBounds(statuslabel, g2);
      double shadow_heigth = stringbounds.getHeight()+2;
      double shadow_width = stringbounds.getWidth()+2;
      JViewport port = mainwindow.getCanvasScrollPanel().getViewport();
      int ymax = port.getHeight()+port.getViewPosition().y;
      int x = port.getViewPosition().x;
      Rectangle2D shadow = new Rectangle2D.Double(x,ymax-shadow_heigth,shadow_width,shadow_heigth);
      setOpacity(g2,0.75f);
      g2.setColor(Color.white);
      g2.fill(shadow);
      g2.setColor(Color.black);
      g2.drawString(statuslabel, x+1,(int)(ymax-1-g2.getFontMetrics().getDescent()));
    }
  }

  private void setOpacity(Graphics2D g2, float op)
  {
    g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, op));
  }

  private int[] getNucleotidePositionFromLasso()
  {
    if (lassoRectangle != null && canvas.map != null && canvas.selectedGene != null)
    {
      int nt1 = canvas.map.convertXPixelToNucleotidePosition((int) lassoRectangle.getX());
      int nt2 = canvas.map.convertXPixelToNucleotidePosition((int) lassoRectangle.getMaxX());

      int realnt1 = nt1;
      int realnt2 = nt2;
      if (canvas.selectedGene.getStrand() == DNAStrand.REVERSE)
      {
        realnt1 = nt2;
        realnt2 = nt1;
      }

      if (realnt1 == 0)
        realnt1 = Math.min(canvas.selectedGene.getFirstKnownPosition(), canvas.selectedGene.getLastKnownPosition());
      if (realnt2 == 0)
        realnt2 = Math.max(canvas.selectedGene.getFirstKnownPosition(), canvas.selectedGene.getLastKnownPosition());

      return new int[] { realnt1, realnt2 };
    }
    return null;
  }

  public void transformLassoRegionIntoHighlightedRegion()
  {
    if (lassoNucleotides != null)
    {
      GPVDataContainer data = GPViz.getInstance().getDataContainer();
      data.addRegionToHighlight(canvas.getSelectedGene().getChromosome(), lassoNucleotides[0], lassoNucleotides[1], canvas.getSelectedGene().getStrand());
      mainwindow.getDisplayPropertiesPanel().refreshRegionSetsPanel(mainwindow);
    }
  }

  public boolean selectExonIfHoveringOverOne(Point point)
  {
    GPVExon selected_exon = canvas.map.getExonAt((int) point.getX(), (int) point.getY());
    if (selected_exon != null)
    {
      // highlight a whole exon
      double x = canvas.map.transformNTToAbsoluteHorizontalPosition(selected_exon.getStartPos());
      double x2 = canvas.map.transformNTToAbsoluteHorizontalPosition(selected_exon.getEndPos());
      int transcript_nr = canvas.map.getTranscriptThatPixelIsIn((int) point.getX(), (int) point.getY());
      Rectangle2D parent_block = canvas.map.getTranscriptBlock(transcript_nr);
      double y = parent_block.getY();
      double height = parent_block.getHeight();

      lassoRectangle = new Rectangle2D.Double(Math.min(x, x2), y, Math.abs(x2 - x), height);
      Rectangle2D.intersect(lassoRectangle, parent_block, lassoRectangle); // trim the lasso in case it was dragged too far left/right
      lassoNucleotides = new int[] { selected_exon.getStartPos(), selected_exon.getEndPos() };
      lassoStart = new Point2D.Double(Math.min(x, x2), y);
      lassoEnd = new Point2D.Double(Math.max(x, x2), y + height);
      return true;
    }
    return false;
  }

  public boolean selectAsPointHighlighter(Point point)
  {
    int transcript = canvas.map.getTranscriptThatPixelIsIn((int) point.getX(), (int) point.getY());
    boolean intophighlighterarea = canvas.map.getContinuousRegionHighlighterBlock().contains(point);
    boolean intoppointhighlighterarea = canvas.map.getPointRegionHighlighterBlock().contains(point);
    Rectangle2D parent_block;
    if (transcript >= 0 || intophighlighterarea || intoppointhighlighterarea)
    {
      if (intophighlighterarea)
      {
        parent_block = canvas.map.getContinuousRegionHighlighterBlock();
      } else if (intoppointhighlighterarea)
      {
        parent_block = canvas.map.getPointRegionHighlighterBlock();
      } else
      {
        parent_block = canvas.map.getTranscriptBlock(transcript);
      }

      lassoRectangle = new Rectangle2D.Double(point.getX(), parent_block.getY(), 1, parent_block.getHeight());
      lassoStart = new Point2D.Double(lassoRectangle.getX(), lassoRectangle.getY());
      int nt = canvas.map.convertXPixelToNucleotidePosition((int) lassoRectangle.getX());
      lassoNucleotides = new int[] { nt, nt };

      return true;
    }

    return false;
  }

  public boolean extendLassoToHere(Point point)
  {
    if (lassoRectangle != null)
    {
      double start_x = lassoRectangle.getX();
      lassoRectangle = new Rectangle2D.Double(Math.min(point.getX(), start_x), lassoRectangle.getY(), Math.abs(start_x - point.getX()),
          lassoRectangle.getHeight());
      lassoNucleotides = getNucleotidePositionFromLasso();
      return true;
    }
    return false;
  }

  private String getWebLinkAt(Point2D point)
  {
    // this could be made faster if we stored the link blocks at some point, and didn't have to reconstruct them all every time the mouse moves..
    // so far it hasn't been a performance issue, though..
    GPVOptions opts = GPViz.getInstance().getOptions();

    if (canvas.getGeneTrackLabelBlock().contains(point))
    {
      return WebLink.resolveGeneLink(opts.getGeneInfoLinkBase(), canvas.selectedGene);
    }
    int ti = 1;
    int pi = 1;
    for (GPVTranscript transcript:canvas.selectedGene.getEnabledTranscriptsSorted())
    {
      if (canvas.getTranscriptTrackLabelBlock(ti).contains(point))
      {
        return WebLink.resolveTranscriptLink(opts.getTranscriptInfoLinkBase(), transcript);
      }
      if (transcript.isCoding())
      {
        Rectangle2D block = canvas.getProteinTrackLabelBlock(pi);
        if (block != null && block.contains(point))
        {
          return WebLink.resolveProteinLink(opts.getProteinInfoLinkBase(), transcript.getResultingProtein());
        }
        pi++;
      }
      ti++;
    }
    
    GPVDataContainer data = GPViz.getInstance().getDataContainer();
    Collection<GPVHighlightedRegion> regions = data.getHighlightedRegionsForGene(canvas.selectedGene, true);
    for(GPVHighlightedRegion region:regions){
      if(canvas.map.getContinuousRegionHighlighter(region.getStart(),region.getStop()).contains(point)){
        return WebLink.resolveRegionLink(opts.getRegionLinkBase(), region);
      }
    }
    
    return null;
  }

  public void setStatus(String status){
      statuslabel=status;
  }
  public void clearStatus(){
    statuslabel = null;
  }

  @Override
  public void mouseClicked(MouseEvent e)
  {

    boolean nothing_happened = true;

    if (e.getButton() == MouseEvent.BUTTON1)
    {
      if (canvas.map != null && canvas.selectedGene != null)
      {
        String link_im_hovering_over = getWebLinkAt(e.getPoint());
        if (link_im_hovering_over != null)
        {
          WebLink.visitLink(link_im_hovering_over);
        } else
        {
          boolean intranscript = canvas.map.getTranscriptThatPixelIsIn(e.getX(), e.getY()) >= 0;
          boolean intopRegionhighlighterarea = canvas.map.getContinuousRegionHighlighterBlock().contains(e.getPoint());
          if (intranscript || intopRegionhighlighterarea)
          {
            if (lassoRectangle != null && lassoRectangle.contains(e.getPoint()))
            {
              // in the lasso
              nothing_happened = false;
              if (e.getClickCount() == 2)
              {
                transformLassoRegionIntoHighlightedRegion();
              }
            } else if (intranscript && (e.getModifiers() - InputEvent.BUTTON1_MASK == InputEvent.CTRL_MASK))
            {
              // Ctrl has been pressed while clicking. we want to select an point
              nothing_happened = !selectAsPointHighlighter(e.getPoint());
            } else if (e.getModifiers() - InputEvent.BUTTON1_MASK == InputEvent.SHIFT_MASK)
            {
              // shift has been pressed. extend the lasso
              nothing_happened = !extendLassoToHere(e.getPoint());
            } else
            {
              // no modifier, select an exon if possible
              nothing_happened = !selectExonIfHoveringOverOne(e.getPoint());
            }
          } else
          {
            if (canvas.map.getPointRegionHighlighterBlock().contains(e.getPoint()))
            {
              nothing_happened = false;
              selectAsPointHighlighter(e.getPoint());
              if (e.getClickCount() == 2)
              {
                transformLassoRegionIntoHighlightedRegion();
              }
            }
          }
        }
      }

      if (nothing_happened)
        clearAll();
      repaint();
    }
  }

  @Override
  public void mousePressed(MouseEvent e)
  {
    if (e.getButton() == MouseEvent.BUTTON1)
    {
      lassoStart = new Point2D.Float(e.getX(), e.getY());
      draggingWithButton1 = true;
    }
  }

  @Override
  public void mouseReleased(MouseEvent e)
  {
    // lassoStart = null;
    draggingWithButton1 = false;
    repaint();
  }

  @Override
  public void mouseEntered(MouseEvent e)
  {

  }

  @Override
  public void mouseExited(MouseEvent e)
  {
  }

  @Override
  public void mouseDragged(MouseEvent e)
  {
    boolean nothing_happened = true;

    if (draggingWithButton1)
    { // i have to check if we are dragging properly, since e.getButton always returns 0
      lassoEnd = new Point2D.Float(e.getX(), e.getY());

      if (canvas.map != null && canvas.selectedGene != null)
      {
        if (lassoStart != null && lassoEnd != null)
        {
          boolean intophighlighterarea = canvas.map.getContinuousRegionHighlighterBlock().contains((int) lassoStart.getX(), (int) lassoStart.getY());
          int transcript_nr = canvas.map.getTranscriptThatPixelIsIn((int) lassoStart.getX(), (int) lassoStart.getY());
          if (transcript_nr >= 0 || intophighlighterarea)
          {
            // we are drawing a lasso in a transcript/gene area
            Rectangle2D parent_block;
            if (intophighlighterarea)
            {
              parent_block = canvas.map.getContinuousRegionHighlighterBlock();
            } else
            {
              parent_block = canvas.map.getTranscriptBlock(transcript_nr);
            }
            double x = Math.min(lassoStart.getX(), lassoEnd.getX());
            double y = parent_block.getY();
            double height = parent_block.getHeight();
            double width = Math.abs(lassoStart.getX() - lassoEnd.getX());

            lassoRectangle = new Rectangle2D.Double(x, y, width, height);
            Rectangle2D.intersect(lassoRectangle, parent_block, lassoRectangle); // trim the lasso in case it was dragged too far left/right
            lassoNucleotides = getNucleotidePositionFromLasso();
            nothing_happened = false;
          }
        }
      }
    }

    if (nothing_happened)
      clearAll();
    repaint();
  }

  @Override
  public void mouseMoved(MouseEvent e)
  {
    if (canvas.map != null && canvas.selectedGene != null)
    {
      String link_im_hovering_over = getWebLinkAt(e.getPoint());
      if (link_im_hovering_over != null)
      {
        this.setCursor(new Cursor(Cursor.HAND_CURSOR));
      } else
      {
        if (canvas.map.getTranscriptThatPixelIsIn(e.getX(), e.getY()) >= 0)
        {
          this.setCursor(new Cursor(Cursor.TEXT_CURSOR));
        } else if (canvas.map.getContinuousRegionHighlighterBlock().contains(e.getPoint())
            || canvas.map.getPointRegionHighlighterBlock().contains(e.getPoint()))
        {
          this.setCursor(new Cursor(Cursor.TEXT_CURSOR));
        } else
        {
          this.setCursor(Cursor.getDefaultCursor());
        }
      }
    } else
    {
      this.setCursor(Cursor.getDefaultCursor());
    }
  }

}
