/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing.canvas;
import java.awt.Shape;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.TreeSet;

import at.icbi.gpviz.data.GPVExon;
import at.icbi.gpviz.data.GPVGene;
import at.icbi.gpviz.data.GPVTranscript;
import at.icbi.gpviz.data.regions.GPVVariant;
import at.icbi.gpviz.data.util.GPVUtils;
import at.icbi.gpviz.drawing.variantshapes.GPVVariantShape;
import at.icbi.gpviz.drawing.variantshapes.GPVVariantShapeScheme;

/**
 * Used to document and find the way on the canvas.
 */
public class GPVCanvasMap
{
  //number of exons and transcripts in the selected gene
  private int numberOfTranscripts = 0;
  private int numberOfCodingTranscripts = 0;
  private int numberOfExons = 0;
  
  private int numberOfCodingExons = 0;
  
  /**
   * Here follows the information about how ranges are scaled. Each array has the same size.
   * Even numbers in the array are exon ranges, odd number are intron ranges. The first range
   * usually has zero lenght, as there is no leading intron displayed
   */
  //the real nucleotide start positions of the ranges (positions on the chromosome)
  private int[] exonIntronRangeStart;
  //the real nucleotide end positions of the ranges (positions on the chromosome)
  private int[] exonIntronRangeStop;
  //the drawing start positions of the ranges as a fraction of the image width
  private float[] exonIntronRangeImageOffset;
  //the ratio used to transform coordinates within the range to their relative horizontal position
  private float[] exonIntronRangeRatio;
  
  /**
   * For proteins it works similar, but since the coordinates differ depending on the transcript, 
   * the following arrays are 2 dimensional, with the first dimension being the transcript (0 is union transcript)
   */
  //the protein start position in the theoretical union protein
  private int[][] proteinRangeStart;
  //the protein start position in the theoretical union protein
  private int[][] proteinRangeEnd;
  //the drawing start positions of the ranges as a fraction of the image width
  private float[] proteinRangeImageOffset;
  //the ratio used to transform coordinates within the range
  private float[] proteinRangeRatio;
  
  //the start/stop codons of all transcripts. stored for convenience.
  private int[] transcriptStartCodons;
  private int[] transcriptStopCodons;
  
  private int[] totalProteinLength;
  
  private int[] proteinToTranscript;;
  
  private VariantShapeAxis variantShapeAxis;
  
  private float unionTranscriptOffset;

  private int width;
  private int height;
  
  private CanvasProperties properties;
  private GPVVariantShapeScheme variantShapes;
  private GPVGene selectedGene;
  
  private boolean negative_strand;
  
  // internal values
  
  private int exon_pos_offset;
  private int gene_length;
  private float gene_width_drawing_ratio;
  private float borderEastWest;;
  private float exon_blowup_ratio;
  private float intron_blowup_ratio;
  
  private float regular_track_height;
  private float protein_track_height;
  private float protein_track_axis_height;
  private float track_offset;
  
  private float borderNorth;
  
  private float exon_space_giveaway;
  
  private int total_exon_len;
  
  private boolean needsAnyLegend;
  
  public GPVCanvasMap(GPVGene gene, CanvasProperties props, int imageheight, int imagewidth, boolean negative_strand, boolean init_proteins, boolean needs_legends){
    this.properties = props;
    this.width = imagewidth;
    this.height = imageheight;
    this.selectedGene = gene;
    
    this.numberOfTranscripts = selectedGene.getEnabledTranscriptsSorted().size();
    this.numberOfExons = selectedGene.getExons().size();
    
    this.negative_strand = negative_strand;
    
    this.transcriptStartCodons=new int[numberOfTranscripts+1];
    this.transcriptStopCodons=new int[numberOfTranscripts+1];
    
    this.numberOfCodingTranscripts=0;
    for(GPVTranscript trans:selectedGene.getEnabledTranscriptsSorted()){
      if(trans.isCoding()){
        this.numberOfCodingTranscripts++;    
      }
    }
    
    this.needsAnyLegend = needs_legends;
    this.totalProteinLength= new int[numberOfCodingTranscripts+1];

    this.variantShapes = properties.getPreferredVariantShapeScheme();
    
    initializeInternalValues();
    initializeExonRanges();
    if(init_proteins)initializeProteinRanges();
  }

  private void initializeInternalValues()
  {
    exon_pos_offset = selectedGene.getFirstKnownPosition();
    gene_length = selectedGene.getLastKnownPosition()-exon_pos_offset;
    gene_width_drawing_ratio = properties.getImageWidthPercent()/(float)gene_length;
    borderEastWest = (1f-properties.getImageWidthPercent())/2f;
    
    // determine the ratio needed to blow it up to use the desired percentage of space
    float percentSpaceExon = properties.getExonSpace();
    total_exon_len = 0;
    for(GPVExon exon:selectedGene.getExons()){
      int exonLength = exon.getEndPos()-exon.getStartPos();
      total_exon_len+=exonLength;
    }
    
    int union_transcript_len = 0;
    
    this.numberOfCodingExons = 0;
    for(GPVExon exon:selectedGene.getAllExonRegionsSeparately(false,true)){
      this.numberOfCodingExons++;
      union_transcript_len+=Math.abs(exon.getEndPos()-exon.getStartPos())+1;
    }
    
    if(numberOfExons>1){
      exon_blowup_ratio = (float)gene_length*percentSpaceExon/total_exon_len;
      intron_blowup_ratio = (float)gene_length*(1f-percentSpaceExon)/(gene_length-total_exon_len);
    }else{
      //keep the scale if there is only 1 exon
      exon_blowup_ratio = 1f;
      intron_blowup_ratio =1f;
    }
    
    
    float track_gap_percent = properties.getPercentTrackGapRelativeToTrack();
    float total_height = properties.getImageTotalGeneAndProteinSpaceHightPercent();
    float legend_height = properties.getLegendHeightPercent();
    if(!needsAnyLegend && properties.isDrawAnyLegend()){
      //if legend drawing is enabled, but we dont actually need it, we can use this space:
      total_height+=properties.getLegendHeightPercent();
      legend_height = 0f;
    }
    
    float num_regular_tracks = (numberOfTranscripts+3); // all transcripts, gene, unionn transcript and union protein
    float num_prot_tracks = (numberOfCodingTranscripts); //all proteins (NOT union protein)
    
    float number_of_tracks = num_regular_tracks + num_prot_tracks;
    float unionTranscriptGapMultiplier = 3f; //the gap between the union transcript and protein is x times wider
    
    
    float greatest_common_factor = (number_of_tracks-2+(unionTranscriptGapMultiplier*3));
    track_offset = (total_height*track_gap_percent)/greatest_common_factor;
    
    if(track_offset<properties.getMinimumTrackGapRelativeToImageHeight()){
      //ok this is too close. we have to make the tracks slimmer in order to get the minimum gap
      track_offset=properties.getMinimumTrackGapRelativeToImageHeight();
      track_gap_percent = (track_offset * greatest_common_factor)/total_height;
    }
    
    unionTranscriptOffset = track_offset*unionTranscriptGapMultiplier;
    float total_tracks_height = (total_height*(1f-track_gap_percent))-unionTranscriptOffset; //-unionTranscriptOffset because we want an additional gap at the end
    float gene_prot_ratio = properties.getGeneProteinTrackRatio();


    
    this.protein_track_height = (total_tracks_height*gene_prot_ratio)/num_prot_tracks;
    this.regular_track_height = (total_tracks_height*(1f-gene_prot_ratio))/num_regular_tracks;
    
    borderNorth = (1f-total_height-properties.getHeaderHeightPercent()-legend_height-properties.getHighlighterBarHeightPercent()-properties.getAxisHeightPercent())/2f;
    
    
    
    //find how much we have to take away from the protein_track_height to display their axis
    if(properties.isShowProteinXAxis()){
      this.protein_track_axis_height = this.protein_track_height*properties.getProteinXAxisPercentOfProteinBlock();
      this.protein_track_axis_height = Math.min(0.02f,protein_track_axis_height); //make sure it doesnt waste too much space
      this.protein_track_height -= this.protein_track_axis_height;
    }
    
    //determine how much space can be given away for smaller exons by reducing large exons
    float total_space_giveaway = 0f;
    for(GPVExon exon:selectedGene.getExons()){
      float exon_length = (exon.getEndPos()-exon.getStartPos());
      float exon_length_to_image_size_ratio = getExonDrawingRatio();
      float exon_length_drawing = exon_length*exon_length_to_image_size_ratio;
      
      if(exon_length_drawing > properties.getExonSpaceSharingCutoff()){
        total_space_giveaway += exon_length_drawing-properties.getExonSpaceSharingCutoff();
      }
    }
    exon_space_giveaway = total_space_giveaway/selectedGene.getExons().size();
    
    this.totalProteinLength[0]  = GPVUtils.nt2Prot(union_transcript_len)+1;
    int i = 1;
    for(GPVTranscript transcript:selectedGene.getEnabledTranscriptsSorted()){
      if(transcript.isCoding()){
        totalProteinLength[i] = transcript.getResultingProtein().getProteinLength();
        i++;
      }
    }
    
  }
  
  private float getExonDrawingRatio(){
    return  exon_blowup_ratio*gene_width_drawing_ratio;
  }
  
  private float getIntronDrawingRatio(){
    return  intron_blowup_ratio*gene_width_drawing_ratio;
  }
  
  private void initializeExonRanges(){
    exonIntronRangeStart = new int[numberOfExons*2];
    exonIntronRangeStop = new int[numberOfExons*2];
    exonIntronRangeRatio = new float[numberOfExons*2];
    exonIntronRangeImageOffset = new float[numberOfExons*2];
    
    
    float horiz_position = borderEastWest;
    int previous_exon_stop = exon_pos_offset;
    int range = 0;
    
    for(GPVExon exon:selectedGene.getExons()){
      int gap = exon.getStartPos()-previous_exon_stop;
      float gap_to_image_size_ratio = getIntronDrawingRatio();
      float gap_length_drawing = (float)gap*gap_to_image_size_ratio;
      
      float exon_length = (exon.getEndPos()-exon.getStartPos());
      float exon_length_to_image_size_ratio = getExonDrawingRatio();
      float exon_length_drawing = exon_length*exon_length_to_image_size_ratio;

      if(exon_length_drawing>properties.getExonSpaceSharingCutoff()){
        exon_length_drawing = properties.getExonSpaceSharingCutoff();
      }
      exon_length_drawing+=exon_space_giveaway;
      exon_length_to_image_size_ratio = exon_length_drawing/exon_length;
      
      float start = horiz_position+gap_length_drawing;
      float stop = start+exon_length_drawing;
      
      //now we found two ranges. we found the range of the exon, and the range of the previous gap
      //first define the intron (gap) range
      exonIntronRangeStart[range]= previous_exon_stop;
      exonIntronRangeStop[range]= exon.getStartPos();
      exonIntronRangeRatio[range] = gap_to_image_size_ratio;
      exonIntronRangeImageOffset[range] = horiz_position;

      range++;
      
      //and now define the exon range
      exonIntronRangeStart[range] = exon.getStartPos();
      exonIntronRangeStop[range] = exon.getEndPos();
      exonIntronRangeRatio[range] = exon_length_to_image_size_ratio;
      exonIntronRangeImageOffset[range] = start;
      
      range++;
      
      horiz_position=stop;
      
      previous_exon_stop = exon.getEndPos();
    }
  }
  
  private void initializeProteinRanges(){
    
    proteinRangeStart = new int[numberOfCodingTranscripts+1][numberOfCodingExons];
    proteinRangeEnd = new int[numberOfCodingTranscripts+1][numberOfCodingExons];
    proteinRangeRatio = new float[numberOfCodingExons];
    proteinRangeImageOffset = new float[numberOfCodingExons];
    proteinToTranscript = new int[numberOfCodingTranscripts+1];
    
    int first_start_codon = selectedGene.getEarliestLowerTranscriptionRegionBoundary();
    int first_stop_codon = selectedGene.getLatestUpperTranscriptionRegionBoundary();
    
    float total_coding_exon_width = 0f;
    
    for(GPVExon exon:selectedGene.getAllExonRegionsSeparately(false,true)){
      float nt_drawing_pos1 = transformNTToRelativeHorizontalPosition(exon.getStartPos());
      float nt_drawing_pos2 = transformNTToRelativeHorizontalPosition(exon.getEndPos());
      float exon_block_width = Math.abs(nt_drawing_pos2-nt_drawing_pos1);
      total_coding_exon_width+=exon_block_width;
    }
    
    //iterate over transcripts
    Iterator<GPVTranscript> transcript_it = selectedGene.getEnabledTranscriptsSorted().iterator();

    int protvar_i = 0;
    for(int i = 0; i<numberOfTranscripts+1&& protvar_i<numberOfCodingTranscripts+1;i++){
      GPVTranscript transcript = null;
      float horiz_position = borderEastWest;
      int start_codon = first_start_codon;
      int stop_codon = first_stop_codon;
      GPVExon[] exons;
      if(i==0){
        //union transcript
        exons = selectedGene.getAllExonRegionsSeparately(false,true).toArray(new GPVExon[0]);
        
        //first, find out how much we are actually losing of the total gene length, due to exons and part of exons not being transcribed
      }else{
        do{
          transcript = transcript_it.next();
          if(!transcript.isCoding()) i++;
        }while(!transcript.isCoding()); //only coding transcripts translate to proteins
       
        proteinToTranscript[protvar_i]=i;         
        exons = transcript.getExonsAlignedToUnion();
        start_codon = transcript.getLowerTranscriptionRegionBoundary();
        stop_codon = transcript.getUpperTranscriptionRegionBoundary();
        //other transcripts start with an offset
        horiz_position = transformPPToRelativeHorizontalPosition(GPVUtils.nt2Prot(start_codon-first_start_codon),0);
      }
      
      
      transcriptStartCodons[i] = start_codon;
      transcriptStopCodons[i] = stop_codon;
      
      int prot_exon_i = 0;
      for(int j = 0; j<exons.length; j++){
        GPVExon exon = exons[j]; 
        if(negative_strand){
          exon = exons[exons.length-j-1];
        }
        //is this a transcript that doesnt contain this exon?
        if(exon!=null){
          //exon position on gene with the start/stop codon cut off
          int start = exon.getStartPos();
          int endpos = exon.getEndPos();

          int prot_startpos;
          int prot_endpos;
          if(negative_strand){
            prot_startpos = transformNTToPP(endpos,transcript);
            prot_endpos = transformNTToPP(start,transcript);
          } else{
            prot_startpos = transformNTToPP(start,transcript);
            prot_endpos = transformNTToPP(endpos,transcript);
          }
          
          proteinRangeStart[protvar_i][prot_exon_i] = prot_startpos;
          
          if(i==0){
            //union transcript
            int protlen = prot_endpos-prot_startpos;

            float nt_drawing_pos1 = transformNTToRelativeHorizontalPosition(start);
            float nt_drawing_pos2 = transformNTToRelativeHorizontalPosition(endpos);
            float nt_drawing_length = Math.abs(nt_drawing_pos2-nt_drawing_pos1);

//            float nt_drawing_percentage = Math.min(1f, nt_drawing_length/(total_coding_exon_width));
            float nt_drawing_percentage = nt_drawing_length/(total_coding_exon_width);
            float pp_drawing_length = properties.getImageWidthPercent()*nt_drawing_percentage;
            
            float protein_drawing_ratio = (pp_drawing_length)/protlen;

            proteinRangeImageOffset[prot_exon_i] = horiz_position;
            proteinRangeRatio[prot_exon_i] = protein_drawing_ratio;
            horiz_position += protlen*proteinRangeRatio[prot_exon_i];
          }
        }else{
          proteinRangeStart[protvar_i][prot_exon_i] = -1;
        }
        prot_exon_i++;
      }
      protvar_i++;
    }
    
    //init proteinRangeEnd
    for(int i = 0; i < proteinRangeEnd.length; i++){
      for(int j = 0; j < proteinRangeEnd[i].length; j++){
        if(proteinRangeStart[i][j]==-1){
          proteinRangeEnd[i][j]=-1;
        }else{
        
          int cor_start = j+1;
          for(; cor_start < proteinRangeStart[i].length && proteinRangeStart[i][cor_start]==-1; cor_start++);
        
          if(cor_start<proteinRangeStart[i].length &&proteinRangeStart[i][cor_start]>-1){
            proteinRangeEnd[i][j] = proteinRangeStart[i][cor_start]-1;
          }else{
            proteinRangeEnd[i][j] = totalProteinLength[i]-1;
          }
        }
      }
    }
    
    return;
  }
  
  public void setVariantShapes(GPVVariantShapeScheme variantShapes)
  {
    this.variantShapes = variantShapes;
  }

  public void initVariantShapeDistribution(TreeSet<GPVVariant> variants){
    TreeMap<GPVVariant,Float> init_positions = new TreeMap<GPVVariant,Float>();
    for(GPVVariant var:variants){
      Float pos = transformNTToRelativeHorizontalPosition(var.getStart());
      init_positions.put(var,pos);
    }
    
    variantShapeAxis = new VariantShapeAxis(init_positions,borderEastWest,properties.getImageWidthPercent(), 0.005f, negative_strand);
    
  }
  
  public Rectangle2D getAvailableSpaceForLegend(){
    float x = wscale(borderEastWest);
    float width = wscale((1f-borderEastWest*2f));
    float y = hscale(getFirstYInGeneArea()+properties.getImageTotalGeneAndProteinSpaceHightPercent()+track_offset*2f);
    
    float height = hscale(properties.getLegendHeightPercent()-track_offset);
    
    
    return new Rectangle2D.Float(x,y,width,height);
  }
  
  private int transformNTToPPOnUnionTranscript(int nt){
    int ret = transformNTToPP(nt,null); 
    return ret;
  }
  /**
   * Calculates the protein position based on the nucleotide position
   * Already considers subtraction of the start codon!
   */
  private int transformNTToPP(int nt, GPVTranscript transcript){
    Collection<GPVExon> exons;
    if(transcript!=null){
      exons = transcript.getAllExonRegionsSeparately(false);
    }else{
      exons = selectedGene.getAllExonRegionsSeparately(false,true);
    }
    return GPVUtils.transformNTToPP(nt, exons, negative_strand);
  }

  public int transformNTToAbsoluteHorizontalPosition(int nucleotidePosition){
    return (int)wscale(transformNTToRelativeHorizontalPosition(nucleotidePosition));
  }
  
  private float transformNTToRelativeHorizontalPosition(int nucleotidePosition){
    for(int i = 0; i < exonIntronRangeStart.length; i++){
      if(nucleotidePosition>=exonIntronRangeStart[i] && nucleotidePosition<=exonIntronRangeStop[i]){
        //we found our range.
        int in_range_position = nucleotidePosition-exonIntronRangeStart[i];
        float ret = exonIntronRangeImageOffset[i]+(in_range_position*exonIntronRangeRatio[i]);
        return flipXIfRev(ret);
      }
    }
//    System.err.println("Unknown nucleotide position. Not in a defined range! "+ nucleotidePosition+ " not in " +selectedGene.getFirstKnownPosition() + "-"+selectedGene.getLastKnownPosition());
    return 0f;
  }
  
  public int transformNTToAbsoluteHorizontalPositionOnTheUnionProtein(int nucleotidePosition){
    int pp;
    if(negative_strand){
      pp = transformNTToPPOnUnionTranscript(nucleotidePosition) ;
    }else{
      pp = transformNTToPPOnUnionTranscript(nucleotidePosition);
    }
    
    return transformPPToAbsoluteHorizontalPosition(pp, 0);
  }
  
  public Shape getExon(GPVExon exon, int transcript){
    return getExon(exon.getStartPos(),exon.getEndPos(),transcript);
  }
  public Shape getExon(int startpos, int endpos, int transcript){
    
    float x1 = transformNTToRelativeHorizontalPosition(startpos);
    float x2 = transformNTToRelativeHorizontalPosition(endpos);
    
    float width = Math.abs(x2-x1);
    
    float y = getFirstYInGeneArea() + (transcript*(regular_track_height+track_offset));
    
    int alt_transcript_nr = transcript;
    if(transcript==numberOfTranscripts+1){
      // this is the union transcript
      y+=unionTranscriptOffset;
      alt_transcript_nr  = 0;
    }
    
    float abs_x = wscale(Math.min(x1,x2));
    float abs_y = hscale(y);
    float abs_width = wscale(width);
    float abs_height = hscale(regular_track_height);
    
    Shape ret;
    if(transcript>0 && transcript <= numberOfTranscripts+1 && (startpos<this.transcriptStartCodons[alt_transcript_nr] || endpos > this.transcriptStopCodons[alt_transcript_nr])){
      // we need to warp this
      double narrowing_rate = properties.getUntranslatedTranscriptRegionNarrowing();
      int startTranslatedRegion = transformNTToAbsoluteHorizontalPosition(this.transcriptStartCodons[alt_transcript_nr]);
      int stopTranslatedRegion = transformNTToAbsoluteHorizontalPosition(this.transcriptStopCodons[alt_transcript_nr]);
//        ret = new Rectangle2D.Float(abs_x,abs_y+(float)(abs_height*(narrowing_rate/2f)), abs_width, (float)(narrowing_rate*(abs_height)));
      ret = new NarrowingTranscriptBox((int)abs_x,(int)abs_y,(int)abs_width,(int)abs_height, Math.min(startTranslatedRegion,stopTranslatedRegion), Math.max(stopTranslatedRegion,startTranslatedRegion), narrowing_rate);
    }else{
      ret = new Rectangle2D.Float(abs_x,abs_y, abs_width, abs_height);
    }
    return ret;
  }
  
  public int transformPPToAbsoluteHorizontalPosition(int proteinPosition, int protein_variation){
    return (int)wscale(transformPPToRelativeHorizontalPosition(proteinPosition,protein_variation));
  }
  
  private float transformPPToRelativeHorizontalPosition(int proteinPosition, int protein_variation){
    int region = getProteinRegionFromPP(proteinPosition, protein_variation);
    if(region>-1){
      //we found our range.
      int in_range_position = proteinPosition-proteinRangeStart[protein_variation][region];
      float ret = proteinRangeImageOffset[region]+(in_range_position*proteinRangeRatio[region]);
      return ret;
    }
    System.err.println("Unknown protein position. Not in a defined range! "+ proteinPosition);
    return 0f;
  }
  
  public Rectangle2D getProteinRegionBlockForExon(GPVExon exon, int protein_index){
    int nt_start = exon.getStartPos();
    int nt_stop = exon.getEndPos();
    if(nt_stop<=transcriptStartCodons[proteinToTranscript[protein_index]] || nt_start>=transcriptStopCodons[proteinToTranscript[protein_index]])return null;
    
    int prot_start = transformNTToPPOnUnionTranscript(nt_start);
    int prot_end = transformNTToPPOnUnionTranscript(nt_stop);
    if(negative_strand) {
      prot_start = transformNTToPPOnUnionTranscript(nt_stop);
      prot_end = transformNTToPPOnUnionTranscript(nt_start);
    }
    
    float x1 = transformPPToRelativeHorizontalPosition(prot_start,0);
    float x2 = transformPPToRelativeHorizontalPosition(prot_end,0);
    
    Rectangle2D parentblock = getProteinVariantBlock(protein_index);
    float abs_y1 = (float)parentblock.getY();
    float abs_height = (float)parentblock.getHeight(); 
    
    return new Rectangle2D.Float(wscale(x1), abs_y1, wscale(x2)-wscale(x1), abs_height);
  }
  
  public Rectangle2D[] getProteinRangeAsSeparateBlock(int pp_start, int pp_end, int variation_number){
    int start_region = getProteinRegionFromPP(pp_start, variation_number);
    int end_region = getProteinRegionFromPP(pp_end, variation_number);
    
    ArrayList<Rectangle2D> retlist = new ArrayList<Rectangle2D>();
    
    int y = (int)getProteinVariantBlock(variation_number).getY();
    int height = getProteinTrackHeight();
    
    int x1 = transformPPToAbsoluteHorizontalPosition(pp_start, variation_number);
    for(int i = start_region; i < end_region; i++){
      if(proteinRangeStart[variation_number][i]==-1 && x1>-1){
        //there is an empty space
        int x2 = transformPPToAbsoluteHorizontalPosition(proteinRangeEnd[variation_number][i-1], variation_number);
        retlist.add(new Rectangle2D.Float(x1,y, x2-x1, height));
        x1=-1;
      }else if(x1 == -1 && proteinRangeStart[variation_number][i]>-1){
        x1 = transformPPToAbsoluteHorizontalPosition(proteinRangeStart[variation_number][i],variation_number);
      }
    }
    
    if(x1==-1){
      x1 = transformPPToAbsoluteHorizontalPosition(proteinRangeStart[variation_number][end_region],variation_number);
    }
    int x2 = transformPPToAbsoluteHorizontalPosition(pp_end, variation_number);
    retlist.add(new Rectangle2D.Float(x1,y, x2-x1, height));
    
    return retlist.toArray(new Rectangle2D[0]);
  }
  
  /**
   * get the positions of merge site lines. these are lines that highlight were exons from different transcripts 
   * have been merged for the union display
   */
  public Line2D[] getMergeSiteLines(GPVExon exon){
    Line2D[] lines = new Line2D[exon.getMergeSites().length];
    int i = 0;
    for(int mergeSite:exon.getMergeSites()){
      float x = transformNTToRelativeHorizontalPosition(mergeSite);
      float y1 = getFirstYInGeneArea()+regular_track_height*0.05f;
      float y2 = getFirstYInGeneArea()+regular_track_height*0.95f;
      Line2D.Float line = new Line2D.Float(wscale(x),hscale(y1),wscale(x),hscale(y2));
      lines[i] = line;
      i++;
    }
    return lines;
  }
  
  /**
   * get position of the exon.
   * @param exon the exon number, starting with 1
   * @param transcript
   * @return
   */
  public Rectangle2D getExonOnGene(int exon){
    
    float x1 = flipXIfRev(exonIntronRangeImageOffset[(exon*2)-1]);
    float x2;
    if(exon==numberOfExons){
      x2 = transformNTToRelativeHorizontalPosition(selectedGene.getLastKnownPosition());
    }else{
      x2 = flipXIfRev(exonIntronRangeImageOffset[(exon*2)]);
    }
      
    float width = Math.abs(x2-x1);
    float y = getFirstYInGeneArea();
    return new Rectangle2D.Float(wscale(Math.min(x1,x2)),hscale(y),wscale(width),hscale(regular_track_height));
  }
  
  public Line2D getTranscriptRail(int transcript){
    float y = getFirstYInGeneArea() + (transcript*(regular_track_height+track_offset)) + (regular_track_height/2f);
    float x1 = borderEastWest;
    float x2 = 1f-borderEastWest;
    Line2D.Float ret = new Line2D.Float(wscale(x1),hscale(y),wscale(x2),hscale(y));
    return ret;
  }
  
  public Line2D getProteinVariantRail(int transcript){
    Rectangle2D block = getProteinVariantBlock(transcript);
    Line2D ret = new Line2D.Double(block.getX(),block.getCenterY(),block.getMaxX(),block.getCenterY());
    return ret;
  }
  
  
  public Line2D getUnionTranscriptRail(){
    Line2D ret = getTranscriptRail(numberOfTranscripts+1);
    //double the track offset for this one
    ret.setLine(ret.getX1(), ret.getY1()+hscale(unionTranscriptOffset), ret.getX2(), ret.getY2()+hscale(unionTranscriptOffset));
    return ret;
  }

  /**
   * this may not even be drawn, but is used for location calculation
   */
  public Rectangle2D getTranscriptBlock(int transcript){
    float y1 = getFirstYInGeneArea() + (transcript*(regular_track_height+track_offset));
    if(transcript==numberOfTranscripts+1){
      //this is the union transcript
      y1+=unionTranscriptOffset;
      
    }
    Rectangle2D.Float ret = new Rectangle2D.Float(wscale(borderEastWest),hscale(y1),wscale(properties.getImageWidthPercent()), hscale(regular_track_height));
    return ret;
  }
  
  public Rectangle2D getUnionTranscriptMarkerBlock(int nt_start, int nt_end, float stroke_width){
    int union_transcript_y = (int)(getTranscriptBlock(numberOfTranscripts+1).getY()+stroke_width);
    int union_transcript_marker_height = (int) (Math.ceil(hscale(regular_track_height)-stroke_width));
    int x1 = transformNTToAbsoluteHorizontalPosition(nt_start);
    int x2 = transformNTToAbsoluteHorizontalPosition(nt_end);
    int len = Math.max(1, Math.abs(x2 - x1));
    return new Rectangle2D.Float(Math.min(x1, x2), union_transcript_y, len, union_transcript_marker_height);
  }
  
  public Rectangle2D getUnionProteinHighlighterBlock(int nt_start, int nt_end, float stroke_width){
    int union_protein_y = (int) (getProteinVariantBlock(0).getY() + stroke_width);
    int union_protein_marker_height = (int) (Math.ceil(hscale(regular_track_height) - stroke_width));
    int prot_x1 = transformNTToAbsoluteHorizontalPositionOnTheUnionProtein(nt_start);
    int prot_x2 = transformNTToAbsoluteHorizontalPositionOnTheUnionProtein(nt_end);
    int len = Math.max(1, Math.abs(prot_x2 - prot_x1));
    return new Rectangle2D.Float(Math.min(prot_x1, prot_x2), union_protein_y, len, union_protein_marker_height);
  }
  
  public Rectangle2D getContinuousRegionHighlighterBlock(){
    int x = (int)wscale(borderEastWest);
    int width = (int)wscale(properties.getImageWidthPercent());
    float padding = properties.getHighlighterBarHeightPercent()*0.10f;
    int y = (int)hscale(borderNorth+properties.getHeaderHeightPercent()+properties.getAxisHeightPercent());
    int height = (int) hscale((properties.getHighlighterBarHeightPercent()-(padding*2f))/2f);
    return new Rectangle2D.Float(x,y,width,height);
  }
  
  public Rectangle2D getPointRegionHighlighterBlock(){
    Rectangle2D block1 = getContinuousRegionHighlighterBlock();
    float padding = properties.getHighlighterBarHeightPercent()*0.10f;
    double y = block1.getMaxY()+hscale(padding);
    return new Rectangle2D.Double(block1.getX(),y,block1.getWidth(),block1.getHeight());
  }
  
  public Rectangle2D getContinuousRegionHighlighter(int nt1, int nt2){
    float x1 = transformNTToRelativeHorizontalPosition(nt1);
    float x2 = transformNTToRelativeHorizontalPosition(nt2);
    float width = Math.abs(x2 - x1);
    Rectangle2D parent_block = getContinuousRegionHighlighterBlock();
    return new Rectangle2D.Float(wscale(Math.min(x1,x2)),(float)parent_block.getY(),wscale(width),(float)parent_block.getHeight());
  }
  
  public GPVVariantShape getVariantHighlighter(GPVVariant variant){
      Rectangle2D parent_block = getPointRegionHighlighterBlock();
      float center_x = variantShapeAxis.getHorizontalPositionForVariant(variant);
      Point2D anchor = new Point2D.Float(wscale(center_x),(float)parent_block.getY());
      
      GPVVariantShape todraw = variantShapes.getShapeFor(variant);
      todraw.init(anchor, parent_block.getHeight()/2d, parent_block.getHeight()/2d);
      
      return todraw;
  }
  
  public Point2D getVariantHighlighterLineDock(GPVVariant variant){
    float center_x = transformNTToRelativeHorizontalPosition(variant.getStart());
    Rectangle2D parent_block = getPointRegionHighlighterBlock();
    return new Point2D.Double(wscale(center_x), parent_block.getMaxY());
  }
  
  public Rectangle2D getXAxisBlock(){
    int x = (int)wscale(borderEastWest);
    int width = (int)wscale(properties.getImageWidthPercent());
    float padding = properties.getAxisHeightPercent()*0.2f;
    int y = (int)hscale(borderNorth+properties.getHeaderHeightPercent()+padding);
    int height = (int) hscale(properties.getAxisHeightPercent()-(padding*2f));
    return new Rectangle2D.Float(x,y,width,height);
  }
  
  public Rectangle2D getGeneBlock(){
    Rectangle2D ret = new Rectangle2D.Float(wscale(borderEastWest), hscale(getFirstYInGeneArea()), wscale(properties.getImageWidthPercent()), hscale(regular_track_height));
    return ret;
  }

  public Point2D getGeneLabelLocation(){
    Rectangle2D block = getGeneBlock();
    Point2D label_loc = new Point2D.Double(block.getX()+wscale(0.01f),block.getMaxY()-hscale(regular_track_height*0.1f));
    return label_loc;
  }
  
  public Point2D getTranscriptLabelLocation(int transcript_index){
    Rectangle2D block = getTranscriptBlock(transcript_index);
    Point2D label_loc = new Point2D.Double(block.getX()+wscale(0.01f),block.getMaxY()-hscale(regular_track_height*0.1f));
    return label_loc;
  }
  
  public Point2D getProteinLabelLocation(int transcript_index){
    boolean union = (transcript_index==0);
    Rectangle2D block = getProteinVariantBlock(transcript_index);
    Point2D label_loc = new Point2D.Double(block.getX()+wscale(0.01f),block.getMaxY()-hscale((union?regular_track_height:protein_track_height)*0.1f));
    return label_loc;
  }
  
  public Rectangle2D getUnionProteinBlock(){
    float y = getFirstYInProteinArea()+protein_track_axis_height;
    Rectangle2D ret = new Rectangle2D.Float(wscale(borderEastWest), hscale(y), wscale(properties.getImageWidthPercent()), hscale(regular_track_height));
    return ret;
  }
  
  public Rectangle2D getProteinVariantBlock(int transcript){
    if(transcript == 0)return getUnionProteinBlock();
    float abs_y = (float)getUnionProteinBlock().getMaxY()+hscale(protein_track_axis_height)+hscale(unionTranscriptOffset);
    abs_y += hscale((transcript-1)*(getProteinTrackHeightWithAxis()+track_offset));
    Rectangle2D ret = new Rectangle2D.Float(wscale(borderEastWest), abs_y, wscale(properties.getImageWidthPercent()), hscale(protein_track_height));
    return ret;
  }
  
  public Rectangle2D getProteinVariantAxisBlock(int transcript){
    float abs_y2 = (float)getProteinVariantBlock(transcript).getY();
    float y = abs_y2-hscale(protein_track_axis_height);
    Rectangle2D ret = new Rectangle2D.Float(wscale(borderEastWest), y, wscale(properties.getImageWidthPercent()), hscale(protein_track_axis_height));
    return ret;
  }
  
  public float getFirstYInGeneArea(){
    float ret = borderNorth+properties.getHeaderHeightPercent()+properties.getAxisHeightPercent()+properties.getHighlighterBarHeightPercent();
    return ret;
  }
  
  
  public float getFirstYInProteinArea(){
    Rectangle2D uniontranscript = getTranscriptBlock(numberOfTranscripts+1);
    float continuous_y = (float)(uniontranscript.getMaxY()/height)+unionTranscriptOffset+track_offset;
    return continuous_y;
  }
  
  public int convertXPixelToNucleotidePosition(int pixel){
    int region = getRegionFromXPixel(pixel);
    if(region>-1){
    //get the relative location
      float relativePixel = flipXIfRev((float)pixel/(float)width);
      float relativePixelIntoTheRegion = relativePixel - exonIntronRangeImageOffset[region];
      //now calculate the nucleotide position based on the start point of the range and the stored ration for that range
      int nucleotidePoisiton = exonIntronRangeStart[region] + (int)(relativePixelIntoTheRegion/exonIntronRangeRatio[region]);
      return nucleotidePoisiton;
    }
    return 0;
  }
  
  
  public int convertXPixelToProteinPosition(int pixel, int proteinnumber){
    int region = getProteinRegionFromXPixel(pixel);
    if(region>-1){
      if(proteinRangeStart[proteinnumber][region]==-1){
        //this region doesnt exist in this particular transcript
        return 0;
      }
      //get the relative location
      float relativePixel = (float)pixel/(float)width;
      float relativePixelIntoTheRegion = relativePixel - proteinRangeImageOffset[region];
      
      //now calculate the nucleotide position based on the start point of the range and the stored ration for that range
      int proteinPositon = proteinRangeStart[proteinnumber][region] + (int)(relativePixelIntoTheRegion/proteinRangeRatio[region]);
      return proteinPositon+1; //plus one because it's common to start counting at 1, but in the code we start at 0
    }
    return 0;
  }
  
  private int getRegionFromXPixel(int pixel){
    if(pixel < wscale(borderEastWest)){
      //before the gene
      return -1;
    }
    if(pixel > wscale(1-borderEastWest)){
      //after the gene
      return -1;
    }
    
    //get the relative location
    float relativePixel = flipXIfRev((float)pixel/(float)width);
    for(int i = 0; i < exonIntronRangeImageOffset.length; i++){
      if(relativePixel>exonIntronRangeImageOffset[i]){
        if(i == exonIntronRangeImageOffset.length-1 || relativePixel<exonIntronRangeImageOffset[i+1]){
          //found the right region
          return i;
        }
      }
    }
    return -1;
  }
  
  private int getProteinRegionFromPP(int pp, int protein_variation){
    for(int i = 0; i < proteinRangeStart[protein_variation].length; i++){
      
      int next_good_index = -1;
      //find next good value
      for(int j = i+1; j < proteinRangeStart[protein_variation].length; j++){
        if(proteinRangeStart[protein_variation][j]>-1){
          next_good_index = j;
          break;
        }
      }
      
      if((pp>=proteinRangeStart[protein_variation][i] && 
          (next_good_index==-1|| pp<proteinRangeStart[protein_variation][next_good_index] || proteinRangeStart[protein_variation][next_good_index]==-1)))
      {
        return i;
      }
    }
    return -1;
  }
  
  private int getProteinRegionFromXPixel(int pixel){
    if(pixel < wscale(borderEastWest)){
      //before the gene
      return -1;
    }
    if(pixel > wscale(1-borderEastWest)){
      //after the gene
      return -1;
    }
    
    //get the relative location
    float relativePixel = (float)pixel/(float)width;
    for(int i = 0; i < proteinRangeImageOffset.length; i++){
      if(relativePixel>proteinRangeImageOffset[i]){
        if(i == proteinRangeImageOffset.length-1 || relativePixel<proteinRangeImageOffset[i+1]){
          //found the right region
          return i;
        }
      }
    }
    return -1;
  }
  
  public boolean isPixelInGene(int pixel_x, int pixel_y){
    return getGeneBlock().contains((double)pixel_x, (double)pixel_y);
  }
  
  /**
   * returns -1 if gene is not in any gene/transcript area, 0 if pixel is in gene, and the transcript number if pixel is in a transcript
   * @param pixel_x
   * @param pixel_y
   * @return
   */
  public int getTranscriptThatPixelIsIn(int pixel_x, int pixel_y){
    //is it in the gene? if yes, return 0
    if(isPixelInGene(pixel_x, pixel_y)) return 0;
    
    //is it in a transcript? if yes, return the transcript number
    for(int i = 1; i <= numberOfTranscripts+1; i++){
      if(getTranscriptBlock(i).contains((double)pixel_x, (double)pixel_y)) return i;
    }    
    //not found? return -1
    return -1;
  }
  
  public int getProteinVariantThatPixelIsIn(int pixel_x, int pixel_y){
    //is it in the gene? if yes, return 0
    
    if(getProteinVariantBlock(0).contains((double)pixel_x, (double)pixel_y)) {
      return 0;
    }
    
    //is it in a transcript? if yes, return the transcript number
    for(int i = 1; i <= numberOfCodingTranscripts; i++){
      if(getProteinVariantBlock(i).contains((double)pixel_x, (double)pixel_y)){
        return i;
      }
    }    
    //not found? return -1
    return -1;
  }

  
  /**
   * returns 0 if not in the area of the area of the genome
   * @param x
   * @param y
   * @return
   */
  public int getNucleotidePositionOn(int x, int y){
    int transcript = getTranscriptThatPixelIsIn(x,y);
    //nowhere. return 0
    if(transcript==-1 && !getContinuousRegionHighlighterBlock().contains(x,y)) return 0;
    
    int ret = convertXPixelToNucleotidePosition(x);
    return ret;
  }
  
  
  /**
   * returns 0 if not in the area of the area of the genome
   * @param x
   * @param y
   * @return
   */
  public int getProteinPositionOn(int x, int y){
    int variant = getProteinVariantThatPixelIsIn(x,y);
    //nowhere, or not in the protein area. return 0
    if(variant==-1) return 0;
    int ret = convertXPixelToProteinPosition(x,variant);
    return ret;
  }
  
  
  public GPVExon getExonAt(int x, int y){
    int transcript_nr = getTranscriptThatPixelIsIn(x, y);
    if(transcript_nr==-1)return null;
    
    int nt = getNucleotidePositionOn(x, y);
    
    TreeSet<GPVExon> exons = null;
    if(transcript_nr>0){
      if(transcript_nr==numberOfTranscripts+1){
        exons = selectedGene.getAllExonRegionsSeparately(true, true);
      }else{
        int i = 1;
        for(GPVTranscript transcript:selectedGene.getEnabledTranscriptsSorted()){
          if(i++==transcript_nr){
            exons = transcript.getExons();
            break;
          }
        }
      }
    }else{
      exons = selectedGene.getExons();
    }

    if(exons!=null){
      for(GPVExon exon:exons){
        if(nt>=exon.getStartPos() && nt<=exon.getEndPos()){
          return exon;
        }
      }
    }
    
    return null;
  }
  
  private float getProteinTrackHeightWithAxis(){
    return this.protein_track_height+protein_track_axis_height;
  }
  
  public int getGeneTrackHeight(){
    return (int)hscale(regular_track_height);
  }
  
  public int getProteinTrackHeight(){
    return (int)hscale(protein_track_height);
  }
  
  /**
   * Flip if this is on the reverse strand. 
   */
  protected float flipXIfRev(float x, float blockwidth){
    if(!negative_strand)return x;
    return 1f-x-Math.abs(blockwidth);     
  }
  protected float flipXIfRev(float x){
    return flipXIfRev(x,0f);
  }
  
  protected float hscale(float coord){
    return coord*height;
  }
  
  protected float wscale(float coord){
    return coord*width;
  }

  protected float getBorderEastWest()
  {
    return borderEastWest;
  }

  protected float getBorderNorth()
  {
    return borderNorth;
  }
  
}
