/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing.canvas;

import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.NoSuchElementException;

public class NarrowingTranscriptBox implements Shape
{

  private int x;
  private int y;
  private int width;
  private int height;
  private int startTranslatedRegion;
  private int stopTranslatedRegion;
  private boolean rounded = true;
  
  private double narrowing_rate;
  
  private double rounding_radius;
  
  private Rectangle2D[] subBlocks;
  
  private boolean[] hasBlock = new boolean[]{true,true,true};
  
  public NarrowingTranscriptBox(int x, int y, int width, int height, int startTranslatedRegion, int stopTranslatedRegion, double narrowing_rate){
    this.x = x;
    this.y = y;
    this.width = width;
    this.height = height;
    
    int x2 = x+width;

    this.rounding_radius = this.height*narrowing_rate*0.25f;
    
    this.startTranslatedRegion = startTranslatedRegion;
    this.stopTranslatedRegion = stopTranslatedRegion;
    
    if(this.startTranslatedRegion>=x2) {
      hasBlock[1] = false;
      hasBlock[2] = false;
    }
    if(this.stopTranslatedRegion <= x){
      hasBlock[1] = false;
      hasBlock[0] = false;
    }
    if(this.stopTranslatedRegion>=x2) {
      hasBlock[2] = false;
    }
    if(this.startTranslatedRegion<=x) {
      hasBlock[0] = false;
    }
    
    if(this.width==0){
      hasBlock[0]= true;
    }
    
    this.narrowing_rate = narrowing_rate;
    
    this.subBlocks = new Rectangle2D[3];
    for(int i = 0; i < subBlocks.length; i++){
      this.subBlocks[i] = getBlock(i);
    }
    
  }
  
  @Override
  public Rectangle getBounds()
  {
    return new Rectangle(x,y,width,height);
  }

  @Override
  public Rectangle2D getBounds2D()
  {
    return new Rectangle2D.Double(x, y, width, height);
  }
  
  private Rectangle2D getBlock(int nr){
    //defaults for middle block
    double height = this.height;
    double x = Math.max(this.startTranslatedRegion,this.x);
    double x2 = Math.min(this.stopTranslatedRegion, this.x+this.width);
    double y = this.y;
    
    //set height if in narrowed region
    if(nr==0 || nr == 2){
      height = this.height * narrowing_rate;
      double difference = Math.abs(this.height - height);
      y = this.y + (difference/2);
    }
    
    if(nr == 0){
      x = this.x;
      if(this.width==0){
        x2 = x;
      }else{
        x2 = Math.min(this.startTranslatedRegion,this.x+this.width);
      }
    }
    
    if(nr == 2){
      x = Math.max(this.x,this.stopTranslatedRegion);
      x2 = this.x+this.width;
    }
    
    
    double width = x2-x;
    return new Rectangle2D.Double(x,y,width, height);
  }

  @Override
  public boolean contains(double x, double y)
  {
    for(Rectangle2D rect:subBlocks){
      if(rect.contains(x,y))return true;
    }
    return false;
  }

  @Override
  public boolean contains(Point2D p)
  {
    return contains(p.getX(),p.getY());
  }

  @Override
  public boolean intersects(double x, double y, double w, double h)
  {
    for(Rectangle2D rect:subBlocks){
      if(rect.intersects(x,y,w,h))return true;
    }
    return false;
  }

  @Override
  public boolean intersects(Rectangle2D r)
  {
    return intersects(r.getX(),r.getY(),r.getWidth(),r.getHeight());
  }

  @Override
  public boolean contains(double x, double y, double w, double h)
  {
    for(Rectangle2D rect:subBlocks){
      if(rect.contains(x,y,w,h))return true;
    }
    return false;
  }

  @Override
  public boolean contains(Rectangle2D r)
  {
    return contains(r.getX(),r.getY(),r.getWidth(),r.getHeight());
  }

  @Override
  public PathIterator getPathIterator(AffineTransform at)
  {
    return new NarrowingTranscriptBoxPathIterator();
  }

  @Override
  public PathIterator getPathIterator(AffineTransform at, double flatness)
  {
    return new NarrowingTranscriptBoxPathIterator();
  }

  class NarrowingTranscriptBoxPathIterator implements PathIterator{

    ArrayList<double[]> path = new ArrayList<double[]>();
    ArrayList<Integer> segment = new ArrayList<Integer>();
    
    private void addCurve(double[] start, double[] center, double[] stop){
      double[] stop1 = calculateStopBeforeCurve(start, center);
      double[] stop2 = calculateStopBeforeCurve(stop, center);;
      double[] curve = new double[]{stop1[0],stop1[1],center[0],center[1],stop2[0],stop2[1]};
      
      path.add(stop1);
      segment.add(PathIterator.SEG_LINETO);
      path.add(curve);
      segment.add(PathIterator.SEG_CUBICTO);
      path.add(stop2);
      segment.add(PathIterator.SEG_LINETO);
    }
    
    private double[] calculateStopBeforeCurve(double[] start, double center[]){
      double[] stop = new double[2];
      if(start[0]!=center[0]){
        double inc_x = center[0] - start[0];
        double myradius = Math.min(Math.abs(inc_x/2),rounding_radius);
        if(inc_x<0){
          inc_x+=myradius;
        }else{
          inc_x-=myradius;
        }
        stop[0] = start[0]+inc_x;
      }else{
        stop[0] = start[0];
      }
      if(start[1] != center[1]){
        double inc_y = center[1] - start[1];
        double myradius = Math.min(Math.abs(inc_y/2),rounding_radius);
        if(inc_y<0){
          inc_y+=myradius;
        }else{
          inc_y-=myradius;
        }
        stop[1] = start[1]+inc_y;
      }else{
        stop[1] = start[1];
      }
      return stop;
    }
    
    private void addLineTo(double[] coords){
      path.add(coords);
      segment.add(PathIterator.SEG_LINETO);
    }
    
    public NarrowingTranscriptBoxPathIterator(){
      
      if(hasBlock[0]){
        if(hasBlock[1] && rounded){
          addCurve(new double[]{subBlocks[1].getX(),subBlocks[1].getY()}, 
              new double[]{subBlocks[0].getMaxX(), subBlocks[0].getY()},  new double[]{subBlocks[0].getX(), subBlocks[0].getY()});
        }else{
          addLineTo(new double[]{subBlocks[0].getMaxX(),subBlocks[0].getY()});
        }
        addLineTo(new double[]{subBlocks[0].getX(), subBlocks[0].getY()});
        
        addLineTo(new double[]{subBlocks[0].getX(),subBlocks[0].getMaxY()});
        
        if(hasBlock[1] && rounded){
          addCurve(new double[]{subBlocks[0].getX(),subBlocks[0].getMaxY()},
              new double[]{subBlocks[0].getMaxX(),subBlocks[0].getMaxY()}, new double[]{ subBlocks[1].getX(),  subBlocks[1].getMaxY()});
        }else{
          addLineTo(new double[]{subBlocks[0].getMaxX(),subBlocks[0].getMaxY()});
        }
      }
      if(hasBlock[1]){
        addLineTo(new double[]{subBlocks[1].getX(),subBlocks[1].getMaxY()});
        addLineTo(new double[]{subBlocks[1].getMaxX(),subBlocks[1].getMaxY()});
      }
      if(hasBlock[2]){
        if(hasBlock[1] && rounded){
          addCurve(new double[]{subBlocks[1].getMaxX(),subBlocks[1].getMaxY()},
              new double[]{subBlocks[2].getX(),subBlocks[2].getMaxY()}, new double[]{ subBlocks[2].getMaxX(),  subBlocks[2].getMaxY()});
        }else{
          addLineTo(new double[]{subBlocks[2].getX(),subBlocks[2].getMaxY()});
        }
        addLineTo(new double[]{subBlocks[2].getMaxX(),subBlocks[2].getMaxY()});
        addLineTo(new double[]{subBlocks[2].getMaxX(),subBlocks[2].getY()});
        if(hasBlock[1] && rounded){
          addCurve(new double[]{subBlocks[2].getMaxX(),subBlocks[2].getY()},
              new double[]{subBlocks[2].getX(),subBlocks[2].getY()}, new double[]{ subBlocks[1].getMaxX(),  subBlocks[1].getY()});
        }else{
          addLineTo(new double[]{subBlocks[2].getX(),subBlocks[2].getY()});
        }
      }
      if(hasBlock[1]){
        addLineTo(new double[]{subBlocks[1].getMaxX(),subBlocks[1].getY()});
        addLineTo(new double[]{subBlocks[1].getX(),subBlocks[1].getY()});
      }
      if(path.size()>0){
        path.add(0,path.get(0));
        segment.add(0,PathIterator.SEG_MOVETO);
      }
    }
    
    private int pos = 0; 
    
    @Override
    public int getWindingRule()
    {
      return PathIterator.WIND_EVEN_ODD;
    }
    

    @Override
    public boolean isDone()
    {
      return pos>path.size();
    }

    @Override
    public void next()
    {
      pos++;
    }

    @Override
    public int currentSegment(float[] coords)
    {
      double[] dcoords = new double[6];
      int ret = currentSegment(dcoords);
      for(int i = 0; i < coords.length; i++){
        coords[i] = (float) dcoords[i];  
      }
      return ret;
    }

    @Override
    public int currentSegment(double[] coords)
    {
      if(isDone()){
        throw new NoSuchElementException("Iterator out of bounds");
      }
      if(pos==path.size()){
        coords[0] = path.get(0)[0];
        coords[1] = path.get(0)[1];  
        return PathIterator.SEG_CLOSE;
      }
      for(int i = 0; i < path.get(pos).length; i++){
        coords[i] = path.get(pos)[i];  
      }
      
      return segment.get(pos);
    }
    
  }
  
}

