/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing.canvas;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.TreeSet;

import at.icbi.gpviz.data.regions.GPVVariant;

public class VariantShapeAxis
{
  
  //stores where the shapes vor variants should be displayed so that they don't overlap each other
  private float[] variantShapeDistribution;
  private GPVVariant[] variants;
  private HashMap<GPVVariant,Float> cachedDistribution;
  
  private float offset;
  private float available_space;
  private float margin;
  private float minimum_distance;
  private float preferred_distance;
  
  private int[][] groups;
  
  public VariantShapeAxis(TreeMap<GPVVariant,Float> init_positions, float offset, float available_space, float shapewidth, boolean reverse_strand){
//    reverse_strand = false;
    this.variantShapeDistribution = new float[init_positions.size()];
    this.variants = new GPVVariant[init_positions.size()];
    this.offset = offset;
    this.available_space = available_space;
    
    int i = 0;
    for(Entry<GPVVariant,Float> entry:init_positions.entrySet()){
      int real_i = reverse_strand?init_positions.size()-i-1:i;
      variantShapeDistribution[real_i] = entry.getValue();
      variants[real_i] = entry.getKey();
      i++;
    }
    
    this.margin=shapewidth;
    
    this.minimum_distance =  shapewidth+margin;
    this.preferred_distance =  shapewidth+margin*2;
  }
  
  public void distribute(){
    
    int[] overlappers = null;
    groups = new int[0][0];

    boolean found_overlappers = false;
    int iteration = 0;
    do{
      found_overlappers = false;
      
      for(int i = 0; i < variantShapeDistribution.length && !found_overlappers; i++){
        overlappers = getOverlappingVariants(i);
        found_overlappers = overlappers.length>1;
      }
      if(found_overlappers){
        iteration++;
        //found overlappers
        overlappers = updateAndExpandGroup(overlappers);
        float[] overlappers_range = getGroupRange(overlappers);
        float space_per_shape = (overlappers_range[2]-overlappers_range[0])/overlappers.length;
        float curr_x = overlappers_range[0];
        
        for(int i = 0; i < overlappers.length; i++){
          variantShapeDistribution[overlappers[i]] = curr_x;
          curr_x+=space_per_shape;//maybe improve distribution
        }
      }
    }while(found_overlappers && groups.length!=1 && iteration<100);
    
    
    cachedDistribution = new HashMap<GPVVariant,Float>();
    for(int i = 0; i < variantShapeDistribution.length; i++){
      cachedDistribution.put(variants[i], variantShapeDistribution[i]);
    }
  }
  
  private int[] updateAndExpandGroup(int[] overlappers){
    HashSet<Integer> overlappers_and_friends = new HashSet<Integer>();
    ArrayList<Integer> obsolete_groups = new ArrayList<Integer>();
    for(int overlapper:overlappers){
      int is_in_group = getGroupThatVariantIsIn(overlapper);
      if(is_in_group>=0){
        //this variant has previously been in an overlapper group. lets find his friends
        for(int groupmember:groups[is_in_group]){
          overlappers_and_friends.add(groupmember);
        }
        obsolete_groups.add(is_in_group);
      }else{
        overlappers_and_friends.add(overlapper);
      }
    }
    
    int[] ret = listToArray(new TreeSet<Integer>(overlappers_and_friends));//sort it before converting it to a list
    int regrouped_i = 0;
    if(obsolete_groups.size()>0){
      int[][] regrouped = new int[groups.length-obsolete_groups.size()+1][];  
      for(int i = 0; i < groups.length; i++){
        if(!obsolete_groups.contains(i)){
          regrouped[regrouped_i++] = Arrays.copyOf(groups[i], groups[i].length);
        }
      }
      regrouped[regrouped_i] = ret;
    }
    
    return ret;
  }
  
  private int[] listToArray(Collection<Integer> list){
    int[] ret = new int[list.size()];
    Iterator<Integer> it = list.iterator();
    for(int i = 0; i< list.size();i++) ret[i] = it.next();
    return ret;
  }
  
  private int getGroupThatVariantIsIn(int var){
    for(int i = 0; i<groups.length; i++){
      for(int j = 0; j < groups[i].length; j++){
        if(groups[i][j]==var) return i;
      }
    }
    return -1;
  }
  
  //returns 3 floats. min, cent, max
  private float[] getGroupRange(int[] vars){
    float min = Float.MAX_VALUE;
    float max = Float.MIN_VALUE;
    for(int var:vars) {
      float pos = variantShapeDistribution[var];
      min = Math.min(min, pos);
      max = Math.max(max, pos);
    }
    
    min = Math.max(offset,min);
    max = Math.min(offset+available_space,max);
    
    
    float center = (max+min)/2f;
    float required_space = vars.length*preferred_distance;
    
    return new float[]{center-required_space/2f,center,center+required_space/2f};
  }
  
  private int[] getOverlappingVariants(int var){
    float pos = variantShapeDistribution[var];
    ArrayList<Integer> retlist = new ArrayList<Integer>();
    for(int i = 0; i < variantShapeDistribution.length; i++){
      float pos2 = variantShapeDistribution[i];
      if(Math.abs(pos2-pos)<minimum_distance){
        //they are too close together
        retlist.add(i);
      }else if((var>i && pos<pos2) || (var<i && pos>pos2)){
        //they are completely cross
        retlist.add(i);
      }
    }
    if(retlist.size()==0) return new int[0];
    int[] ret = listToArray(retlist);
    return ret;
  }
  
  
  public float getHorizontalPositionForVariant(GPVVariant variant){
    if(cachedDistribution==null){
      distribute();
    }
    return cachedDistribution.get(variant);
  }

}

