/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing.options;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.TreeSet;

import javax.swing.JComponent;
import javax.swing.JPanel;

import at.icbi.gpviz.data.GPVChromosome;
import at.icbi.gpviz.data.GPVDataContainer;
import at.icbi.gpviz.data.GPVDomain;
import at.icbi.gpviz.data.GPVDomainSet;
import at.icbi.gpviz.data.GPVExon;
import at.icbi.gpviz.data.GPVGene;
import at.icbi.gpviz.data.GPVProtein;
import at.icbi.gpviz.data.GPVTranscript;
import at.icbi.gpviz.drawing.colorscheme.AbstractGPVColorScheme;
import at.icbi.gpviz.swing.canvas.GPVCanvas;
import at.tugraz.genome.biojava.seq.DNAStrand;

public class CanvasPreviewPanel extends JPanel
{

  private static final long serialVersionUID = 30908239959008078L;
  private GPVCanvas canvas;
  private GPVDataContainer container;
  private JComponent parent;
  
  /**
   * Create the dialog.
   */
  public CanvasPreviewPanel(JComponent parent)
  {
    this.parent = parent;
    setLayout(new BorderLayout());
    this.container = new GPVDataContainer();
    GPVGene  gene = createDummyGene();
    HashMap<String,GPVGene> genes = new HashMap<String,GPVGene>();
    genes.put(gene.getGeneID(),gene);
    this.container.setGenes(genes);
    this.canvas = new GPVCanvas(this.container);
    this.canvas.setSelectedGene(gene);
    
    addDummyHighlighter();
    addDummyDomains();
    
    canvas.getProperties().setDrawStartStopCodonLines(true);
    add(this.canvas,BorderLayout.CENTER);
    
    parent.addComponentListener(new ComponentListener()
    {
      @Override public void componentShown(ComponentEvent e){ }
      @Override public void componentMoved(ComponentEvent e){}
      @Override public void componentHidden(ComponentEvent e){}
      @Override public void componentResized(ComponentEvent e)
      {
        setPreferredSize(new Dimension(400,(int)(e.getComponent().getHeight()*0.5f)));
        repaint();
      }
    });
  }
  
  public void slideVisible(final boolean vis){
    if(vis) setVisible(true);
    //make a sliding animation
    new Thread(){
      public void run(){
        int steps = 10;
        int animation_time = 50;
        int steptime = animation_time/steps;
        int dest = (int)(parent.getHeight()*0.5f);
        if(!vis){ 
          dest = 0;
        }
        
        int height = getHeight();
        int incr = (height-dest)/steps;
        for(int i = 0; i< steps; i++){
          height = height-incr;
          setPreferredSize(new Dimension(400,height));
          refreshCanvas();
          try{
            Thread.sleep(steptime);
          }catch (InterruptedException e){}
        }
        setPreferredSize(new Dimension(400,dest));
        
        if(!vis)setVisible(false);
      }
    }.start();
    
    
  }
  
  public void changeColorScheme(AbstractGPVColorScheme previewScheme){
    canvas.getProperties().setColorScheme(previewScheme);
  }

  private GPVGene createDummyGene(){
    GPVChromosome chrom = new GPVChromosome("chrX");
    GPVGene gene = new GPVGene("ABCB7",chrom);
    GPVTranscript tran1 = new GPVTranscript("NM_004299");
    TreeSet<GPVExon> exons = new TreeSet<GPVExon>();
    tran1.setStopCodon(74273207);
    tran1.setStartCodon(74376105);
    exons.add(new GPVExon(74273105,74273420, DNAStrand.REVERSE));
    exons.add(new GPVExon(74280058,74280165, DNAStrand.REVERSE));
    exons.add(new GPVExon(74282163,74282266, DNAStrand.REVERSE));
    exons.add(new GPVExon(74284905,74285076, DNAStrand.REVERSE));
    exons.add(new GPVExon(74288842,74288971, DNAStrand.REVERSE));
    exons.add(new GPVExon(74289126,74289289, DNAStrand.REVERSE));
    exons.add(new GPVExon(74290200,74290357, DNAStrand.REVERSE));
    exons.add(new GPVExon(74291344,74291518, DNAStrand.REVERSE));
    exons.add(new GPVExon(74293524,74293611, DNAStrand.REVERSE));
    exons.add(new GPVExon(74293703,74293791, DNAStrand.REVERSE));
    exons.add(new GPVExon(74295197,74295465, DNAStrand.REVERSE));
    exons.add(new GPVExon(74296357,74296489, DNAStrand.REVERSE));
    exons.add(new GPVExon(74318777,74318896, DNAStrand.REVERSE));
    exons.add(new GPVExon(74332721,74332807, DNAStrand.REVERSE));
    exons.add(new GPVExon(74334589,74334669, DNAStrand.REVERSE));
    exons.add(new GPVExon(74375940,74376132, DNAStrand.REVERSE));
    tran1.setExons(exons);
    tran1.setParent(gene);
    tran1.setResultingProtein(new GPVProtein("NP_004290"));
    HashMap<String,GPVTranscript> trans = new HashMap<String,GPVTranscript>();
    trans.put(tran1.getTranscriptId(),tran1);
    gene.setTranscripts(trans);
    gene.setStrand(DNAStrand.REVERSE);
    gene.updateFirstAndLastPosition();
    tran1.setStrand(DNAStrand.REVERSE);
    return gene;
  }
  
  private void addDummyHighlighter(){
    container.addRegionToHighlight(container.getChromosomeByName("chrX"), 74318777, 74332807, DNAStrand.REVERSE);
    container.addRegionToHighlight(container.getChromosomeByName("chrX"), 74293703, 74295465, DNAStrand.REVERSE);
    container.addRegionToHighlight(container.getChromosomeByName("chrX"), 74293550, 74293550, DNAStrand.REVERSE);
  }
  
  private void addDummyDomains(){
    GPVDomainSet set_a = new GPVDomainSet();
    set_a.setSetName("specific");
    GPVDomain domain = new GPVDomain();
    domain.setParentSet(set_a);
    domain.setDescription("ABCC_ATM1_transporter");
    domain.setProteinId("NP_004290");
    domain.setProteinStartPos(473);
    domain.setProteinEndPos(709);
    set_a.addDomain(domain);
    GPVDomainSet set_b = new GPVDomainSet();
    set_b.setSetName("superfamily");
    domain = new GPVDomain();
    domain.setParentSet(set_b);
    domain.setDescription("P-loop_NTPase superfamily");
    domain.setProteinId("NP_004290");
    domain.setProteinStartPos(473);
    domain.setProteinEndPos(709);
    set_b.addDomain(domain);
    domain = new GPVDomain();
    domain.setParentSet(set_b);
    domain.setDescription("ABC_membrane superfamily");
    domain.setProteinId("NP_004290");
    domain.setProteinStartPos(141);
    domain.setProteinEndPos(412);
    set_b.addDomain(domain);
    GPVDomainSet set_c = new GPVDomainSet();
    set_c.setSetName("multi-dom");
    set_b.addDomain(domain);
    domain = new GPVDomain();
    domain.setParentSet(set_c);
    domain.setDescription("AMT1");
    domain.setProteinId("NP_004290");
    domain.setProteinStartPos(211);
    domain.setProteinEndPos(706);
    set_c.addDomain(domain);
    ArrayList<GPVDomainSet> sets = new ArrayList<GPVDomainSet>();
    sets.add(set_a);
    sets.add(set_b);
    sets.add(set_c);
    container.setDomainSets(sets);
  }
  
  public void refreshCanvas(){
    canvas.refresh();
  }
}
