/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing.options;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.lang.reflect.Method;
import java.util.List;
import java.util.Random;

import javax.swing.AbstractCellEditor;
import javax.swing.ComboBoxEditor;
import javax.swing.JButton;
import javax.swing.JColorChooser;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JToggleButton;
import javax.swing.plaf.ButtonUI;
import javax.swing.plaf.metal.MetalButtonUI;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;

import at.icbi.gpviz.GPViz;
import at.icbi.gpviz.data.GPVOptions;
import at.icbi.gpviz.drawing.colorscheme.AbstractGPVColorScheme;
import at.icbi.gpviz.drawing.colorscheme.DefaultGPVColorScheme;
import at.icbi.gpviz.drawing.colorscheme.GPVColor;
import at.icbi.gpviz.drawing.colorscheme.GPVColorScheme;
import at.icbi.gpviz.drawing.colorscheme.UserDefinedGPVColorScheme;

public class ColorSchemePanel extends JPanel implements ItemListener, ActionListener, ProgramOptionTab, ComponentListener, KeyListener
{
  private static final long serialVersionUID = -1406974474802399699L;
  private static int COLUMN_NAME = 0;
  private static int COLUMN_COLOR = 1;
  
  private JTable table;
  private JComboBox<AbstractGPVColorScheme> comboBox = new JComboBox<AbstractGPVColorScheme>();
  private JButton btnNew = new JButton("New");
  private JButton btnSave = new JButton("Save");
  private JLabel messageLabel = new JLabel("");
  private final JButton btnDelete = new JButton("Delete");
  private final JPanel panel_2 = new JPanel();
  private final DomainColorChooserPanel domainColorsPanel = new DomainColorChooserPanel(this);

  private static final String NEW_SCHEME_NAME = "New Color Scheme";
  private final JToggleButton btnPreview = new JToggleButton("Preview");
  
  private UserDefinedGPVColorScheme previewScheme;
  
  protected CanvasPreviewPanel preview = null;
  
  /**
   * Create the panel.
   */
  @SuppressWarnings("serial")
  public ColorSchemePanel(JDialog parent)
  {
    setLayout(new BorderLayout(0, 10));

    JPanel panel = new JPanel();
    add(panel, BorderLayout.NORTH);
    panel.setLayout(new BorderLayout(5, 0));

    JLabel lblNewLabel = new JLabel("Color scheme:");
    panel.add(lblNewLabel, BorderLayout.WEST);
    
    panel.add(comboBox);
    comboBox.addItemListener(this);
    comboBox.setEditable(true);
    comboBox.setEditor(new ColorSchemeEditor());

    panel.add(btnNew, BorderLayout.EAST);
    btnNew.addActionListener(this);

    JPanel midpanel = new JPanel();
    preview = new CanvasPreviewPanel(midpanel);
    midpanel.setLayout(new BorderLayout(0,0));
    JScrollPane scrollPane = new JScrollPane();
    midpanel.add(scrollPane, BorderLayout.CENTER);
    midpanel.add(preview,BorderLayout.SOUTH);
    add(midpanel, BorderLayout.CENTER);

    table = new JTable();
    table.setCellSelectionEnabled(false);
    table.setRowSelectionAllowed(false);
    table.setColumnSelectionAllowed(false);
    table.setModel(new DefaultTableModel(
        new Object[][] {
        },
        new String[] {
            "Item", "Color"
        }
        ) {
      Class[] columnTypes = new Class[] {
          String.class, ColorButton.class
      };
      public Class getColumnClass(int columnIndex) {
        return columnTypes[columnIndex];
      }
      @Override
      public boolean isCellEditable(int row, int column) {
         return column==COLUMN_COLOR;
      }
      
    });

    table.getColumnModel().getColumn(COLUMN_NAME).setPreferredWidth(270);
    
    table.getColumnModel().getColumn(COLUMN_COLOR).setCellRenderer(new ColorRenderer());
    table.getColumnModel().getColumn(COLUMN_COLOR).setCellEditor(new ColorEditor(this));
    if(GPViz.AIMONK){
      table.addKeyListener(this);
    }
    
    scrollPane.setViewportView(table);

    JPanel panel_1 = new JPanel();
    add(panel_1, BorderLayout.SOUTH);
    panel_1.setLayout(new BorderLayout(0, 5));
    panel_1.add(domainColorsPanel, BorderLayout.CENTER);
    FlowLayout flowLayout = (FlowLayout) panel_2.getLayout();
    flowLayout.setAlignment(FlowLayout.RIGHT);
    
    panel_1.add(panel_2, BorderLayout.SOUTH);
    panel_2.add(messageLabel);
    
    messageLabel.setFont(new Font("Dialog", Font.BOLD, 10));
    
    panel_2.add(btnPreview);
    panel_2.add(btnDelete);
    panel_2.add(btnSave);
    btnSave.addActionListener(this);
    btnDelete.addActionListener(this);
    btnPreview.addActionListener(this);

    this.addComponentListener(this);
    
    initColorSchemeList();
    initTable();
    initProteinDomainColorList();
    updateEnabledDisabledStatus();
  }


  private AbstractGPVColorScheme getSelectedScheme(){
    return (AbstractGPVColorScheme) this.comboBox.getSelectedItem();
  }

  private void initColorSchemeList(){
    List<UserDefinedGPVColorScheme> all_color_schemes = UserDefinedGPVColorScheme.loadAll();
    this.comboBox.addItem(new DefaultGPVColorScheme());
    for(UserDefinedGPVColorScheme scheme:all_color_schemes){
      this.comboBox.addItem(scheme);
    }

  }

  private void initTable(){
    DefaultTableModel model = (DefaultTableModel) table.getModel();
    while(model.getRowCount()>0){
      model.removeRow(0);
    }

    AbstractGPVColorScheme selected = getSelectedScheme();
    if(selected!=null){

      Method[] methods = GPVColorScheme.class.getMethods();

      for(Method getter:methods){
        if(getter.isAnnotationPresent(GPVColor.class)){
          if(getter.getName().startsWith("get") && getter.getReturnType()==Color.class){
            String description = getter.getAnnotation(GPVColor.class).description();
            try
            {
              Color color = (Color) getter.invoke(selected, (Object[])null);
              ColorButton button = null;
              
              if(selected instanceof UserDefinedGPVColorScheme){
                String settername = "set"+getter.getName().substring(3);
                Method setter = selected.getClass().getMethod(settername, new Class[]{Color.class});
                if(setter!=null){
                  button = new ColorButton(description, setter, color);
                }
              }else{
                button = new ColorButton(description, null, color);
              }
              
              if(button!=null){
                model.addRow(new Object[]{description,button});
              }
              
              
            }
            catch (ReflectiveOperationException e)
            {
              System.err.println("Reflection error invoking getter for color " +description);
              e.printStackTrace();
            }
            catch (SecurityException e)
            {
              System.err.println("Reflection error finding setter for color " +description);
              e.printStackTrace();
            }
          }
        }
      }
    }
  }
  
  private void redrawPreview(){
    if(preview!=null && preview.isVisible())preview.refreshCanvas();
  }
  
  private void initProteinDomainColorList(){
    AbstractGPVColorScheme selected = getSelectedScheme();
    if(selected!=null){
      Color[] colors = selected.getDomainSetColorProvider().getColors();
      domainColorsPanel.setColors(colors);
    }
  }
  
  private void updateEnabledDisabledStatus(){
    AbstractGPVColorScheme selected = getSelectedScheme();
    if(selected == null || selected instanceof DefaultGPVColorScheme){
      messageLabel.setText("Click \"New\" to create a new color scheme");
      btnSave.setEnabled(false);
      btnDelete.setEnabled(false);
      domainColorsPanel.setEnabled(false);
    }else{
      messageLabel.setText("");
      btnSave.setEnabled(true);
      btnDelete.setEnabled(true);
      domainColorsPanel.setEnabled(true);
    }
  }
  
  
  public void saveCurrentColorScheme(){
    AbstractGPVColorScheme selected = getSelectedScheme();
    saveToColorScheme(selected);
  }
  
  
  public void savePreviewColorScheme(){
    saveToColorScheme(previewScheme);
    preview.changeColorScheme(previewScheme);
    preview.refreshCanvas();
  }
  
  public void saveToColorScheme(AbstractGPVColorScheme selected){
    if(selected!=null){
      if(selected instanceof UserDefinedGPVColorScheme){
        UserDefinedGPVColorScheme tosave =  ((UserDefinedGPVColorScheme)selected);
        
        for(int i = 0; i < table.getModel().getRowCount(); i++){
          ColorButton color = (ColorButton)table.getValueAt(i, COLUMN_COLOR);
          color.saveToColorScheme(tosave);
        }
        
        if(tosave.getName()==null || tosave.getName().equals("")){
          tosave.setName(getNewSchemeName());
        }
        
        Color[] domainColors = domainColorsPanel.getColors();
        if(domainColors.length<=0)domainColors = new DefaultGPVColorScheme().getDomainSetColorProvider().getColors();
        tosave.setDomainSetColors(domainColors);
        UserDefinedGPVColorScheme.save(tosave);
      }
    }
  }
  
  
  public void deleteCurrentColorScheme(){
    AbstractGPVColorScheme selected = getSelectedScheme();
    if(selected!=null){
      if(selected instanceof UserDefinedGPVColorScheme){
        UserDefinedGPVColorScheme todelete =  ((UserDefinedGPVColorScheme)selected);
        
        int answer = JOptionPane.showConfirmDialog(this, "Delete color scheme "+todelete.getName()+"?", "Confirm deletion", JOptionPane.YES_NO_OPTION);
        if(answer==JOptionPane.YES_OPTION){
          UserDefinedGPVColorScheme.delete(todelete);
          comboBox.removeItem(todelete);
        }
      }
    }
  }

  @Override
  public void itemStateChanged(ItemEvent e)
  {
    if(e.getSource()==this.comboBox){
      initTable();
      initProteinDomainColorList();
      updateEnabledDisabledStatus();
      previewScheme = getSelectedScheme().clone();
      preview.changeColorScheme(previewScheme);
      redrawPreview();
    }
  }

  private String getNewSchemeName(){
    String ret = NEW_SCHEME_NAME;
    boolean ok;
    do{
      ok = true;
      for(int i = 0; i < comboBox.getItemCount(); i++){
        String name =  comboBox.getItemAt(i).getName();
        if(name.equals(ret)){
          ok=false;
          String num = name.substring(NEW_SCHEME_NAME.length()).trim();
          if(num.length()>0){
            Integer number = Integer.parseInt(num);
            ret = NEW_SCHEME_NAME+" "+(number+1);
          }else{
            ret = NEW_SCHEME_NAME+" "+2;
          }
        }
      }
    }while(!ok);
    
    return ret;
  }
  
  @Override
  public void actionPerformed(ActionEvent e)
  {
    if(e.getSource()==btnNew){
      AbstractGPVColorScheme last = getSelectedScheme();
      if(last==null) last = new DefaultGPVColorScheme();
      UserDefinedGPVColorScheme newscheme = last.clone();
      newscheme.setName(getNewSchemeName());
      this.comboBox.addItem(newscheme);
      this.comboBox.setSelectedItem(newscheme);
      this.comboBox.requestFocus();
      initTable();
    }else if(e.getSource()==btnSave){
      saveCurrentColorScheme();
    }else if(e.getSource()==btnDelete){
      deleteCurrentColorScheme();
    }else if(e.getSource()==btnPreview){
      preview.slideVisible(btnPreview.isSelected());
      if(preview.isVisible()) preview.refreshCanvas();
    }
  }

  @Override
  public String getTitle()
  {
    return "Color schemes";
  }

  @Override
  public void fillWithOptions(GPVOptions source)
  {
    if(source!=null){
      AbstractGPVColorScheme preferred = source.getPreferredColorScheme();
      if(preferred!=null){
        for(int i = 0; i < comboBox.getItemCount(); i++){
          AbstractGPVColorScheme scheme = comboBox.getItemAt(i);
          if(scheme.getName().equals(preferred.getName())){
            comboBox.setSelectedIndex(i);
          }
        }
      }
    }
  }

  @Override
  public boolean validateOptions()
  {
    return true;
  }

  public void save(GPVOptions options){
    saveCurrentColorScheme();
    AbstractGPVColorScheme selected = getSelectedScheme();
    if(selected==null)selected = new DefaultGPVColorScheme();
    options.setPreferredColorScheme(selected);
  }

  @Override public void componentResized(ComponentEvent e){}
  @Override public void componentMoved(ComponentEvent e){}
  @Override public void componentShown(ComponentEvent e){}

  @Override
  public void componentHidden(ComponentEvent e)
  {
    if(e.getSource()==this){
      if(preview!=null)preview.setVisible(false);
    }
  }


  @Override public void keyTyped(KeyEvent e){}
  @Override public void keyPressed(KeyEvent e)
  {
    if(e.getSource()==table){
      if(e.isAltDown() && (e.getKeyChar()=='r' || e.getKeyChar()=='R')){
        Random r = new Random();
        for(int i = 0; i < table.getRowCount(); i++){
          Object val = table.getValueAt(i, COLUMN_COLOR);
          if(val instanceof ColorButton){
            ColorButton but = (ColorButton) val;
            if(!(but.purpose.endsWith("background")||but.purpose.endsWith("borders")||but.purpose.endsWith("text color")||but.purpose.endsWith("rails")) || !e.isShiftDown()){
              but.selectedColor = new Color(r.nextFloat(), r.nextFloat(), r.nextFloat());
            }
            but.saveToColorScheme(previewScheme);
          }
        }
        Color[] domains = new Color[6];
        for(int i = 0; i < 6; i++){
          domains[i] = new Color(r.nextFloat(), r.nextFloat(), r.nextFloat());
        }
        domainColorsPanel.setColors(domains);
        previewScheme.setDomainSetColors(domains);
        table.repaint();
        savePreviewColorScheme();
        redrawPreview();
      }else if(e.isAltDown() && e.getKeyChar()=='e'){
        Random r = new Random();
        Object val = table.getValueAt(table.getSelectedRow(), COLUMN_COLOR);
        if(val instanceof ColorButton){
          ColorButton but = (ColorButton) val;
          but.selectedColor = new Color(r.nextFloat(), r.nextFloat(), r.nextFloat());
          but.saveToColorScheme(previewScheme);
        }
        table.repaint();
        redrawPreview();
      }
      
    }
  }
  @Override public void keyReleased(KeyEvent e){}
}

class ColorButton {
  Color selectedColor;
  Method setter;
  String purpose;

  public ColorButton(String purpose, Method setter, Color initcolor){
    super();
    this.purpose = purpose;
    this.setter = setter;
    this.selectedColor = initcolor;
  }
  
  public void saveToColorScheme(UserDefinedGPVColorScheme colorscheme){
    try
    {
      setter.invoke(colorscheme, selectedColor);
    }
    catch (ReflectiveOperationException e)
    {
      System.err.println("Reflection error setting color "+setter.getName()+" on "+colorscheme.getClass().getName());
      e.printStackTrace();
    }
    catch (IllegalArgumentException e)
    {
      System.err.println("Reflection error setting color "+setter.getName()+" on "+colorscheme.getClass().getName());
      e.printStackTrace();
    }
  }
}

class ColorEditor extends AbstractCellEditor implements TableCellEditor, ActionListener {
  private static final long serialVersionUID = 8620361554807778793L;
  private ColorButton currentColor;
  private JButton button;
  private ColorSchemePanel parent;

  public ColorEditor(ColorSchemePanel parent) {
    this.button = new JButton();
    this.button.addActionListener(this);
    button.setBorderPainted(false);
    this.parent = parent;
  }

  public void actionPerformed(ActionEvent e) {
    button.setBackground(currentColor.selectedColor);
    Color chosen = JColorChooser.showDialog(this.parent, "Select color for "+currentColor.purpose,currentColor.selectedColor);
    if(chosen != null) currentColor.selectedColor = chosen;
    parent.savePreviewColorScheme();
    fireEditingStopped(); //Make the renderer reappear.
  }

  public Object getCellEditorValue() {
    return currentColor;
  }

  public Component getTableCellEditorComponent(JTable table,
      Object value,
      boolean isSelected,
      int row,
      int column) {
    currentColor = (ColorButton)value;
    button.setEnabled(currentColor.setter!=null);
    return button;
  }
}

class ColorRenderer extends JButton implements TableCellRenderer {
  private static final long serialVersionUID = -6897481695264763448L;
  private static final ButtonUI buttonUI = new ColorButtonUI(); 

  public Component getTableCellRendererComponent(
      JTable table, Object color,
      boolean isSelected, boolean hasFocus,
      int row, int column) {
    ColorButton newColor = (ColorButton)color;
    //We need to set the ButtonUI, or else the color buttons won't show properly on a mac
    this.setUI(buttonUI);
    this.setBackground(newColor.selectedColor);
    this.setBorderPainted(false);
    this.setEnabled(newColor.setter!=null);
    return this;
  }
}

class ColorSchemeEditor implements ComboBoxEditor{
  JTextField tf = new JTextField();  
  AbstractGPVColorScheme toEdit;
  
  @Override
  public Component getEditorComponent()
  {
    return tf;
  }

  @Override
  public void setItem(Object o)
  {
    if ( o != null && o instanceof AbstractGPVColorScheme) {
      toEdit = ((AbstractGPVColorScheme)o);
      tf.setText(toEdit.getName());
    }  
  }

  @Override
  public Object getItem()
  {
    if ( toEdit != null && toEdit instanceof UserDefinedGPVColorScheme) {
      ((UserDefinedGPVColorScheme)toEdit).setName(tf.getText());
    }
    return toEdit;
  }

  @Override
  public void selectAll()
  {
    tf.selectAll();
  }
  @Override
  public void addActionListener(ActionListener l)
  {
    tf.addActionListener( l );  
  }
  @Override
  public void removeActionListener(ActionListener l)
  {
    tf.removeActionListener(l);
  }
  
}

class ColorButtonUI extends MetalButtonUI{
  public ColorButtonUI()
  {
  }
  
  @Override
  public void paint(Graphics g, JComponent b){
    super.paint(g, b);
    
  }
}

