/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.swing.options;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.io.Serializable;

import javax.swing.BorderFactory;
import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.DropMode;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JColorChooser;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListCellRenderer;
import javax.swing.ListSelectionModel;
import javax.swing.ScrollPaneConstants;
import javax.swing.TransferHandler;
import javax.swing.border.Border;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

public class DomainColorChooserPanel extends JPanel implements ListSelectionListener, ActionListener
{
  private JList<Color> list;
  private JButton btnAdd = new JButton("Add");
  private JButton btnRemove = new JButton("Remove");
  private ColorSchemePanel parent;
  
  /**
   * Create the panel.
   */
  public DomainColorChooserPanel(ColorSchemePanel parent)
  {
    this.parent = parent;
    setLayout(new BorderLayout(0, 0));
    
    JPanel panel = new JPanel();
    FlowLayout flowLayout = (FlowLayout) panel.getLayout();
    flowLayout.setVgap(0);
    flowLayout.setHgap(0);
    add(panel, BorderLayout.EAST);
    
    panel.add(btnAdd);
    panel.add(btnRemove);
    btnAdd.addActionListener(this);
    btnRemove.addActionListener(this);
    btnRemove.setEnabled(false);
    
    JLabel lblNewLabel = new JLabel("Choose Domain Colors:");
    add(lblNewLabel, BorderLayout.NORTH);
    
    JScrollPane scrollPane = new JScrollPane();
    scrollPane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_NEVER);
    scrollPane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);
    add(scrollPane, BorderLayout.CENTER);

    list = new JList<Color>();
    scrollPane.setViewportView(list);
    list.setCellRenderer(new ImageCellRenderer());
    list.setLayoutOrientation(JList.HORIZONTAL_WRAP);
    list.setVisibleRowCount(1);
    list.setSelectionMode(ListSelectionModel.SINGLE_INTERVAL_SELECTION);
    list.setFixedCellWidth(30);
    list.setFixedCellHeight(30);
    list.setDragEnabled(true);
    list.setDropMode(DropMode.INSERT);
    list.setTransferHandler(new ImageTransferHandler(list));

    list.setBackground(this.getBackground());
    list.addListSelectionListener(this);
    this.validate();
  }
  
  public void setColors(Color[] initColors){
    DefaultListModel<Color> listModel  = new DefaultListModel<Color>();
    
    for(Color color:initColors){
      listModel.addElement(color);
    }
    
    list.setModel(listModel);
  }
  
  public void setEnabled(boolean enabled){
    super.setEnabled(enabled);
    list.setEnabled(enabled);
    btnAdd.setEnabled(enabled);
    btnRemove.setEnabled(enabled && list.getMinSelectionIndex()>=0);
  }

  @Override
  public void valueChanged(ListSelectionEvent e)
  {
    if(e.getSource()==list){
      btnRemove.setEnabled(list.getMinSelectionIndex()>=0 && this.isEnabled());
    }
    parent.savePreviewColorScheme();
  }

  @Override
  public void actionPerformed(ActionEvent e)
  {
    if(e.getSource()==btnAdd){
      Color chosen = JColorChooser.showDialog(this, "Choose color for protein domains", Color.red);
      if(chosen!=null){
        ((DefaultListModel<Color>)list.getModel()).addElement(chosen);
      }
      parent.savePreviewColorScheme();
    }else if(e.getSource()==btnRemove){
      int selected_from = list.getMinSelectionIndex();
      int selected_to = list.getMaxSelectionIndex();
      for(int i = selected_from; i <= selected_to; i++){
        ((DefaultListModel<Color>)list.getModel()).remove(selected_from);
      }
      parent.savePreviewColorScheme();
    }
  }
  
  public Color[] getColors(){
    DefaultListModel<Color> listModel = (DefaultListModel<Color>) list.getModel();
    Color[] ret = new Color[listModel.getSize()];
    for(int i = 0; i < ret.length; i++){
      ret[i] = listModel.get(i);
    }
    return ret;
  }
}

class ImageTransferHandler extends TransferHandler {

  private static final long serialVersionUID = 5552119483728468611L;
  private static final DataFlavor DATA_FLAVOUR = new DataFlavor(ColorIcon.class, "Images");

  private final JList<Color> previewList;
  private boolean inDrag;

  ImageTransferHandler(JList<Color> previewList) {
      this.previewList = previewList;
  }

  public int getSourceActions(JComponent c) {
      return TransferHandler.MOVE;
  }

  protected Transferable createTransferable(JComponent c) {
      inDrag = true;
      return new Transferable() {
          public DataFlavor[] getTransferDataFlavors() {
              return new DataFlavor[] {DATA_FLAVOUR};
          }

          public boolean isDataFlavorSupported(DataFlavor flavor) {
              return flavor.equals(DATA_FLAVOUR);
          }

          public Color getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException {
              return previewList.getSelectedValue();
          }
      };
  }

  public boolean canImport(TransferSupport support) {
      if (!inDrag || !support.isDataFlavorSupported(DATA_FLAVOUR)) {
          return false;
      }

      JList.DropLocation dl = (JList.DropLocation)support.getDropLocation();
      if (dl.getIndex() == -1) {
          return false;
      } else {
          return true;
      }
  }

  public boolean importData(TransferSupport support) {
      if (!canImport(support)) {
          return false;
      }

      Transferable transferable = support.getTransferable();
      try {
          Color draggedImage = (Color)transferable.getTransferData(DATA_FLAVOUR);

          JList.DropLocation dl = (JList.DropLocation)support.getDropLocation();
          DefaultListModel<Color> model = (DefaultListModel<Color>)previewList.getModel();
          int dropIndex = dl.getIndex();
          if (model.indexOf(draggedImage) < dropIndex) {
              dropIndex--;
          }
          model.removeElement(draggedImage);
          model.add(dropIndex, draggedImage);
          return true;
      } catch (Exception e) {
          e.printStackTrace();
          return false;
      }
  }

  protected void exportDone(JComponent source, Transferable data, int action) {
      super.exportDone(source, data, action);
      inDrag = false;
  }
}

class ImageCellRenderer extends JPanel implements ListCellRenderer<Color> {

  private static final long serialVersionUID = 7615446646316804605L;
  DefaultListCellRenderer defaultListCellRenderer = new DefaultListCellRenderer();
  JLabel imageLabel = new JLabel();

  ImageCellRenderer() {
      setLayout(new BorderLayout());
      Border emptyBorder = BorderFactory.createEmptyBorder(2, 2, 2, 2);
      imageLabel.setBorder(emptyBorder);
      add(imageLabel, BorderLayout.CENTER);
  }

  public ImageCellRenderer getListCellRendererComponent(JList<? extends Color> list, Color value, int index, boolean isSelected, boolean cellHasFocus) {
      defaultListCellRenderer.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
      setBorder(defaultListCellRenderer.getBorder());
      setBackground(defaultListCellRenderer.getBackground());
      imageLabel.setIcon((Icon)new ColorIcon(value));
      return this;
  }
}

class ColorIcon implements Icon, Serializable {
  private static final long serialVersionUID = 5968708688430808140L;
  private Color color;

  ColorIcon(Color color) {
      this.color = color;
  }

  public void paintIcon(Component c, Graphics g, int x, int y) {
      g.setColor(color);
      g.fillRect(x, y, getIconWidth(), getIconHeight());
  }

  public int getIconWidth() {
      return 30;
  }

  public int getIconHeight() {
      return 30;
  }

  public boolean equals(Object o) {
      if (o == null || getClass() != o.getClass()) {
          return false;
      }
      return color.equals(((ColorIcon)o).color);
  }

}



