/**
 *   GPViz: software for dynamic gene-centered visualization of genomic regions and/or variants
 *    Copyright (C) 2013  Rene Snajder
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package at.icbi.gpviz.weblinks;

import java.awt.Desktop;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;

import at.icbi.gpviz.data.GPVGene;
import at.icbi.gpviz.data.GPVProtein;
import at.icbi.gpviz.data.GPVTranscript;
import at.icbi.gpviz.data.regions.GPVHighlightedRegion;

public class WebLink
{
  public static final String VARIABLE_GENEID = "gene_id";
  public static final String VARIABLE_UNIGENEID = "unigene_id";
  public static final String VARIABLE_GENENAME = "gene_name";
  public static final String VARIABLE_GENESYMBOL = "gene_symbol";
  public static final String VARIABLE_TRANSCRIPT_REFSEQ = "transcript_refseq";
  public static final String VARIABLE_PROTEIN_REFSEQ = "protein_refseq";
  public static final String VARIABLE_PROTEIN_ID = "protein_id";
  
  public static final String VARIABLE_CHROMOSOME = "chrom";
  public static final String VARIABLE_STARTPOS = "start";
  public static final String VARIABLE_STOPPOS = "stop";
  
  
  public static String[] getVariablesForGeneLink(){
    return new String[]{VARIABLE_GENEID, VARIABLE_UNIGENEID, VARIABLE_GENESYMBOL, VARIABLE_GENENAME};
  }
  
  public static String[] getVariablesForTranscriptLink(){
    return new String[]{VARIABLE_TRANSCRIPT_REFSEQ};
  }
  
  public static String[] getVariablesForProteinLink(){
    return new String[]{VARIABLE_PROTEIN_REFSEQ, VARIABLE_PROTEIN_ID};
  }
  
  public static String[] getVariablesForRegionLink(){
    return new String[]{VARIABLE_CHROMOSOME, VARIABLE_STARTPOS, VARIABLE_STOPPOS};
  }
  
  /**
   * Adds the variable stuff ${} around a variable.
   */
  public static String decorateVariable(String variable){
    return "${"+variable+"}";
  }
  
  private static boolean containsVariable(String link, String[] variables){
    for(String variable:variables){
      if(link.contains("$"+variable) || link.contains("${"+variable+"}")){
        return true;
      }
    }
    return false;
  }

  private static String replaceVariable(String link, String variable, String value){
    if(value!=null){
      link = link.replaceAll("\\$"+variable, value);
      link = link.replaceAll("\\$\\{"+variable+"\\}", value);
    }
    return link;
  }

  public static String resolveGeneLink(String link, GPVGene gene){

    String ret = replaceVariable(link,VARIABLE_GENEID, gene.getGeneID());
    ret = replaceVariable(ret,VARIABLE_UNIGENEID, gene.getUniGeneID());
    ret = replaceVariable(ret,VARIABLE_GENESYMBOL, gene.getGeneSymbol());
    ret = replaceVariable(ret,VARIABLE_GENENAME, gene.getGeneFullName());
    if(containsVariable(ret,getVariablesForGeneLink())){
      //there are variables left. That means we are missing information...
      return null;
    }
    return ret;
  }
  public static String resolveTranscriptLink(String link, GPVTranscript transcript){

    String ret = replaceVariable(link,VARIABLE_TRANSCRIPT_REFSEQ, transcript.getTranscriptId());
    if(containsVariable(ret,getVariablesForTranscriptLink())){
      //there are variables left. That means we are missing information...
      return null;
    }
    return ret;
  }
  public static String resolveProteinLink(String link, GPVProtein protein){
    String ret = replaceVariable(link,VARIABLE_PROTEIN_REFSEQ, protein.getProteinID());
    ret = replaceVariable(ret,VARIABLE_PROTEIN_ID, protein.getProteinUniprotId());
    if(containsVariable(ret,getVariablesForProteinLink())){
      //there are variables left. That means we are missing information...
      return null;
    }
    return ret;
  }
  
  public static String resolveRegionLink(String link, GPVGene gene){
    return resolveRegionLink(link, gene.getChromosome().getTrimmedName(), gene.getFirstKnownPosition(), gene.getLastKnownPosition());
  }
  
  public static String resolveRegionLink(String link, GPVHighlightedRegion region){
    return resolveRegionLink(link, region.getChromosome().getTrimmedName(), region.getStart(), region.getStop());
  }
  public static String resolveRegionLink(String link, String chrom, int start, int stop){
    String ret = replaceVariable(link,VARIABLE_CHROMOSOME, chrom);
    ret = replaceVariable(ret,VARIABLE_STARTPOS, start+"");
    ret = replaceVariable(ret,VARIABLE_STOPPOS, stop+"");
    if(containsVariable(ret,getVariablesForProteinLink())){
      //there are variables left. That means we are missing information...
      return null;
    }
    return ret;
  }

  public static void visitLink(String link_im_hovering_over)
  {
    try
    {
      Desktop.getDesktop().browse(new URI(link_im_hovering_over));
    }
    catch (IOException e1)
    {
      e1.printStackTrace();
    }
    catch (URISyntaxException e1)
    {
      e1.printStackTrace();
    }
  }
  
}
